/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.junit.lib;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.InputStream;
import java.io.StringWriter;
import java.util.Arrays;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.jabhts.runtime.html.HtmlPrintWriter;
import org.jabhts.runtime.lib.TemplateJavaMD5Exception;
import org.jabhts.translator.compiler.htmlLoader.HtmlLoader;
import org.jabhts.translator.compiler.utils.ByteBuffer;
import org.jabhts.translator.compiler.utils.CompilerLog;
import org.jabhts.translator.compiler.utils.FileUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

abstract public class TestBase
{
	public final HtmlPrintWriter pw;
	private final StringWriter sw;

	protected final String defaultHtmlBasename = getClass().getSimpleName();
	protected String htmlBasename = defaultHtmlBasename;

	public static File baseDirF;

	protected TestBase()
	{
		sw = new StringWriter();
		pw = new HtmlPrintWriter(sw);
	}

	public void test() throws Exception
	{
		try {
			emitTemplate();
			validate();
		}
		catch (TemplateJavaMD5Exception ex) {
			throw ex;
			//Assert.fail("MD5 mismatch: " + ex.getMessage());
		}
	}

	abstract protected void emitTemplate() throws Exception;

	protected void validate() throws Exception
	{
		pw.flush();
		String actual = sw.toString();
		//
		File actualF = getHtmlF(true);
		softSave(actualF, actual); // save so we can use html5lib
		actual = getReformattedHtml(actualF);
		softSave(actualF, actual); // and then save the html5lib version
		//
		File expectedF = getHtmlF(false);
		//String expected = new String(FileUtils.readFile(expectedF), "utf-8");
		String expected = getReformattedHtml(expectedF);
		//
		if (actual.equals(expected))
			return;
		//
		System.err.println("####### Failure in " + getClass().getName());
		if (!htmlBasename.equals(defaultHtmlBasename))
			System.err.println("   using html basename: " + htmlBasename);
		String tmpName = expectedF.getName().replaceFirst("(\\.[^.]+)$", "-tmp$1");
		File tmpF = new File(actualF.getParentFile(), tmpName);
		try {
			FileUtils.saveFile(tmpF, expected.getBytes("utf-8"));
			makeDiff(tmpF,  actualF);
		}
		finally {
			tmpF.delete();
		}
		throw new SilentTestException();
	}

	protected File getHtmlF(boolean actual) throws Exception
	{
		String basename = (actual ? defaultHtmlBasename : htmlBasename + "-expected");
		//String basedir = System.getenv("JAT_BASEDIR");
		//if (basedir == null)
		//	throw new Exception("Missing environment variable JAT_BASEDIR");
		File f = new File(getClass().getName().replace('.', '/')).getParentFile();
		if (actual)
			f = new File(f.getParentFile().getParentFile(), "actualOutput");
		return new File(baseDirF, f.getPath() + "/" + basename + ".html");
	}

	protected void softSave(File f, String contents) throws Exception
	{
		if (!contents.contains("charset=utf-8")) {
			int idx = contents.indexOf("<head>");
			if (idx > 0)
				contents = contents.substring(0, idx + 6)
					+ "\n<meta http-equiv='Content-Type' content='text/html; charset=utf-8'>\n"
					+ contents.substring(idx + 6);
			//System.err.println(contents);
		}
		byte[] newData = contents.getBytes("utf-8");
		if (f.exists()) {
			byte[] oldData = FileUtils.readFile(f);
			if (Arrays.equals(newData, oldData))
				return;
		}
		FileUtils.saveFile(f, newData);
	}

	protected String getReformattedHtml(File f) throws Exception
	{
		Document doc = new HtmlLoader(new CompilerLog("reformat")).load(f).doc;
		cleanDocumentForComparison(doc);
		StringWriter output = new StringWriter(1000);
		DOMSource domSource = new DOMSource(doc);
		StreamResult streamResult = new StreamResult(output);
		TransformerFactory tf = TransformerFactory.newInstance();
		Transformer serializer = tf.newTransformer();
		serializer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
		//serializer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM,"users.dtd");
		serializer.setOutputProperty(OutputKeys.INDENT, "yes");
		serializer.transform(domSource, streamResult);
		String s = output.toString();
		return s.replaceAll("\\n([ \\t]*\\n)+", "\n")
				.replaceAll("([^<>\\s])[ \\t]*\\n</div>", "$1</div>")
				.replaceAll("\\n[ \\t]+", "\n");
	}

	protected void cleanDocumentForComparison(Document doc) throws Exception
	{
		Element htmlE = doc.getDocumentElement();
		if (htmlE == null)
			return;
		Element headE = getHead(htmlE);
		if (headE == null)
			return;
		Node next;
		for (Node child = headE.getFirstChild(); child != null; child = next) {
			next = child.getNextSibling();
			if (child.getNodeType() != Node.ELEMENT_NODE) {
				headE.removeChild(child);
				continue;
			}
			if (!child.getNodeName().equalsIgnoreCase("meta"))
				continue;
			Element childE = (Element)child;
			if (!"Content-Type".equalsIgnoreCase(childE.getAttribute("http-equiv")))
				continue;
			headE.removeChild(child);
		}
	}

	protected Element getHead(Element htmlE)
	{
		for (Node n = htmlE.getFirstChild(); n != null; n = n.getNextSibling()) {
			if (n.getNodeType() == Node.ELEMENT_NODE && n.getNodeName().equalsIgnoreCase("head"))
				return (Element)n;
		}
		return null;
	}

	protected void makeDiff(File expectedF, File actualF) throws Exception
	{
		Process p = Runtime.getRuntime().exec(new String[]{
				"diff",
				"--unified",
				//"--ignore-space-change",
				//"--ignore-blank-lines",
				"--report-identical-files",
				expectedF.getPath(),
				actualF.getPath()});
		p.getOutputStream().close();
		InputStream is = new BufferedInputStream(p.getInputStream(), 2048);
		p.getErrorStream().close();
		int i;
		ByteBuffer bb = new ByteBuffer(100);
		while ((i = is.read()) != -1)
			bb.append(i);
		p.destroy();
		System.err.println(bb.toString("utf-8"));
	}
}
