/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.runtime.html;

import org.jetbrains.annotations.Nullable;

public class HtmlAttr
{
	/**
	 * Includes single quotes in returned string, if needed.
	 */
	public static String quoted(@Nullable String s)
	{
		if (s == null || s.isEmpty())
			return "''";
		if (!needsQuoting(s))
			return s;
		int len = s.length();
		StringBuilder sb = new StringBuilder(len + 6);
		quoted(sb, s);
		return sb.toString();
	}

	/**
	 * Does always include single quotes in returned string.
	 */
	public static String forceQuoted(@Nullable String s)
	{
		if (s == null || s.isEmpty())
			return "''";
		int len = s.length();
		StringBuilder sb = new StringBuilder(len + 6);
		sb.append('\'');
		add(sb, s, false);
		sb.append('\'');
		return sb.toString();
	}

	/**
	 * Never any quotes, you must *always* supply those.
	 */
	public static String escape(@Nullable String s)
	{
		if (s == null)
			return "";
		if (!needsEscaping(s))
			return s;
		int len = s.length();
		StringBuilder sb = new StringBuilder(len + 6);
		escape(sb, s);
		return sb.toString();
	}

	public static void escape(StringBuilder sb, @Nullable String s)
	{
		if (s == null || s.isEmpty())
			return;
		if (!needsEscaping(s)) {
			sb.append(s);
			return;
		}
		add(sb, s, true);
	}

	public static void quoted(StringBuilder sb, @Nullable String s)
	{
		if (s == null || s.isEmpty()) {
			sb.append("''");
			return;
		}
		if (!needsQuoting(s)) {
			sb.append(s);
			return;
		}
		sb.ensureCapacity(sb.length() + s.length() + 6);
		sb.append('\'');
		add(sb, s, false);
		sb.append('\'');
	}

	/**
	 * html 4.0.1, section 3.2.2 Attributes
	 */
	public static boolean needsQuoting(@Nullable String s)
	{
		if (s == null)
			return true;
		int len = s.length();
		if (len == 0)
			return true;
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch >= '0' && ch <= '9')
				continue;
			if (ch >= 'a' && ch <= 'z')
				continue;
			if (ch >= 'A' && ch <= 'Z')
				continue;
			if (ch == '.' || ch == '-' || ch == '_')
				continue;
			return true;
		}
		return false;
	}

	public static boolean needsEscaping(@Nullable String s)
	{
		if (s == null)
			return true;
		int len = s.length();
		if (len == 0)
			return true;
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			switch (ch) {
				case '\'':
				case '"':
				case '&':
				case '<':
				case '>':
				case (char)127:
					return true;
			}
		}
		return false;
	}

	private static void add(StringBuilder sb, @Nullable String s, boolean escapeDoubleQuote)
	{
		if (s == null)
			return;
		int len = s.length();
		sb.ensureCapacity(sb.length() + len);
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			switch (ch) {
				case '&':
					sb.append("&amp;");
					break;
				case '<':
					sb.append("&lt;");
					break;
				case '>':
					sb.append("&gt;");
					break;
				case '\'':
					sb.append("&#39;");
					break;
				case '"':
					if (escapeDoubleQuote)
						sb.append("&#34;");
					else
						sb.append(ch);
					break;
				default:
					sb.append(ch);
					break;
			}
		}
	}
}
