/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.buildsystem;

import java.io.File;
import org.jabhts.annotation.global.TCompile;
import org.jabhts.translator.compiler.AnnotationCompiler;
import org.jabhts.translator.compiler.utils.AnnotationHelper;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jabhts.translator.compiler.utils.CompilerLog;
import org.jabhts.translator.compiler.utils.CompilerUtils;
import org.jabhts.translator.compiler.utils.FileUtils;
import org.jabhts.translator.compiler.utils.LogLevel;
import org.jabhts.translator.globals.CompilerOptions;
import org.jabhts.translator.globals.FileInfo;
import org.jabhts.translator.globals.PersistentCompilationInfo;
import org.jabhts.translator.globals.ProgressInfo;

public class FileCompiler
{
	public final CompilerOptions options;
	public final FileInfo javaFI;
	public final String dottedPackageName, shortClassname, fullClassname;
	public final CompilerLog log;
	public final SourceInfo oldSourceInfo;
	public final SourceInfo newSourceInfo;

	public Class<?> loadedJavaClass;
	public TCompileSet tcompileSet;

	public final ProgressInfo progressInfo;
	public static final String marker = "@" + TCompile.class.getSimpleName();

	public FileCompiler(ProgressInfo progressInfo, File javaF, String dottedPackageName, CompilerOptions options)
	{
		this.progressInfo = progressInfo;
		this.options = options;
		this.javaFI = new FileInfo(javaF);
		this.dottedPackageName = dottedPackageName;
		this.shortClassname = FileUtils.basename(javaFI.f);
		this.fullClassname = dottedPackageName + "." + shortClassname;
		this.log = new CompilerLog("Log for " + fullClassname);
		this.oldSourceInfo = options.sourceInfoMap.get(javaFI.f);
		this.newSourceInfo = new SourceInfo(javaFI, dottedPackageName);
	}

	public void compileFile() throws Exception
	{
		if (!shouldCompileThisClass())
			return;
		try {
			if (oldSourceInfo != null) {
				FileInfo info = oldSourceInfo.getModificatedSourceFile();
				if (info != null) {
					log.log(LogLevel.Verbose, "Source file changed: " + info.f.getPath());
					log.log(LogLevel.Verbose, "old = " + info.moddate + ",   now = " + info.f.lastModified() + "  =  " + (info.f.lastModified() - info.moddate));
				} else if (options.forceRecompilation) {
					log.log(LogLevel.Verbose, "No source files changed: forced recompilation");
				} else {
					log.log(LogLevel.Verbose, "No source files changed: no recompilation.");
					progressInfo.numNonCompiledTemplates += oldSourceInfo.compilationMap.size();
					return;
				}
			}
			String contents = new String(FileUtils.readFile(javaFI.f), options.sourceEncoding);
			boolean hasTHtml = contents.contains(marker);
			if (!hasTHtml) {
				progressInfo.numIgnoredSources++;
				if (oldSourceInfo != null) {
					log.log(LogLevel.Notice, "File doesn't contain " + marker + " anymore.");
					options.sourceInfoMap.remove(javaFI.f);
				}
				return;
			}
			progressInfo.startsFile(newSourceInfo.javaFI);
			if (oldSourceInfo == null)
				log.log(LogLevel.Verbose, "No source info: must recompile.");
			compileClass();
			if (!log.hasError())
				options.sourceInfoMap.put(javaFI.f, newSourceInfo);
			return;
		}
		catch (Exception ex) {
			log.log(LogLevel.Error, "Got exception: " + CompilerUtils.getSomeMessage(ex));
			throw ex;
		}
		finally {
			emitLog();
		}
	}

	public boolean shouldCompileThisClass() throws CompilerException
	{
		return options.compileClassFilter == null
				|| options.compileClassFilter.contains(fullClassname)
				|| options.compileClassFilter.contains(shortClassname);
	}

	protected void compileClass() throws CompilerException
	{
		//noinspection UnusedCatchParameter
		try {
			loadedJavaClass = Class.forName(fullClassname, false, options.classloader);
		}
		catch (ClassNotFoundException ex) {
			log.log(LogLevel.Error, "Can't load class: " + fullClassname);
			return;
		}
		tcompileSet = new TCompileSet(log, loadedJavaClass);
		if (tcompileSet.isEmpty())
			return;
		if (log.hasError())
			return;
		for (CompilationInfo compilation : tcompileSet.getCompilationInfos()) {
			progressInfo.startsTCompile(compilation.annot);
			String key = compilation.scopeClass.getSimpleName() + "##" + compilation.annotDescription;
			log.setLabel("  ------------ " + AnnotationHelper.toString(compilation.annot)
					+ (compilation.scopeClass.getEnclosingClass() != null ? "  " + compilation.scopeClass.getSimpleName() : ""));
			if (oldSourceInfo != null && !options.forceRecompilation) {
				PersistentCompilationInfo pci = oldSourceInfo.compilationMap.get(key);
				if (pci != null && pci.getModificatedSourceFile() == null) {
					newSourceInfo.compilationMap.put(key, pci);
					progressInfo.numNonCompiledTemplates++;
					continue;
				}
			}
			progressInfo.numCompiledTemplates++;
			PersistentCompilationInfo pci = new PersistentCompilationInfo(key);
			AnnotationCompiler jc = makeAnnotationCompiler(pci, compilation);
			jc.compile(progressInfo);
			newSourceInfo.compilationMap.put(key, pci);
		}
	}

	protected AnnotationCompiler makeAnnotationCompiler(PersistentCompilationInfo pci, CompilationInfo compilationInfo)
	{
		return new AnnotationCompiler(options, pci, javaFI, loadedJavaClass, log, compilationInfo);
	}

	public void emitLog()
	{
		log.purgeToStderr(progressInfo, options.verbose);
	}
}
