/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.htmlLoader;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import nu.validator.htmlparser.common.DoctypeExpectation;
import nu.validator.htmlparser.common.XmlViolationPolicy;
import nu.validator.htmlparser.dom.HtmlDocumentBuilder;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jabhts.translator.compiler.utils.CompilerLog;
import org.jabhts.translator.compiler.utils.CompilerUtils;
import org.jabhts.translator.compiler.utils.FileUtils;
import org.jabhts.translator.compiler.utils.LogLevel;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.xml.sax.InputSource;

public class HtmlLoader
{
	public final CompilerLog log;
	private HtmlDocumentBuilder builder;

	public HtmlLoader(CompilerLog log)
	{
		this.log = log;
	}

	public HtmlDocument load(File htmlF) throws CompilerException
	{
		try {
			if (htmlF.getName().toLowerCase().endsWith(".xhtml"))
				return loadXhtml(htmlF);
			if (!htmlF.exists())
				throw new CompilerException("No such file: " + htmlF.getPath());
			if (builder == null) {
				builder = new HtmlDocumentBuilder(XmlViolationPolicy.ALTER_INFOSET);
				builder.setIgnoringComments(false);
				builder.setDoctypeExpectation(DoctypeExpectation.AUTO);
				builder.parse(new ByteArrayInputStream("<div></div>".getBytes())); // NPE bug workaround.
			}
			builder.setErrorHandler(new LoggingErrorHandler(log));
			long moddate = htmlF.lastModified();
			FileInputStream fis = new FileInputStream(htmlF);
			Document doc;
			try {
				int sz = (int)Math.max(128, Math.min(16394, htmlF.length()));
				doc = builder.parse(new InputSource(new BufferedInputStream(fis, sz)));
			}
			finally {
				fis.close();
			}
			String docType = "<!DOCTYPE HTML>";
			return new HtmlDocument(htmlF, moddate, doc, docType);
		}
		catch (CompilerException ex) {
			throw ex;
		}
		catch (Exception ex) {
			log.log(LogLevel.Error, "Can't load html template: " + CompilerUtils.getSomeMessage(ex) + "\n" +
					"   " + htmlF.getPath());
			System.err.println("Load html exception: " + ex.getMessage());
			ex.printStackTrace();
			return null;
		}
	}

	public HtmlDocument loadXhtml(File htmlF) throws Exception
	{
		if (!htmlF.exists())
			throw new CompilerException("No such file: " + htmlF.getPath());
		long moddate = htmlF.lastModified();
		byte[] b = FileUtils.readFile(htmlF);
		XHtmlDomParser parser = XHtmlDomParser.makeSimpleParser();
		Document doc = parser.parse(b);
		DocumentType docType = doc.getDoctype();
		return new HtmlDocument(htmlF, moddate, doc, docType != null ? docType.getName() : null);
	}
}
