/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.utils;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import org.jabhts.annotation.global.TCompile;

public class AnnotationHelper
{
	public final CompilerLog log;
	public final Annotation annotation;
	public final Class<?> cls;
	public final Field field;
	public final Method method;

	public AnnotationHelper(CompilerLog log, Annotation annotation, Class<?> cls, Field field, Method method)
	{
		this.log = log;
		this.annotation = annotation;
		this.cls = cls;
		this.field = field;
		this.method = method;
	}

	public boolean getAnnotationBooleanParam(String name) throws InternalCompilerException
	{
		try {
			Class<?> atype = annotation.annotationType();
			Method m = atype.getMethod(name);
			return (Boolean)m.invoke(annotation);
		}
		catch (Exception ex) {
			throw new InternalCompilerException("Can't get boolean annotation value (" + name + "): "
					+ CompilerUtils.getSomeMessage(ex), ex);
		}
	}

	public String[] getAnnotationStringArrayParam(String name) throws InternalCompilerException
	{
		try {
			Object o = annotation.annotationType().getMethod(name).invoke(annotation);
			if (o == null)
				return CompilerUtils.emptyStringArray;
			if (!(o instanceof String[])) {
				log.log(LogLevel.Error, "Annotation parameter '" + name + "' is not String[]: "
						+ CompilerUtils.annotationToString(annotation, cls, field, method));
				return CompilerUtils.emptyStringArray;
			}
			String[] array = (String[])o;
			for (String s : array) {
				if (s == null || s.isEmpty()) {
					log.log(LogLevel.Error, "Annotation parameter '" + name + "' may not contain nulls or empty strings: "
							+ CompilerUtils.annotationToString(annotation, cls, field, method));
				}
			}
			return array;
		}
		catch (Exception ex) {
			throw new InternalCompilerException("Can't get String[] annotation value (" + name + "): "
					+ CompilerUtils.getSomeMessage(ex), ex);
		}
	}

	public String annotationToString()
	{
		StringBuilder sb = new StringBuilder(20);
		sb.append(toString(annotation)).append(' ');
		if (field != null || method != null) {
			if (field != null)
				sb.append(field.getName());
			if (method != null) {
				sb.append(method.getReturnType().getSimpleName()).append(' ');
				sb.append(method.getName()).append("()");
			}
			//sb.append(" in ").append(CompilerUtils.getInnerclassName(cls));
			sb.append(" in ");
			CompilerUtils.addClass(sb, cls);
		} else if (cls != null) {
			sb.append("class ");
			CompilerUtils.addClass(sb, cls);
		}
		return sb.toString();
	}

	public static String toString(TCompile a)
	{
		StringBuilder sb = new StringBuilder();
		sb.append('@').append(a.annotationType().getSimpleName());
		String sep = "(";
		if (!a.filename().isEmpty()) {
			sb.append(sep).append("filename='").append(a.filename()).append("'");
			sep = ", ";
		}
		if (!a.fragment().isEmpty()) {
			sb.append(sep).append("fragmentTree='").append(a.fragment()).append("'");
			sep = ", ";
		}
		if (a.fragmentChilds()) {
			sb.append(sep).append("fragmentChilds");
			sep = ", ";
		}
		if (!a.tpname().isEmpty()) {
			sb.append(sep).append("tpname='").append(a.tpname()).append("'");
			sep = ", ";
		}
		if (!sep.equals("("))
			sb.append(")");
		return sb.toString();
	}

	public static String toString(Annotation a)
	{
		return '@' + a.annotationType().getSimpleName();
	}
}
