/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.io;

import java.io.IOException;
import java.io.InputStream;

public class Base64InputStream extends InputStream
{
	private final InputStream is;

	private static final String encoded =
			"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

	private int encodedIndex;
	private int bufferedBits;

	private boolean eof;

	public Base64InputStream(InputStream is)
	{
		this.is = is;
	}

	public int available() throws IOException
	{
		return is.available() * 3 / 4;
	}

	public int read() throws IOException
	{
		if (eof)
			return -1;
		while (true) {
			int ch = readNonWhitespace();
			if (ch < 0) {
				eof = true;
				if (encodedIndex == 0)
					return -1;
				throw new IOException("Invalid BASE64 data: Got EOF inside encoded chunk.");
			}
			if (ch == '=') {
				switch (encodedIndex) {
					case 0:
					case 1:
						throw new IOException("Invalid BASE64 data: Got '=' at encoded char pos 0/1.");
					case 2:
						ch = readNonWhitespace();
						if (ch != '=')
							throw new IOException("Invalid BASE64 data: Invalid char after trailing '='.");
						ch = readNonWhitespace(); // make sure the src stream sees EOF.
						break;
					case 3:
						ch = readNonWhitespace();
				}
				eof = true;
				if (ch >= 0)
					throw new IOException("Invalid BASE64 data: got garbage after trailing '='");
				return -1;
			}
			int pos = encoded.indexOf((char)ch);
			if (pos < 0)
				throw new IOException("Invalid BASE64 data: invalid char = chr(" + ch + ")");
			switch (encodedIndex) {
				case 0:
					bufferedBits = pos << 2;
					encodedIndex = 1;
					continue; // need 1 more byte.
				case 1:
					ch = bufferedBits | (pos >>> 4);
					bufferedBits = (pos & 0x0F) << 4;
					encodedIndex = 2;
					return ch;
				case 2:
					ch = bufferedBits | (pos >>> 2);
					bufferedBits = (pos & 0x03) << 6;
					encodedIndex = 3;
					return ch;
				case 3:
					encodedIndex = 0;
					return bufferedBits | pos;
				default:
					throw new IOException("Invalid internal state.");
			}
		}
	}

	private int readNonWhitespace() throws IOException
	{
		int ch;
		do {
			ch = is.read();
		} while (Character.isWhitespace(ch));
		return ch;
	}
}
