/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.io;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.util.Arrays;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.misc.ArrayUtils;
import dk.vitality.lib.foundation.misc.ByteBuffer;

/**
 * Simple layer on top of {@link Process} which provides input/output
 * as strings (or byte arrays) instead of streams, and which makes
 * logging of stdout/stderr to System.err easy.
 * <p/>
 * Intended for simple executing of unix command line programs.
 * The class is not thread safe and is intended for use by a single thread only.
 * <p/>
 * Default is to echo stdout/err output to System.err, but not to collect
 * the output in a String due to memory considerations. If you want the
 * output as a String, call {@link #setWantsStdoutAsString},
 * {@link #setWantsStdoutAsBytes} and/or {@link #setWantsStderrAsString}
 */
public class SimpleUnixExec
{
	public static String backtick(String... cmd) throws Exception
	{
		SimpleUnixExec sue = new SimpleUnixExec(cmd);
		sue.setLogStdoutToSystemErr(false);
		sue.setWantsStdoutAsString(true);
		sue.setLogStdoutToSystemErr(false);
		sue.setLogStderrToSystemErr(true);
		sue.run();
		int r = sue.getResultCode();
		if (r != 0)
			throw new Exception("cmd exited with resultcode " + r + ": " + Arrays.toString(cmd));
		return sue.getStdoutAsString().trim();
	}

	/**
	 * @param cmd First item must be path to program.
	 */
	public SimpleUnixExec(String... cmd)
	{
		this.cmd = cmd;
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < cmd.length; i++) {
			if (i > 0)
				sb.append(' ');
			appendShellQuotedParam(sb, cmd[i]);
		}
		commandDesc = sb.toString();
	}

	/**
	 * This method replaces the FULL environment.
	 *
	 * @param env Must include SH, PATH, etc.
	 */
	public void setEnvironment(String[] env) throws IOException
	{
		checkNotStarted();
		this.env = env;
	}

	/**
	 * Adds to the environment specified by {@link #setEnvironment}.
	 * If no environment has been specified beforehand, an empty environment is used.
	 */
	public void addToEnvironment(String[] additionalVars) throws IOException
	{
		checkNotStarted();
		if (env == null) {
			setEnvironment(additionalVars);
			return;
		}
		env = ArrayUtils.concat(env, additionalVars);
	}

	public String[] getEnvironmentCopy()
	{
		String[] ss = new String[env.length];
		System.arraycopy(env, 0, ss, 0, env.length);
		return ss;
	}

	public void setWorkingDir(File dirF) throws IOException
	{
		checkNotStarted();
		this.workingDirF = dirF;
	}

	public void setLogStderrToSystemErr(boolean flag) throws IOException
	{
		checkNotStarted();
		this.logStderrToSystemErr = flag;
	}

	public void setLogStdoutToSystemErr(boolean flag) throws IOException
	{
		checkNotStarted();
		this.logStdoutToSystemErr = flag;
	}

	public void setWantsStdoutAsString(boolean flag) throws IOException
	{
		checkNotStarted();
		this.wantsStdoutAsString = flag;
	}

	public void setStdoutStream(OutputStream os) throws IOException
	{
		checkNotStarted();
		this.stdoutStream = os;
	}

	public void setWantsStdoutAsBytes(boolean flag) throws IOException
	{
		checkNotStarted();
		this.wantsStdoutAsBytes = flag;
		this.logStdoutToSystemErr = false;
	}

	public void setWantsStderrAsString(boolean flag) throws IOException
	{
		checkNotStarted();
		this.wantsStderr = flag;
		this.logStdoutToSystemErr = false;
	}

	public void setDebugFlag(boolean flag) throws IOException
	{
		checkNotStarted();
		this.debugFlag = flag;
	}

	public String getStdoutAsString() throws IOException
	{
		checkDone();
		return stdoutSB != null ? stdoutSB.toString() : null;
	}

	public byte[] getStdoutAsBytes() throws IOException
	{
		checkDone();
		return stdoutBB != null ? stdoutBB.toBytes() : null;
	}

	public String getStderr() throws IOException
	{
		checkDone();
		return stderrSB != null ? stderrSB.toString() : null;
	}

	public void setStdin(byte[] b) throws IOException
	{
		checkNotStarted();
		stdinData = b;
	}

	/**
	 * Run the sub process.
	 * <p/>
	 * This method does not throw an exception if the process returns a non-zero result code.
	 */
	public void run() throws Exception
	{
		checkNotStarted();
		doRun();
	}

	public int getResultCode() throws IOException
	{
		checkDone();
		return resultCode;
	}

	public interface StderrFilter
	{
		boolean include(String line);
	}

	public void setStderrFilter(StderrFilter filter)
	{
		this.stderrFilter = filter;
	}

	/**
	 * Charset used for converting stdout/stderr into String. Default is UTF-8.
	 */
	public void setCharset(String charset)
	{
		this.charset = charset;
	}
	//========================= private parts


	private static int nextProgramId = 100;
	private final int programId = nextProgramId++;
	private boolean logStderrToSystemErr = true, logStdoutToSystemErr = true, debugFlag;
	private final boolean debugVerbose = false;
	private boolean hasPrintedProgramId = false;
	private final String[] cmd;
	private String[] env;
	private final String commandDesc;
	private File workingDirF;
	private Process theProcess;
	private boolean hasDestroyed = false;
	private StringBuilder stdoutSB, stderrSB;
	private ByteBuffer stdoutBB;
	private boolean wantsStdoutAsString, wantsStdoutAsBytes, wantsStderr;
	private OutputStream stdoutStream;
	private InputStream is = null, es = null;
	private OutputStream os = null;
	private int resultCode = -1;
	private byte[] stdinData;
	private long startTime;
	private boolean done;
	private StderrFilter stderrFilter;
	private String charset = "UTF-8";

	private void doRun() throws Exception
	{
		startTime = System.currentTimeMillis();
		if (debugFlag) {
			logLine("Running..");
			if (env != null) {
				for (int i = 0; i < env.length; i++)
					System.err.println("env[" + i + "]: " + env[i]);
			}
		}
		if (wantsStderr)
			stderrSB = new StringBuilder();
		long st = System.currentTimeMillis();
		theProcess = Runtime.getRuntime().exec(cmd, env, workingDirF);
		BufferedReader reader = null;
		Thread stdoutThread = null;
		try {
			os = theProcess.getOutputStream();
			is = new BufferedInputStream(theProcess.getInputStream());
			es = new BufferedInputStream(theProcess.getErrorStream(), 128);
			reader = new BufferedReader(new InputStreamReader(es, charset));
			stdoutThread = new Thread(new StdoutReader());
			stdoutThread.start();
			Thread stdinThread = null;
			if (stdinData != null) {
				stdinThread = new Thread(new StdinWriter());
				stdinThread.start();
			} else {
				os.close();
			}
			while (true) {
				String s = reader.readLine();
				if (s == null)
					break;
				if (stderrFilter != null && !stderrFilter.include(s))
					continue;
				if (stderrSB != null) {
					stderrSB.append(s);
					stderrSB.append("\n");
				}
				if (logStderrToSystemErr)
					logLine(s);
			}
			if (debugFlag)
				logLine("Done reading stderr");
			es.close();
			if (stdinThread != null)
				stdinThread.join();
			stdoutThread.join();
			if (debugFlag)
				logLine("Done waiting for stdout");
			reader.close();
			is.close();
			if (debugFlag)
				logLine("calls waitFor()");
			resultCode = theProcess.waitFor();
			if (debugFlag)
				logLine("waitFor() done, returned " + resultCode);
		}
		finally {
			try {
				if (reader != null)
					reader.close();
				if (os != null)
					os.close();
				if (is != null)
					is.close();
				if (es != null)
					es.close();
				destroy();
			}
			catch (IOException dummy) {
				System.err.println("Got exception while closing process: " + dummy.getMessage());
				// ignore it
			}
		}
		destroy();
		if (stdoutThread != null)
			stdoutThread.join();
		long dur = System.currentTimeMillis() - st;
		if (debugFlag)
			logLine("Done running in " + dur + " msecs, resultcode = " + resultCode);
		done = true;
	}

	private void checkDone() throws IOException
	{
		if (startTime == 0)
			throw new IOException("Sub process has not been launched yet.");
		if (!done)
			throw new IOException("Sub-process is not done running.");
	}

	private void checkNotStarted() throws IOException
	{
		if (startTime > 0)
			throw new IOException("Sub process has already been launched.");
	}

	protected void logLine(String s)
	{
		if (!hasPrintedProgramId) {
			hasPrintedProgramId = true;
			logLine(commandDesc);
		}
		long t = System.currentTimeMillis() - startTime;
		System.err.println("##[" + programId + "." + t + "] " + s);
	}

	private void destroy()
	{
		if (theProcess != null && !hasDestroyed) {
			hasDestroyed = true;
			if (debugFlag)
				logLine("calls destroy()");
			theProcess.destroy();
			if (debugFlag)
				logLine("destroy done.");
		}
	}

	private class StdoutReader implements Runnable
	{
		private StringBuilder sb;

		public void run()
		{
			try {
				readOutput();
			}
			catch (Exception ex) {
				System.err.println("Got exception from reading stdout: " + ex);
			}
		}

		void readOutput() throws Exception
		{
			if (debugVerbose)
				logLine("StdoutReader starting");
			int i;
			if (stdoutStream != null) {
				byte[] buffer = new byte[65536];
				int num = 0;
				int numReads = 0;
				while ((i = is.read(buffer)) >= 0) {
					if (i == 0)
						continue;
					stdoutStream.write(buffer, 0, i);
					numReads++;
					num += i;
				}
				stdoutStream.flush();
				if (debugFlag) {
					logLine("Stdout reader done, " + num + " bytes in " + numReads + " reads"
							+ (numReads > 0 ? " = " + (num / numReads) + " bytes/read" : ""));
				}
			} else if (wantsStdoutAsBytes) {
				stdoutBB = new ByteBuffer(256);
				while ((i = is.read()) != -1) {
					stdoutBB.append((byte)i);
					logByte(i);
				}
			} else if (wantsStdoutAsString) {
				BufferedReader reader = new BufferedReader(new InputStreamReader(is, charset));
				stdoutSB = new StringBuilder(100);
				while ((i = reader.read()) != -1) {
					stdoutSB.append((char)i);
					logByte(i);
				}
			} else if (logStdoutToSystemErr) {
				BufferedReader reader = new BufferedReader(new InputStreamReader(is, charset));
				String s;
				while ((s = reader.readLine()) != null)
					logLine(s);
			}
			if (logStdoutToSystemErr && sb != null && sb.length() > 0)
				logByte('\n');
			if (debugVerbose)
				logLine("StdoutReader all done");
		}

		private void logByte(int ch)
		{
			if (!logStdoutToSystemErr)
				return;
			if (sb == null)
				sb = new StringBuilder();
			if (ch == '\n') {
				logLine(sb.toString());
				sb.setLength(0);
			} else {
				sb.append((char)ch);
			}
		}
	}

	private class StdinWriter implements Runnable
	{
		public void run()
		{
			try {
				if (debugVerbose)
					logLine("StdinWriter starting");
				writeInput();
			}
			catch (Exception ex) {
				System.err.println("Got exception from writing stdin: " + ex);
			}
			if (debugVerbose)
				logLine("StdinWriter all done");
		}

		void writeInput() throws Exception
		{
			if (stdinData != null)
				os.write(stdinData);
			os.close();
		}
	}

	public static String quoteArguments(@Nullable String[] cmd)
	{
		if (cmd == null)
			return "";
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < cmd.length; i++) {
			if (i > 0)
				sb.append(' ');
			SimpleUnixExec.appendShellQuotedParam(sb, cmd[i]);
		}
		return sb.toString();
	}

	public static void appendShellQuotedParam(@NotNull StringBuilder sb, @Nullable String s)
	{
		if (s == null || s.isEmpty()) {
			sb.append("''");
			return;
		}
		if (shellSafe(s)) {
			sb.append(s);
			return;
		}
		int len = s.length();
		sb.append("'");
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch == '\'' || ch == '\\')
				sb.append('\\');
			else if (ch < ' ') {
				sb.append('\\');
				ch += '@';
			}
			sb.append(ch);
		}
		sb.append("'");
	}

	private static boolean shellSafe(String s)
	{
		int len = s.length();
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if ((ch >= 'a' && ch <= 'z')
					|| (ch >= 'A' && ch <= 'Z')
					|| (ch >= '0' && ch <= '9')
					|| ch == '-' || ch == '_' || ch == '.'
					|| ch == '/' || ch == ':' || ch == ','
					|| ch == '=' || ch == '+') {
				continue;
			}
			if (ch == '#' && i > 0) // safe as previous can't be whitespace.
				continue;
			return false;
		}
		return true;
	}
}
