/* Copyright (c) 2006-2012, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.misc;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public class CollectionUtils
{
	/**
	 * Doesn't keep reference to list if it's empty, thus reduzing memory
	 * usage if list is an empty ArrayList.
	 */
	@NotNull
	public static <T> List<T> unmodifiableOrEmptyList(@Nullable List<? extends T> list)
	{
		if (list == null || list.isEmpty())
			return Collections.emptyList();
		else
			return Collections.unmodifiableList(list);
	}

	@NotNull
	public static <T> Set<T> unmodifiableOrEmptySet(@Nullable Set<? extends T> set)
	{
		if (set == null || set.isEmpty())
			return Collections.emptySet();
		else
			return Collections.unmodifiableSet(set);
	}

	@NotNull
	public static <K, V> Map<K,V> unmodifiableOrEmptyMap(@Nullable Map<? extends K, ? extends V> map)
	{
		if (map == null || map.isEmpty())
			return Collections.emptyMap();
		else
			return Collections.unmodifiableMap(map);
	}

	@NotNull
	public static <T> List<T> concat(@Nullable List<T> a, @Nullable List<T> b)
	{
		if (a == null || a.isEmpty())
			return unmodifiableOrEmptyList(b);
		if (b == null || b.isEmpty())
			return Collections.unmodifiableList(a);
		List<T> list = new ArrayList<>(a.size() + b.size());
		list.addAll(a);
		list.addAll(b);
		return unmodifiableOrEmptyList(list);
	}

	@NotNull
	public static <K, V> Map<K,V> concat(@Nullable Map<? extends K, ? extends V> a,
										 @Nullable Map<? extends K, ? extends V> b)
	{
		if (a == null || a.isEmpty())
			return unmodifiableOrEmptyMap(b);
		if (b == null || b.isEmpty())
			return Collections.unmodifiableMap(a);
		Map<K, V> map = new HashMap<>(a.size() + b.size());
		map.putAll(a);
		map.putAll(b);
		return unmodifiableOrEmptyMap(map);
	}

	public static final List<Boolean> falseTrue;

	static {
		List<Boolean> list = new ArrayList<>(2);
		list.add(false);
		list.add(true);
		falseTrue = Collections.unmodifiableList(list);
	}
}
