/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.misc;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.function.Function;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.ref.FatPair;

public class StringUtils
{
	public static boolean triEquals(@Nullable String a, @Nullable String b)
	{
		return a == null ? b == null : a.equals(b);
	}

	public static boolean triEqualsIgnoreCase(@Nullable String a, @Nullable String b)
	{
		return a == null ? b == null : a.equalsIgnoreCase(b);
	}

	public static int triCompare(@Nullable String a, @Nullable String b)
	{
		if (a == null)
			return b == null ? 0 : -1;
		if (b == null)
			return 1;
		return a.compareTo(b);
	}

	@Nullable
	public static String replaceSubstring(@Nullable String txt, @Nullable String newText, int offset, int len)
	{
		if (txt == null)
			return null;
		if (newText == null)
			newText = "";
		StringBuilder sb = new StringBuilder(txt.length() + newText.length() - len);
		if (offset > 0)
			sb.append(txt.substring(0, offset));
		sb.append(newText);
		int end = offset + len;
		if (end < txt.length())
			sb.append(txt.substring(end));
		return sb.toString();
	}

	@Nullable
	public static String replaceAll(@Nullable String txt, @NotNull String oldText, @Nullable String newText)
	{
		if (oldText == null)
			throw new IllegalArgumentException("oldText argument is null!");
		if (txt == null)
			return null;
		if (oldText.equals(newText)) // special-case.
			return txt;
		if (oldText.isEmpty())
			throw new IllegalArgumentException("oldText is an empty string");
		int startPos = txt.indexOf(oldText);
		if (startPos < 0)
			return txt;
		StringBuilder sb = new StringBuilder(txt.length());
		int prev = 0;
		while (true) {
			sb.append(txt.substring(prev, startPos));
			if (startPos >= txt.length())
				break;
			if (newText != null)
				sb.append(newText);
			prev = startPos + oldText.length();
			int idx = txt.indexOf(oldText, prev);
			startPos = (idx < 0) ? txt.length() : idx;
		}
		return sb.toString();
	}

	@Nullable
	public static String replaceFirst(@Nullable String txt, @NotNull String oldText, @Nullable String newText)
	{
		if (oldText == null)
			throw new IllegalArgumentException("oldText argument is null!");
		if (txt == null)
			return null;
		if (oldText.equals(newText)) // special-case.
			return txt;
		if (oldText.isEmpty())
			throw new IllegalArgumentException("oldText is an empty string");
		int idx = txt.indexOf(oldText);
		return (idx < 0) ? txt : replaceSubstring(txt, newText, idx, oldText.length());
	}

	@NotNull
	public static String replaceMulti(@NotNull String txt, String... replacements)
	{
		if (replacements == null)
			return txt;
		for (int i = 0; i < replacements.length; i += 2) {
			txt = StringUtils.replaceAll(txt, replacements[i], replacements[i + 1]);
		}
		return txt;
	}

	@Nullable
	public static String zapPrefix(@Nullable String s, @Nullable String prefix)
	{
		if (s == null)
			return null;
		if (prefix != null && !prefix.isEmpty() && s.startsWith(prefix))
			s = s.substring(prefix.length());
		return s;
	}

	@Nullable
	public static String zapRepeatedPrefix(@Nullable String s, @NotNull String prefix)
	{
		if (s == null)
			return null;
		while (s.startsWith(prefix))
			s = s.substring(prefix.length());
		return s;
	}

	@Nullable
	public static String zapSuffix(@Nullable String s, @NotNull String suffix)
	{
		if (s == null)
			return null;
		if (s.endsWith(suffix))
			s = s.substring(0, s.length() - suffix.length());
		return s;
	}

	@Nullable
	public static String zapRepeatedSuffix(@Nullable String s, @NotNull String suffix)
	{
		if (s == null)
			return null;
		while (s.endsWith(suffix))
			s = s.substring(0, s.length() - suffix.length());
		return s;
	}

	/**
	 * Changes runs of whitespace into a single space. <br>
	 * aka "s/\s+/ /g" in perl<br>
	 * So cr and lf becomes a space.
	 * And removes initial/trailing whitespace.
	 * Translates all ascii control chars into spaces.
	 */
	@Nullable
	public static String compactWhitespaces(@Nullable String s)
	{
		if (s == null)
			return null;
		int len = s.length();
		StringBuilder sb = new StringBuilder(len);
		boolean hadWhitespace = false;
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch <= ' ' || ch == 0xA0) { // A0 == nbsp
				hadWhitespace = true;
				continue;
			}
			if (hadWhitespace && sb.length() > 0)
				sb.append(' ');
			hadWhitespace = false;
			sb.append(ch);
		}
		return sb.toString();
	}

	@Nullable
	public static String zapAllWhitespace(@Nullable String s)
	{
		if (s == null)
			return null;
		int len = s.length();
		StringBuilder sb = new StringBuilder(len);
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch <= ' ' || ch == 0xA0)  // A0 == nbsp
				continue;
			sb.append(ch);
		}
		return sb.toString();
	}

	@Nullable
	public static String cleanWindowsLinefeeds(@Nullable String s)
	{
		if (s == null)
			return null;
		int len = s.length();
		StringBuilder sb = new StringBuilder(len);
		boolean lastWasCR = false;
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch == '\r') {
				sb.append('\n');
				lastWasCR = true;
				continue;
			}
			if (ch == '\n' && lastWasCR) {
				lastWasCR = false;
				continue;
			}
			sb.append(ch);
			lastWasCR = false;
		}
		return sb.toString();
	}

	@Nullable
	public static String concat(@Nullable String a, @Nullable String b)
	{
		if (a == null)
			return b;
		else if (b == null)
			return a;
		else
			return a + b;
	}

	@Nullable
	public static String concat(@Nullable String a, @Nullable String b, @Nullable String c)
	{
		return concat(concat(a, b), c);
	}

	@NotNull
	public static String join(@Nullable String sep, @Nullable String... values)
	{
		if (values == null || values.length == 0)
			return "";
		if (values.length == 1)
			return values[0];
		int len = 0;
		for (String s : values)
			len += (s != null) ? s.length() : 0;
		if (len == 0)
			return "";
		if (sep != null)
			len += (values.length - 1) * sep.length();
		StringBuilder sb = new StringBuilder(len);
		if (sep == null) {
			for (String s : values) {
				if (s != null)
					sb.append(s);
			}
		} else {
			boolean first = true;
			for (String s : values) {
				if (!first)
					sb.append(sep);
				first = false;
				if (s != null)
					sb.append(s);
			}
		}
		return sb.toString();
	}

	@NotNull
	public static String joinNonEmptyTrimmed(@Nullable String sep, @Nullable String... values)
	{
		if (values == null || values.length == 0)
			return "";
		if (values.length == 1)
			return values[0];
		StringBuilder sb = new StringBuilder();
		for (String s : values) {
			if (s == null || s.isEmpty())
				continue;
			s = s.trim();
			if (s.isEmpty())
				continue;
			if (sb.length() > 0)
				sb.append(sep);
			sb.append(s);
		}
		return sb.toString();
	}

	@NotNull
	public static String join(@Nullable String sep, @Nullable Collection<String> values)
	{
		if (values == null || values.isEmpty())
			return "";
		int len = 0;
		for (String s : values)
			len += (s != null) ? s.length() : 0;
		if (sep != null)
			len += (values.size() - 1) * sep.length();
		StringBuilder sb = new StringBuilder(len);
		if (sep == null) {
			for (String s : values) {
				if (s != null)
					sb.append(s);
			}
		} else {
			boolean first = true;
			for (String s : values) {
				if (!first)
					sb.append(sep);
				first = false;
				if (s != null)
					sb.append(s);
			}
		}
		return sb.toString();
	}

	@NotNull
	public static String join(@Nullable String sep, @NotNull Stream<String> values)
	{
		List<String> list = values.collect(Collectors.toList());
		return join(sep, list);
	}

	@Nullable
	public static String joinIntegers(@Nullable String sep, @Nullable Collection<Integer> list)
	{
		if (list == null)
			return null;
		if (list.isEmpty())
			return "";
		if (list.size() == 1 && list instanceof List)
			return ((List)list).get(0).toString();
		StringBuilder sb = new StringBuilder(list.size() * 6);
		for (Integer x : list) {
			if (sb.length() > 0 && sep != null)
				sb.append(sep);
			if (x != null)
				sb.append(x);
		}
		return sb.toString();
	}

	@Nullable
	public static String toStringJoined(@Nullable String sep, @Nullable Iterable<?> list)
	{
		if (list == null)
			return null;
		StringBuilder sb = new StringBuilder();
		for (Object obj : list) {
			String s = obj.toString();
			if (sb.length() > 0 && sep != null)
				sb.append(sep);
			if (s != null)
				sb.append(s);
		}
		return sb.toString();
	}

	@Nullable
	public static String join(@Nullable String sep, @Nullable int[] array)
	{
		if (array == null)
			return null;
		if (array.length == 1)
			return Integer.toString(array[0]);
		StringBuilder sb = new StringBuilder(array.length * 4);
		for (int v : array) {
			if (sb.length() > 0 && sep != null)
				sb.append(sep);
			sb.append(v);
		}
		return sb.toString();
	}

	@Nullable
	public static String makeFirstCharacterUpperCase(@Nullable String s)
	{
		if (s == null || s.isEmpty())
			return s;
		if (s.length() == 1)
			return s.toUpperCase();
		return Character.toUpperCase(s.charAt(0)) + s.substring(1);
	}

	@Nullable
	public static List<String> splitByCharToList(@Nullable String s, char ch)
	{
		if (s == null)
			return null;
		int len = s.length();
		if (len == 0)
			return Collections.emptyList();
		int num = 1;
		int idx = 0;
		while (idx < len) {
			idx = s.indexOf(ch, idx);
			if (idx < 0)
				break;
			num++;
			idx++;
		}
		List<String> list = new ArrayList<>(num);
		idx = 0;
		for (int part = 0; part < num; part++) {
			int idx2 = s.indexOf(ch, idx);
			list.add((idx2 < 0) ? s.substring(idx) : s.substring(idx, idx2));
			idx = idx2 + 1;
		}
		return list;
	}

	@NotNull
	public static String[] splitByChar(@Nullable String s, char ch)
	{
		if (s == null)
			return ArrayUtils.emptyStringArray;
		int len = s.length();
		if (len == 0)
			return ArrayUtils.emptyStringArray;
		int num = 1;
		int idx = 0;
		while (idx < len) {
			idx = s.indexOf(ch, idx);
			if (idx < 0)
				break;
			num++;
			idx++;
		}
		String[] array = new String[num];
		idx = 0;
		for (int part = 0; part < num; part++) {
			int idx2 = s.indexOf(ch, idx);
			array[part] = (idx2 < 0) ? s.substring(idx) : s.substring(idx, idx2);
			idx = idx2 + 1;
		}
		return array;
	}

	@NotNull
	public static List<String> splitByString(@NotNull String s, @NotNull String sep)
	{
		List<String> list = new ArrayList<>();
		int prev = 0;
		while (true) {
			int idx = s.indexOf(sep, prev);
			if (idx < 0) {
				list.add(s.substring(prev));
				return list;
			}
			list.add(s.substring(prev, idx));
			prev = idx + sep.length();
		}
	}

	@NotNull
	public static List<String> splitByRE(@NotNull String s, @NotNull Pattern re)
	{
		List<String> list = new ArrayList<>();
		int prev = 0;
		Matcher m = re.matcher(s);
		while (m.find(prev)) {
			list.add(s.substring(prev, m.start()));
			prev = m.end();
		}
		list.add(s.substring(prev));
		return list;
	}

	@Nullable
	public static String intern(@Nullable String s)
	{
		return (s != null) ? s.intern() : null;
	}

	@NotNull
	public static String makeFinderSortkey(@Nullable String s)
	{
		if (s == null)
			return "";
		int len = s.length();
		StringBuilder sb = new StringBuilder(len);
		for (int i = 0; i < len; ) {
			char ch = s.charAt(i);
			if (ch < '0' || ch > '9') {
				sb.append(ch);
				i++;
				continue;
			}
			int j = i + 1; // one beyond last digit
			while (j < len) {
				char ch2 = s.charAt(j);
				if (ch2 < '0' || ch2 > '9')
					break;
				j++;
			}
			for (int k = j - i; k < 10; k++)
				sb.append('0');
			while (i < j)
				sb.append(s.charAt(i++));
		}
		return sb.toString();
	}

	public static <T> void sortAsFinder(@NotNull List<T> list,
	                                    @NotNull Function<T, String> keyer)
	{
		List<FatPair<T, String>> tmp = new ArrayList<>(list.size());
		tmp.addAll(list.stream().map(t -> new FatPair<>(t, keyer.apply(t))).collect(Collectors.toList()));
		Collections.sort(tmp, (x, y) -> x.b.compareToIgnoreCase(y.b));
		list.clear();
		tmp.stream().map(p -> p.a).forEach(list::add);
	}

	public static <T> void sortAsFinder(@NotNull T[] list,
	                                    @NotNull Function<T, String> keyer)
	{
		List<FatPair<T, String>> tmp = new ArrayList<>(list.length);
		for (T t : list)
			tmp.add(new FatPair<>(t, keyer.apply(t)));
		Collections.sort(tmp, (x, y) -> x.b.compareToIgnoreCase(y.b));
		for (int i = 0; i < list.length; i++)
			list[i] = tmp.get(i).a;
	}


	@Nullable
	public static String zapFollowingLines(@Nullable String s)
	{
		if (s == null)
			return null;
		int idx = s.indexOf('\n');
		return (idx < 0) ? s : s.substring(0, idx).trim();
	}

	@NotNull
	public static String indent(@Nullable String s, @NotNull String indent)
	{
		if (s == null || s.isEmpty())
			return "";
		int numNewlines = 1;
		for (int i = 0; i < s.length(); i++) {
			if (s.charAt(i) == '\n')
				numNewlines++;
		}
		StringBuilder sb = new StringBuilder(s.length() + numNewlines * indent.length());
		boolean isAfterNewline = true;
		for (int i = 0; i < s.length(); i++) {
			char ch = s.charAt(i);
			if (ch == '\n') {
				isAfterNewline = true;
			} else if (isAfterNewline) {
				sb.append(indent);
				isAfterNewline = false;
			}
			sb.append(ch);
		}
		return sb.toString();
	}

	@SuppressWarnings("UnusedDeclaration")
	public static void unitTest()
	{
		testReplace("hello world!", " ", ".", "hello.world!");
		testReplace(null, " ", ".", null);
		testReplace("hello world, this is a text", " ", ".", "hello.world,.this.is.a.text");
		testReplace("hello world, this is a world of text", "world", "funny", "hello funny, this is a funny of text");
	}

	private static void testReplace(@Nullable String txt, @NotNull String oldText, @Nullable String newText,
	                                @Nullable String expected)
	{
		String actual = replaceAll(txt, oldText, newText);
		if (triEquals(actual, expected))
			return;
		System.err.println("replaceAll failed:");
		System.err.println("txt: '" + txt + "'");
		System.err.println("oldText: '" + oldText + "'");
		System.err.println("newText: '" + newText + "'");
		System.err.println("expected: '" + expected + "'");
		System.err.println("  actual: '" + actual + "'");
		System.exit(1);
	}

}
