/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.misc;

import java.util.Comparator;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;

public class VersionStringComparator implements Comparator<String>
{
	public static final VersionStringComparator instance = new VersionStringComparator();

	public int compare(String o1, String o2)
	{
		Decoder a = new Decoder(o1);
		Decoder b = new Decoder(o2);
		while (a.hasMore() || b.hasMore()) {
			int diff = a.next() - b.next();
			if (diff < 0)
				return -1;
			else if (diff > 0)
				return 1;
		}
		return 0;
	}

	private static class Decoder
	{
		@NotNull private final String s;
		private final int len;
		private int idx;

		Decoder(@Nullable String s)
		{
			if (s == null) {
				this.s = "";
				this.len = 0;
				return;
			}
			this.s = s;
			this.len = s.length();
			if (len > 0 && s.charAt(0) == '.')
				throw new RuntimeException("Version string starts with dot: '" + s + "'");
		}

		boolean hasMore()
		{
			return idx < len;
		}

		int next()
		{
			int stage = stage_final;
			if (idx >= len)
				return stage;
			int x = 0;
			int start = idx;
			while (idx < len) {
				char ch = s.charAt(idx++);
				if (ch == '.') {
					if (idx < len && s.charAt(idx) == '.')
						throw new RuntimeException("Got '..' in version string: '" + s + "'");
					break;
				}
				if (ch >= '0' && ch <= '9') {
					x = x * 10 + ch - '0';
					continue;
				}
				switch (ch) {
					case 'd':
					case 'D':
						stage = stage_developer;
						break;
					case 'a':
					case 'A':
						stage = stage_alpha;
						break;
					case 'b':
					case 'B':
						stage = stage_beta;
						break;
					default:
						throw new RuntimeException("Invalid char '" + ch + "' in version string: '" + s + "'");
				}
				if (idx <= start + 1)
					throw new RuntimeException("Expected digit before '" + ch + "' in version string: '" + s + "'");
				break;
			}
			return x * 8 + stage;
		}
	}

	private static final int stage_developer = 0;
	private static final int stage_alpha = 1;
	private static final int stage_beta = 2;
	private static final int stage_final = 3;

	public static void main(String[] args)
	{
		test("1.3", "1.3.0", 0);
		test("1.3", "1.3.0.0", 0);
		test("1.3", "1.3.1", -1);
		test("1.3.1", "1.3", 1);
		test("1.3.1", "1.3.0", 1);
		test("1.3.1", "1.3.0.0.0", 1);
		test("1.3b1", "1.3.0", -1);
		test("1.3b1", "1.3a2", 1);
		test("1.3a2", "1.3b1", -1);
		test("1.3a2", "1.3d3", 1);
		test("1.3b2", "1.3d3", 1);
		test("2.3d1", "1.3d3", 1);
		test("1.2.3.4b1", "1.2.3.4d2", 1);
		test("1.2.3.4b1", "1.2", 1);
	}

	private static void test(String a, String b, int expected)
	{
		try {
			int r = instance.compare(a, b);
			if (r == expected) {
				System.err.println("ok: '" + a + "' <=> '" + b + "' = " + r);
				return;
			}
			System.err.println("FAILED: '" + a + "' <=> '" + b + "' = " + r + ", expected " + expected);
		}
		catch (Exception ex) {
			System.err.println("FAILED: can't parse '" + a + "' or '" + b + "': " + ExceptionUtils.getExMsg(ex));
		}
	}
}
