/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.xml.create;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;
import dk.vitality.lib.foundation.misc.StringUtils;

public class XmlCreateUtils
{
	private static DocumentBuilder myDocumentBuilder = null;
	protected Document doc;

	public XmlCreateUtils(Document theDoc)
	{
		doc = theDoc;
	}

	public XmlCreateUtils(Element e)
	{
		doc = e.getOwnerDocument();
	}

	public static synchronized Document createDocument() throws ParserConfigurationException
	{
		if (myDocumentBuilder == null) {
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			myDocumentBuilder = dbf.newDocumentBuilder();
		}
		return myDocumentBuilder.newDocument();
	}

	public Document getDocument()
	{
		return doc;
	}

	public Element createElement(String tag) throws Exception
	{
		try {
			return doc.createElement(tag);
		}
		catch (DOMException ex) {
			throw new Exception("Can't create DOM element named '" + tag + "'", ex);
		}
	}

	public static void setOptAttr(Element elem, String name, String value)
	{
		if (name == null || name.isEmpty())
			throw new NullPointerException("name is null");
		if (value == null || value.isEmpty())
			elem.removeAttribute(name);
		else
			elem.setAttribute(name, value);
	}

	public static void setAttr(Element elem, String name, String value)
	{
		if (name == null || name.isEmpty())
			throw new NullPointerException("name is null");
		elem.setAttribute(name, value);
	}

	public static void setAttr(Element elem, String name, boolean value)
	{
		setAttr(elem, name, value ? "true" : "false");
	}

	public static void setAttr(Element elem, String name, int value)
	{
		setAttr(elem, name, Integer.toString(value));
	}

	public static void setAttr(Element elem, String name, long value)
	{
		setAttr(elem, name, Long.toString(value));
	}

	public static void setAttr(Element elem, String name, double value)
	{
		setAttr(elem, name, Double.toString(value));
	}

	public static void removeAttr(Element elem, String name)
	{
		elem.removeAttribute(name);
	}

	public Element addSoftChildWithTextNode(Element parentE, String tagName, String text)
	{
		if (text == null || text.isEmpty())
			return null;
		return addChildWithTextNode(parentE, tagName, text);
	}

	public Element addChildWithTextNode(Element parentE, String tagName, String value)
	{
		if (tagName == null || tagName.isEmpty())
			throw new RuntimeException("Empty tagname not valid");
		Element e = doc.createElement(tagName);
		if (value != null && !value.isEmpty())
			e.appendChild(doc.createTextNode(value));
		parentE.appendChild(e);
		return e;
	}

	public void addSoftTextNode(Element parentE, String value)
	{
		if (value == null || value.isEmpty())
			return;
		parentE.appendChild(doc.createTextNode(value));
	}

	public void addCDataNodes(Element elem, String value)
	{
		if (value == null || value.isEmpty())
			return;
		int startIndex = 0;
		int len = value.length();
		while (startIndex < len) {
			int idx = value.indexOf("]]>");
			if (idx < 0) {
				elem.appendChild(doc.createCDATASection(value.substring(startIndex)));
				return;
			}
			String s = value.substring(startIndex, idx);
			elem.appendChild(doc.createCDATASection(s));
			elem.appendChild(doc.createTextNode("]]>"));
			startIndex = idx + 3;
		}
	}

	public void addCommentNode(Element elem, String value)
	{
		if (value == null || value.isEmpty())
			return;
		elem.appendChild(doc.createComment(value));
	}

	/**
	 * moves childs of one node to another node
	 */
	public static void moveChildsToNodeFromNode(Node newParent, Node oldParent)
	{
		while (true) {
			Node child = oldParent.getFirstChild();
			if (child == null)
				break;
			newParent.appendChild(child);
		}
	}

	public static void removeAllChildren(Node parentNode)
	{
		while (true) {
			Node child = parentNode.getFirstChild();
			if (child == null)
				return;
			parentNode.removeChild(child);
		}
	}

	public static void removeChildrenByNodeName(Node parentNode, String nodeName)
	{
		Node child = parentNode.getFirstChild();
		while (child != null) {
			Node nextChild = child.getNextSibling();
			if (child.getNodeName().equals(nodeName))
				parentNode.removeChild(child);
			child = nextChild;
		}
	}

	public static void removeNode(Node n)
	{
		if (n == null)
			return;
		Node parentNode = n.getParentNode();
		if (parentNode != null)
			parentNode.removeChild(n);
	}

	public static void removeNodes(Node[] list)
	{
		if (list == null)
			return;
		for (Node n : list)
			removeNode(n);
	}

	//---------------------------------------------------------------------
	public static void zapWhitespaces(Node node) throws Exception
	{
		int type = node.getNodeType();
		switch (type) {
			case Node.DOCUMENT_NODE:
			case Node.ELEMENT_NODE:
				break;
			default:
				return;
		}
		Node child = node.getFirstChild();
		while (child != null) {
			Node nextChild = child.getNextSibling();
			if (isJustWhitespace(child))
				node.removeChild(child);
			else
				zapWhitespaces(child);
			child = nextChild;
		}
	}

	public static boolean childsAreJustWhitespace(Node n) throws Exception
	{
		if (n == null)
			return true;
		Node child = n.getFirstChild();
		while (child != null) {
			if (!isJustWhitespace(child))
				return false;
			child = child.getNextSibling();
		}
		return true;
	}

	public static boolean isJustWhitespace(Node n) throws Exception
	{
		int type = n.getNodeType();
		if (type != Node.TEXT_NODE)
			return false;
		String s = n.getNodeValue();
		if (s == null)
			return true;
		for (int i = s.length() - 1; i >= 0; i--) {
			if (s.charAt(i) > 32)
				return false;
		}
		return true;
	}

	public static void whitespaceTrimAllTextNodes(Node n) throws Exception
	{
		int type = n.getNodeType();
		switch (type) {
			case Node.DOCUMENT_NODE:
			case Node.ELEMENT_NODE:
				break;
			case Node.TEXT_NODE:
				whitespaceTrimOneTextNode(n);
				return;
			default:
				return;
		}
		Node child = n.getFirstChild();
		while (child != null) {
			Node nextChild = child.getNextSibling();
			whitespaceTrimAllTextNodes(child);
			child = nextChild;
		}
	}

	private static void whitespaceTrimOneTextNode(Node n) throws Exception
	{
		Text t = (Text) n;
		String s = t.getData();
		String s2 = s.trim();
		if (s2.isEmpty()) {
			XmlCreateUtils.removeNode(n);
			return;
		}
		if (s2.length() == s.length())
			return;
		t.setData(s2);
	}

	public static void combineAdjacentTextNodes(Node parentNode) throws Exception
	{
		Node child = parentNode.getFirstChild();
		while (child != null) {
			Node nextChild = child.getNextSibling();
			switch (child.getNodeType()) {
				case Node.ELEMENT_NODE:
					combineAdjacentTextNodes(child);
					break;
				case Node.TEXT_NODE:
					if (nextChild != null && nextChild.getNodeType() == Node.TEXT_NODE) {
						Text t1 = (Text) child;
						Text t2 = (Text) nextChild;
						t1.appendData(t2.getData());
						XmlCreateUtils.removeNode(nextChild);
						continue; // keep 'child' active if more text nodes.
					}
					break;
				default:
					break;
			}
			child = nextChild;
		}
	}

	public static void normalizeAllWhitespace(Node n) throws Exception
	{
		int type = n.getNodeType();
		switch (type) {
			case Node.DOCUMENT_NODE:
			case Node.ELEMENT_NODE:
				Node nextChild;
				for (Node child = n.getFirstChild(); child != null; child = nextChild) {
					nextChild = child.getNextSibling();
					normalizeAllWhitespace(child);
				}
				return;
			case Node.TEXT_NODE:
				break;
			default:
				return;
		}
		Text t = (Text)n;
		String s = t.getData();
		if (s.isEmpty())
			return;
		boolean a = Character.isWhitespace(s.charAt(0));
		boolean b = Character.isWhitespace(s.charAt(s.length() - 1));
		s = StringUtils.compactWhitespaces(s);
		if (b && !s.isEmpty())
			s += " ";
		if (a)
			s = " " + s;
		t.setData(s);
	}

	public static void zapComments(Node node) throws Exception
	{
		int type = node.getNodeType();
		switch (type) {
			case Node.DOCUMENT_NODE:
			case Node.ELEMENT_NODE:
				break;
			default:
				return;
		}
		Node child = node.getFirstChild();
		while (child != null) {
			Node nextChild = child.getNextSibling();
			if (child.getNodeType() == Node.COMMENT_NODE)
				node.removeChild(child);
			else
				zapComments(child);
			child = nextChild;
		}
	}
}
