/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.xml.parse;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.InputStream;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Document;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import dk.vitality.lib.foundation.exception.ExceptionUtils;

public class DomParser implements ErrorHandler
{
	protected DocumentBuilder parser;
	protected Exception fatalErrorException;

	private static DocumentBuilderFactory advancedFactory = null, simpleFactory = null;

	protected DomParser(DocumentBuilder parser)
	{
		this.parser = parser;
		parser.setErrorHandler(this);
	}

	@SuppressWarnings({"UnusedDeclaration"})
	public static synchronized DomParser getAdvanced() throws ParserConfigurationException
	{
		if (advancedFactory == null)
			advancedFactory = createdAdvancedFactory();
		return new DomParser(advancedFactory.newDocumentBuilder());
	}

	@SuppressWarnings({"UnusedDeclaration"})
	public static synchronized DomParser getNonValidating() throws ParserConfigurationException
	{
		if (simpleFactory == null)
			simpleFactory = createdSimpleFactory();
		return new DomParser(simpleFactory.newDocumentBuilder());
	}

	public static DocumentBuilderFactory createdAdvancedFactory()
	{
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setExpandEntityReferences(true);
		factory.setIgnoringComments(false);
		factory.setIgnoringElementContentWhitespace(false);
		factory.setNamespaceAware(true);
		factory.setValidating(true);
		factory.setXIncludeAware(true);
		setAttribute(factory, "http://apache.org/xml/features/dom/defer-node-expansion", false);
		setAttribute(factory, "http://apache.org/xml/features/validation/schema", true);
		setAttribute(factory, "http://apache.org/xml/features/validation/schema-full-checking", true);
		setAttribute(factory, "http://apache.org/xml/features/validate-annotations", false);
		setAttribute(factory, "http://apache.org/xml/features/warn-on-duplicate-entitydef", true);
		setAttribute(factory, "http://apache.org/xml/features/validation/warn-on-duplicate-attdef", true);
		return factory;
	}

	public static DocumentBuilderFactory createdSimpleFactory()
	{
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setExpandEntityReferences(true);
		factory.setIgnoringComments(false);
		factory.setIgnoringElementContentWhitespace(false);
		factory.setNamespaceAware(true);
		factory.setValidating(false);
		factory.setXIncludeAware(false);
		setAttribute(factory, "http://apache.org/xml/features/dom/defer-node-expansion", false);
		setAttribute(factory, "http://apache.org/xml/features/validation/schema", false);
		setAttribute(factory, "http://apache.org/xml/features/validation/schema-full-checking", false);
		setAttribute(factory, "http://apache.org/xml/features/validate-annotations", false);
		setAttribute(factory, "http://apache.org/xml/features/warn-on-duplicate-entitydef", true);
		setAttribute(factory, "http://apache.org/xml/features/validation/warn-on-duplicate-attdef", true);
		return factory;
	}

	private static void setAttribute(DocumentBuilderFactory factory, String attr, boolean value)
	{
		try {
			factory.setAttribute(attr, value);
		}
		catch (IllegalArgumentException ex) {
			String msg = "Ignores XML parser attr (" + ExceptionUtils.getExMsg(ex) + "): " + attr;
			if (!warnedAttrs.containsKey(msg))
				System.err.println(msg);
			else
				warnedAttrs.put(msg, "1");
		}
	}

	private static final Map<String, String> warnedAttrs = new ConcurrentHashMap<>();

	public Document parse(byte[] b) throws Exception
	{
		return parse(new ByteArrayInputStream(b));
	}

	public Document parse(InputStream is) throws Exception
	{
		return parse(new InputSource(is));
	}

	public Document parse(String s) throws Exception
	{
		return parse(s.getBytes("utf-8"));
	}

	public Document parse(InputSource is) throws Exception
	{
		Document doc = getParser().parse(is);
		if (fatalErrorException != null)
			throw fatalErrorException;
		return doc;
	}

	public Document parse(File f) throws Exception
	{
		Document doc = getParser().parse(f);
		if (fatalErrorException != null)
			throw fatalErrorException;
		return doc;
	}

	public DocumentBuilder getParser() throws ParserConfigurationException
	{
		return parser;
	}

	public void warning(SAXParseException ex)
	{
		System.err.println("[Warning] " + getLocationString(ex) + ": " + ex.getMessage());
	}

	public void error(SAXParseException ex)
	{
		System.err.println("[Error] " + getLocationString(ex) + ": " + ex.getMessage());
		if (fatalErrorException == null)
			fatalErrorException = ex;
	}

	public void fatalError(SAXParseException ex) throws SAXException
	{
		System.err.println("[Fatal Error] " + getLocationString(ex) + ": " + ex.getMessage());
		throw ex;
	}

	public static String getLocationString(SAXParseException ex)
	{
		StringBuilder sb = new StringBuilder();
		String systemId = ex.getSystemId();
		if (systemId != null) {
			int index = systemId.lastIndexOf('/');
			if (index != -1)
				systemId = systemId.substring(index + 1);
			sb.append(systemId);
			sb.append(':');
		}
		sb.append(ex.getLineNumber());
		sb.append(':');
		sb.append(ex.getColumnNumber());
		return sb.toString();
	}
}
