/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "VersionListJsonParser.h"
#import "GBDebug.h"
#include <sys/sysctl.h>
#include <SystemConfiguration/SystemConfiguration.h>

static BOOL verbose = NO;
static NSString* gbVersion = NULL;
static NSString* javaVersion = NULL;
static SInt32 osxMajor = 0, osxMinor = 0, osxBugfix = 0;

static void gbLog(NSString *format, ...)
{
	if (!verbose)
		return;
	va_list args;
	va_start(args, format);
	NSLogv(format, args);
	va_end(args);
}

static void gbDie(NSString *format, ...)
{
	if (verbose) {
		va_list args;
		va_start(args, format);
		NSLogv(format, args);
		va_end(args);
	}
	exit(1);
}

static NSString *getVersionFromBundleWithPath(NSString* path)
{
	path = [path stringByAppendingString:@"/Contents/Info.plist"];
	NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:path];
	if (!dict) {
		gbLog(@"Missing plist: %@", path);
		return @"0.0";
	}
	NSString *vers = [dict objectForKey:@"CFBundleVersion"]; // CFBundleShortVersionString has wrong content for Java
	if (!vers) {
		gbLog(@"CFBundleVersion is missing in %@", path);
		vers = [dict objectForKey:@"CFBundleShortVersionString"];
	}
	return vers ? vers : @"0.0";
}

static void addUrlArg(NSMutableString *sb, NSString* name, NSString* value)
{
	if ([sb length])
		[sb appendString:@"&"];
	[sb appendString:name];
	[sb appendString:@"="];
	NSString *s = [value stringByAddingPercentEncodingWithAllowedCharacters:[NSCharacterSet URLQueryAllowedCharacterSet]];
		[sb appendString:s];
}

static void addSysIntUrlArg(NSMutableString *sb, NSString *paramName, const char *sysname)
{
	int value;
	size_t length = sizeof(value);
	int error = sysctlbyname(sysname, &value, &length, NULL, 0);
	if (error) {
		gbLog(@"sysctlbyname(%s) returned %ld", sysname, error);
		return;
	}
	addUrlArg(sb, paramName, [@(value) stringValue]);
}

static void addSysStringUrlArg(NSMutableString *sb, NSString *paramName, const char *sysname)
{
	size_t length = 0;
	int error = sysctlbyname(sysname, NULL, &length, NULL, 0);
	if (error) {
		gbLog(@"sysctlbyname(%s, NULL) returned %ld", sysname, error);
		return;
	}
	char *s = (char *)malloc(sizeof(char) * length + 10);
	if (!s)
		gbDie(@"Can't malloc(%ld)", length);
	error = sysctlbyname(sysname, s, &length, NULL, 0);
	if (error) {
		gbLog(@"sysctlbyname(%s, buffer) returned %ld", sysname, error);
		free(s);
		return;
	}
	addUrlArg(sb, paramName, @(s));
	free(s);
}

static NSURL *addQuery(NSURL *baseUrl)
{
	NSURLComponents *urlb = [NSURLComponents componentsWithURL:baseUrl resolvingAgainstBaseURL:NO];
	NSMutableString *sb = [NSMutableString stringWithCapacity:100];
	//
	addUrlArg(sb, @"appVersion", gbVersion);
	addUrlArg(sb, @"javaVersion", javaVersion);
	addUrlArg(sb, @"osVersion", [NSString stringWithFormat:@"%d.%d.%d", osxMajor, osxMinor, osxBugfix]);
	//
	addSysIntUrlArg(sb, @"cputype", "hw.cputype");
	addSysIntUrlArg(sb, @"cpusubtype", "hw.cpusubtype");
	addSysIntUrlArg(sb, @"ncpu", "hw.ncpu");
	addSysStringUrlArg(sb, @"model", "hw.model");
	//
	unsigned long ulvalue;
	size_t length = sizeof(ulvalue);
	int error = sysctlbyname("hw.cpufrequency", &ulvalue, &length, NULL, 0);
	if (error)
		gbLog(@"sysctlbyname(hw.cpufrequency) returned %ld", error);
	else
		addUrlArg(sb, @"cpuFreqMHz", [@(ulvalue / 1000000) stringValue]);
	//
	unsigned long lvalue;
	length = sizeof(lvalue);
	error = sysctlbyname("hw.memsize", &lvalue, &length, NULL, 0);
	if (error)
		gbLog(@"sysctlbyname(hw.memsize) returned %ld", error);
	else
		addUrlArg(sb, @"ramMB", [@(lvalue / 1024.0 / 1024.0) stringValue]);
	//
	const char *names[] = { "hw.cpu64bit_capable", "hw.optional.x86_64", "hw.optional.64bitops" /* PowerPC */, NULL };
	int value = 0;
	for (const char **name = names; *name; name++) {
		length = sizeof(value);
		error = sysctlbyname(*name, &value, &length, NULL, 0);
		gbLog(@"sysctlbyname(%s) -> %d", *name, value);
		if (!error)
			break;
		value = 0;
	}
	addUrlArg(sb, @"cpu64bit", [@(value ? 0 : 1) stringValue]);
	//
	NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
	NSArray *languages = [defs objectForKey:@"AppleLanguages"];
	if ([languages count] > 0)
		addUrlArg(sb, @"lang", [languages objectAtIndex:0]);
	//
	urlb.query = sb;
	return [urlb URL];
}

BOOL isUserCurrentlyOnConsole()
{
	// https://developer.apple.com/library/mac/qa/qa1133/_index.html
	// https://developer.apple.com/library/mac/documentation/Networking/Reference/SCDynamicStoreCopySpecific/
	CFStringRef cfConsole = SCDynamicStoreCopyConsoleUser(NULL, NULL, NULL);
	if (cfConsole == NULL) {
		gbLog(@"SCDynamicStoreCopyConsoleUser() returned NULL");
		return NO;
	}
	NSString *console = (__bridge NSString *)cfConsole;
	NSString *myself = NSUserName();
	BOOL result = [myself isEqualToString:console];
	CFRelease(cfConsole);
	gbLog(@"console = %@, myself = %@  => %d", console, myself, result);
	return result;
}

int main(int argc, const char * argv[])
{
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
	Gestalt(gestaltSystemVersionMajor, &osxMajor);
	Gestalt(gestaltSystemVersionMinor, &osxMinor);
	Gestalt(gestaltSystemVersionBugFix, &osxBugfix);
	// NSOperatingSystemVersion osv = [processInfo operatingSystemVersion]; only in 10.10 and later.
#pragma GCC diagnostic pop
	@autoreleasepool {
		if ([NSRunningApplication runningApplicationsWithBundleIdentifier:INSTALLER_BUNDLE_ID].count) {
			gbLog(@"Upgrader app is already running, so abort checking.");
			return 0;
		}
		int idx = 0;
		// defaults read org.glimmerblocker.installer.suite
		// defaults delete org.glimmerblocker.installer.suite
		NSUserDefaults* defaults = [[NSUserDefaults alloc] initWithSuiteName:PREFS_SUITE_ID_INSTALLER];
		NSDate *lastSysInfoDate = [NSDate dateWithTimeIntervalSinceReferenceDate:[defaults doubleForKey:PREFSKEY_LASTDATE_MACINFO]];
		BOOL addMacInfo = (-[lastSysInfoDate timeIntervalSinceNow] >= 7 * 24 * 60 * 60);
		while (++idx < argc) {
			if (!strcmp(argv[idx], "-v")) {
				verbose = YES;
				continue;
			}
			if (!strcmp(argv[idx], "-m")) {
				addMacInfo = YES;
				continue;
			}
			gbLog(@"Unexpected arg: %s", argv[idx]);
		}
		if (!isUserCurrentlyOnConsole()) {
			gbLog(@"User not on console. Skips check for upgrade.");
			return 0;
		}
		NSString *gbBundlePath = @"/Library/PreferencePanes/GlimmerBlocker.prefPane";
		gbVersion = getVersionFromBundleWithPath(gbBundlePath);
		javaVersion = getVersionFromBundleWithPath([gbBundlePath stringByAppendingString:@"/Contents/GlimmerBlockerProxy.app/Contents/PlugIns/jre"]);
		//
		NSString *ua = [NSMutableString stringWithFormat:@"GlimmerBlockerUpdater/%@ OSX/%d.%d.%d",
						gbVersion,
						osxMajor, osxMinor, osxBugfix];
		//
		NSURL *url = [VersionListJsonParser versionListURL];
		if (addMacInfo)
			url = addQuery(url);
		gbLog(@"url: %@", url);
		gbLog(@"ua: %@", ua);
		//
		NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:url];
		[request setCachePolicy:NSURLRequestReloadIgnoringLocalCacheData];
		[request setHTTPShouldHandleCookies:NO];
		[request setValue:ua forHTTPHeaderField:@"user-agent"];
		NSData *previousResponseBody = [defaults dataForKey:PREFSKEY_RESPONSE_BODY];
		NSString *ifModifiedSince = [previousResponseBody length] ? [defaults stringForKey:PREFSKEY_IF_MODIFIED_SINCE] : NULL;
		gbLog(@"Got saved if-modified-since = %@", ifModifiedSince);
		if (ifModifiedSince)
			[request setValue:ifModifiedSince forHTTPHeaderField:@"If-Modified-Since"];
		//
		NSError *error = NULL;
		NSURLResponse *response = NULL;
		NSData *data = [NSURLConnection sendSynchronousRequest:request returningResponse:&response error:&error];
		if (error != NULL)
			gbDie(@"Got http error: %@", error);
		//
		if (addMacInfo)
			[defaults setDouble:[[NSDate date] timeIntervalSinceReferenceDate] forKey:PREFSKEY_LASTDATE_MACINFO ];
		NSHTTPURLResponse *httpResponse = (NSHTTPURLResponse*)response;
		if (httpResponse.statusCode == 304) {
			if (![previousResponseBody length]) {
				gbDie(@"Got 304 http status code while NOT having previous body.");
			}
			data = previousResponseBody; // need to check once more, OSX might be upgraded and allows newer GB version.
			gbLog(@"Not modified, reuses previous %ld bytes response body.", [data length]);
			[defaults synchronize];
		} else if (httpResponse.statusCode == 200) {
			NSString *lastModified = [httpResponse.allHeaderFields objectForKey:@"Last-Modified"];
			if (lastModified) {
				gbLog(@"Updates moddate: %@ -> %@", ifModifiedSince, lastModified);
				[defaults setObject:data forKey:PREFSKEY_RESPONSE_BODY];
				[defaults setObject:lastModified forKey:PREFSKEY_IF_MODIFIED_SINCE];
				[defaults synchronize];
			}
		} else {
			gbDie(@"Server returned http status %ld", httpResponse.statusCode);
		}
		VersionListJsonParser* parser = [[VersionListJsonParser alloc] initWithData:data
																   installedVersion:gbVersion
																			verbose:verbose
																			  error:&error];
		if (error != NULL)
			gbDie(@"Got error parsing JSON: %@", error);
		if (!parser)
			gbDie(@"Failed creating version parser");
		if (parser.versionList.count == 0) {
			gbLog(@"Got no version from server.");
			return 0;
		}
		GbVersionInfo *info = [parser.versionList objectAtIndex:0];
		if (!info.isUpgrade) {
			gbLog(@"Success, latest version is %@ which isn't an upgrade. Nothing to do.", info.version);
			return 0;
		}
		NSString *skippedVersion = [defaults stringForKey:PREFSKEY_SKIPPED_VERSION];
		if ([info.version isEqualToString:skippedVersion]) {
			gbLog(@"Success, got version %@ which user has selected to skip.", info.version, gbVersion);
			return 0;
		}
		if ([NSRunningApplication runningApplicationsWithBundleIdentifier:INSTALLER_BUNDLE_ID].count) {
			gbLog(@"Success, got version %@ while installed = %@, but upgrader app is already running.", info.version, gbVersion);
			return 0;
		}
		gbLog(@"Success, got version %@ while installed = %@, so launch upgrader.", info.version, gbVersion);
		[GBDebug executeCommandlineTool:@"open" withArgs:@[@"-a", [GBDebug updaterBundlePath]]];
	}
	return 0;
}
