/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "VersionListJsonParser.h"
#import "GBDebug.h"

@interface GbVersionInfo()
@end

@implementation GbVersionInfo

- (instancetype)init
{
	if (!(self = [super init]))
		return NULL;
	return self;
}

@end

//=========================================================================================================

NSString *PREFS_SUITE_ID_INSTALLER = @"org.glimmerblocker.installer.suite";
NSString *PREFSKEY_IF_MODIFIED_SINCE = @"if-modified-since";
NSString *PREFSKEY_RESPONSE_BODY = @"response-body";
NSString *PREFSKEY_SKIPPED_VERSION = @"skipped-version";
NSString *PREFSKEY_LASTDATE_MACINFO = @"last-sysinfo-date";

//=========================================================================================================

@interface VersionListJsonParser()

@property BOOL verbose;
@property NSString* installedVersion;
@property NSString* installedVersionKey;
@property SInt32 osxMajor, osxMinor, osxBugfix;
@property NSRegularExpression* osxVersionRegexp;

- (void)log:(NSString *)format, ...;
- (void)parseVersion:(id)obj;

@end


@implementation VersionListJsonParser


- (instancetype)initWithData:(NSData*)data
			installedVersion:(NSString*)installedVersion
					 verbose:(BOOL)verbose
					   error:(NSError **)error;
{
	if (!(self = [super init]))
		return NULL;
	_verbose = verbose;
	_installedVersion = installedVersion;
	_installedVersionKey = [VersionListJsonParser makeVersionSortKey:_installedVersion];
	_versionList = [NSMutableArray arrayWithCapacity:30];
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
	_osxMajor = 0;
	_osxMinor = 0;
	_osxBugfix = 0;
	Gestalt(gestaltSystemVersionMajor, &_osxMajor);
	Gestalt(gestaltSystemVersionMinor, &_osxMinor);
	Gestalt(gestaltSystemVersionBugFix, &_osxBugfix);
	// NSOperatingSystemVersion osv = [processInfo operatingSystemVersion]; only in 10.10 and later.
#pragma GCC diagnostic pop
	_osxVersionRegexp = [NSRegularExpression regularExpressionWithPattern:@"^(\\d+)\\.(\\d+)(?:\\.(\\d+))?"
																  options:0
																	error:error];
	if (*error || !_osxVersionRegexp) {
		[self log:@"error in osxVersionRegexp"];
		return NULL;
	}
	id json = [NSJSONSerialization JSONObjectWithData:data options:0 error:error];
	if (*error) {
		[self log:@"Can't parse json, error: %@", *error];
		return self;
	}
	//[self log:@"json: %@", json];
	if (![json isKindOfClass:[NSArray class]]) {
		[self log:@"json response top object is not array"];
		return self;
	}
	NSArray* list = (NSArray*)json;
	for (id obj in list) {
		[self parseVersion:obj];
	}
	return self;
}

- (void)log:(NSString *)format, ...
{
	if (!_verbose)
		return;
	va_list args;
	va_start(args, format);
	NSLogv(format, args);
	va_end(args);
}

- (BOOL)currentOsxIsAtLeastMajor:(int)reqMajor  minor:(int)reqMinor  bugfix:(int)reqBugfix
{
	if (_osxMajor > reqMajor)
		return YES;
	if (_osxMajor < reqMajor)
		return NO;
	//
	if (_osxMinor > reqMinor)
		return YES;
	if (_osxMinor < reqMinor)
		return NO;
	//
	return (_osxBugfix >= reqBugfix);
}

- (NSString*)getStringNamed:(NSString*)key defaultValue:(NSString*)defaultValue in:(NSDictionary*)dict
{
	NSString *s = [dict objectForKey:key];
	if (!s) {
		if (!defaultValue)
			[self log:@"Missing '%@' in dict: %@", key, dict];
		return defaultValue;
	}
	if (!s.length) {
		if (!defaultValue)
			[self log:@"Empty '%@' in dict: %@", key, dict];
		return defaultValue;
	}
	return s;
}

- (NSString*)getStringNamed:(NSString*)key in:(NSDictionary*)dict
{
	return [self getStringNamed:key defaultValue:NULL in:dict];
}

- (void)parseVersion:(id)obj
{
	if (![obj isKindOfClass:[NSDictionary class]]) {
		[self log:@"Bad type of item in version list: %@", obj];
		return;
	}
	NSDictionary *dict = (NSDictionary*)obj;
	GbVersionInfo *info = [[GbVersionInfo alloc] init];
	info.version = [self getStringNamed:@"v" in:dict];
	if (!info.version)
		return;
	info.sortkey = [VersionListJsonParser makeVersionSortKey:info.version];
	info.isUpgrade = [info.sortkey compare:_installedVersionKey] > 0;
	info.isInstalled = [info.version isEqualToString:_installedVersion];
	NSString *reqOSX = [self getStringNamed:@"osx" in:dict];
	if (!reqOSX)
		return;
	NSTextCheckingResult *r = [_osxVersionRegexp firstMatchInString:reqOSX options:0 range:NSMakeRange(0, reqOSX.length)];
	if (!r) {
		[self log:@"Can't match min-osx = '%@'", reqOSX];
		return;
	}
	int reqMajor = [reqOSX substringWithRange:[r rangeAtIndex:1]].intValue;
	int reqMinor = [reqOSX substringWithRange:[r rangeAtIndex:2]].intValue;
	NSRange r3 = [r rangeAtIndex:3];
	int reqBugfix = (r3.location != NSNotFound) ? [reqOSX substringWithRange:r3].intValue : 0;
	//NSLog(@"GB '%@' requires OSX %d.%d.%d\n", vers, reqMajor, reqMinor, reqBugfix);
	if (![self currentOsxIsAtLeastMajor:reqMajor minor:reqMinor  bugfix:reqBugfix]) {
		[self log:@"This OSX (%d.%d.%d) is not compatible with GB %@ which requires OSX %d.%d.%d or later",
			  _osxMajor, _osxMinor, _osxBugfix, info.version, reqMajor, reqMinor, reqBugfix];
		return;
	}
	NSString* baseUrl = [self getStringNamed:@"url" in:dict];
	if (!baseUrl)
		return;
	if (![baseUrl hasPrefix:@"http"])
		baseUrl = [@"https://glimmerblocker.org/upgrades" stringByAppendingString:baseUrl];
	info.baseUrl = [NSURL URLWithString:baseUrl];
	//
	NSString *applypatch = [self getStringNamed:@"applypatch" defaultValue:@"/applypatch.gz" in:dict];
	NSAssert(applypatch, @"applypatch");
	info.applyPatchUrl = [NSURL URLWithString:[baseUrl stringByAppendingString:applypatch]];
	//
	NSString *releaseNotesUrl = [self getStringNamed:@"notes" defaultValue:@"/release-notes.html" in:dict];
	info.releaseNotesUrl = [NSURL URLWithString:[baseUrl stringByAppendingString:releaseNotesUrl]];
	//
	NSString *ch = [self getStringNamed:@"ch" defaultValue:@"final" in:dict];
	NSAssert(ch, @"channel");
	if ([ch isEqualToString:@"dev"])
		info.channel = ChannelDev;
	else if ([ch isEqualToString:@"beta"])
		info.channel = ChannelBeta;
	else
		info.channel = ChannelFinal;
	//
	[_versionList addObject:info];
}

+ (UpgradeChannel)wantedChannel
{
	NSFileManager *fm = [NSFileManager defaultManager];
	if ([fm fileExistsAtPath:CHANNEL_DEV_PATH])
		return ChannelDev;
	else if ([fm fileExistsAtPath:CHANNEL_BETA_PATH])
		return ChannelBeta;
	else
		return ChannelFinal;

}

+ (NSURL*)versionListURLForChannel:(UpgradeChannel)channel
{
	NSString *s;
	switch (channel) {
		case ChannelFinal:
			s = @"final";
			break;
		case ChannelBeta:
			s = @"beta";
			break;
		case ChannelDev:
			s = @"dev";
			break;
		default:
			NSLog(@"Invalid channel param = %ld", (long)channel);
			s = @"final";
	}
	return [NSURL URLWithString:[NSString stringWithFormat:@"https://glimmerblocker.org/upgrades/info/channel-%@.json", s]];
	
}

+ (NSURL*)versionListURL
{
	return [VersionListJsonParser versionListURLForChannel:[VersionListJsonParser wantedChannel]];
}

static NSRegularExpression *sortkeyRE = NULL;

+ (NSString*)makeVersionSortKey:(NSString*)version
{
	if (!version)
		return NULL;
	if (!sortkeyRE) {
		NSError *error = NULL;
		sortkeyRE = [NSRegularExpression regularExpressionWithPattern:@"(\\d+)(?:(\\w)(\\d+))?"
															  options:0
																error:&error];
		if (error) {
			NSLog(@"Error in regexp in vsk: %@", error);
			return version;
		}
	}
	NSMutableString *sb = [NSMutableString stringWithCapacity:version.length * 4];
	NSUInteger prev = 0;
	for (NSTextCheckingResult *r in [sortkeyRE matchesInString:version options:0 range:NSMakeRange(0, version.length)]) {
		[sb appendString:[version substringWithRange:NSMakeRange(prev, r.range.location - prev)]];
		prev = r.range.location + r.range.length;
		//
		NSRange r1 = [r rangeAtIndex:1];
		[sb appendFormat:@"%03d", [version substringWithRange:r1].intValue];
		NSRange r2 = [r rangeAtIndex:2];
		if (r2.location == NSNotFound) {
			[sb appendString:@"z"];
		} else {
			[sb appendString:[version substringWithRange:r2]];
		}
		NSRange r3 = [r rangeAtIndex:3];
		if (r3.location != NSNotFound)
			[sb appendFormat:@"%04d", [version substringWithRange:r3].intValue];
	}
	[sb appendString:[version substringFromIndex:prev]];
	return sb;
}

@end

