/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "HostTextFieldCell.h"
#import "Xml.h"
#import "Filter.h"
#import "Rule.h"

@interface HostTextFieldCell()

@property NSXMLElement* ruleE;
@property Filter* filter;

@end

@implementation HostTextFieldCell

// Just returning an NSAttributedString in objectValueForTableColumn
// makes Cocoa draw the text using weird antialiasing - very ugly and the text is not very readable.
// So the colorized text drawing is done manually in 3 steps rather than using NSAttributedString.

- (void)setRuleE:(NSXMLElement*)ruleE filter:(Filter*)filter
{
	_ruleE = ruleE;
	_filter = filter;
}

- (NSString*)truncate:(NSString*)s forWidth:(CGFloat)maxWidth withAttributes:(NSDictionary*)dict
{
	if ([s sizeWithAttributes:dict].width <= maxWidth)
		return s;
	NSInteger len = [s length];
	for (int idx = 1; idx < len; idx++) {
		NSString* s2 = [@"…" stringByAppendingString:[s substringFromIndex:idx]];
		if ([s2 sizeWithAttributes:dict].width <= maxWidth)
			return s2;
	}
	return @"?"; // column narrower than a single char?
}

- (BOOL)useWhiteInView:(NSView*)controlView
{
	NSWindow* window = [controlView window];
	return [self isHighlighted] && [window isKeyWindow] && [window firstResponder] == controlView;
}

- (BOOL)isDottedIP:(NSString*)host
{
	NSInteger len = [host length];
	int minDotIndex = 1;
	for (int i = 0; i < len; i++) {
		unichar ch = [host characterAtIndex:i];
		if (ch >= '0' && ch <= '9')
			continue;
		if (ch != '.')
			return NO;
		if (i < minDotIndex)
			return NO;
		minDotIndex = i + 2;
	}
	return YES;
}

static NSColor* darkOrange;
static NSColor* darkGreen;

static NSMutableDictionary* twoLevelDomains;

static void addDomains(NSString* encoded)
{
	for (NSString* s in [encoded componentsSeparatedByString:@";"])
		twoLevelDomains[s] = @"1";
}

- (BOOL)drawColorized:(NSString*)host
			  atPoint:(NSPoint)pt
	   withAttributes:(NSMutableDictionary*)dict
	  withDomainColor:(NSColor*)domainColor
{
	if (!twoLevelDomains) {
		twoLevelDomains = [NSMutableDictionary dictionaryWithCapacity:10];
		// tons of magic TLDs:  http://wiki.greasespot.net/Magic_TLD
		addDomains(@"co.uk;org.uk;edu.uk");
		addDomains(@"com.au;org.au;edu.au");
	}
	NSInteger len = [host length];
	NSInteger lastDot = len - 1;
	while (true) {
		if (lastDot <= 0)
			return NO;
		if ([host characterAtIndex:lastDot] == '.')
			break;
		lastDot--;
	}
	NSInteger firstDot = lastDot - 1;
	while (true) {
		if (firstDot < 0)
			break;
		if ([host characterAtIndex:firstDot] == '.')
			break;
		firstDot--;
	}
	if (twoLevelDomains[[host substringFromIndex:firstDot + 1]]) {
		lastDot = firstDot--;
		while (true) {
			if (firstDot < 0)
				break;
			if ([host characterAtIndex:firstDot] == '.')
				break;
			firstDot--;
		}
	}
	NSString* s1 = (firstDot < 0) ? @"" : [host substringToIndex:firstDot + 1];
	NSString* s2 = [host substringWithRange:NSMakeRange(firstDot + 1, lastDot - firstDot - 1)];
	NSString* s3 = [host substringFromIndex:lastDot];
	//
	[s1 drawAtPoint:pt withAttributes:dict];
	pt.x += [s1 sizeWithAttributes:dict].width;
	//
	CGFloat w2 = [s2 sizeWithAttributes:dict].width;
	[s3 drawAtPoint:NSMakePoint(pt.x + w2, pt.y) withAttributes:dict];
	//
	dict[NSForegroundColorAttributeName] = domainColor;
	[s2 drawAtPoint:pt withAttributes:dict];
	return YES;
}

static NSString* phoneText = @"\u260E  "; // old-style phone symbol.

- (void)drawInteriorWithFrame:(NSRect)cellFrame inView:(NSView*)controlView
{
	if (!darkOrange) {
		darkOrange = [NSColor colorWithDeviceRed:0.51f green:0.32f blue:0 alpha:1];
		darkGreen = [NSColor colorWithDeviceRed:0 green:0.63f blue:0 alpha:1];
	}
	NSMutableDictionary* dict = [NSMutableDictionary dictionaryWithCapacity:3];
	dict[NSFontAttributeName] = [self font];
	NSPoint pt = NSMakePoint(cellFrame.origin.x, cellFrame.origin.y);
	CGFloat horzSpace = cellFrame.size.width;
	if ([Xml getBoolAttribute:_ruleE withName:@"mobile"]) {
		[phoneText drawAtPoint:pt withAttributes:dict];
		CGFloat w = [phoneText sizeWithAttributes:dict].width;
		pt.x += w;
		horzSpace -= w;
	}
	RuleType type = [Rule getRuleType:_ruleE];
	if (type == KeywordRuleType) {
		NSString* s = soft([Xml getAttribute:_ruleE withName:@"keyword"]);
		NSString* keywordType = [Xml getAttribute:_ruleE withName:@"keyword-type"];
		if ([keywordType isEqual:@"starts-with"])
			s = [s stringByAppendingString:@"*"];
		s = [self truncate:s forWidth:horzSpace withAttributes:dict];
		pt.x += horzSpace - [s sizeWithAttributes:dict].width;
		if ([keywordType isEqual:@"regexp"])
			dict[NSForegroundColorAttributeName] = darkOrange;
		else
			dict[NSForegroundColorAttributeName] = darkGreen;
		[s drawAtPoint:pt withAttributes:dict];
		return;
	}
	NSString* s = soft([Xml getAttribute:_ruleE withName:@"host"]);
	NSString* hostType = [Xml getAttribute:_ruleE withName:@"host-type"];
	if (!hostType && ![s length])
		s = @"*";
	else if ([hostType isEqual:@"domain"])
		s = [NSString stringWithFormat:@"*.%@", s];
	else if (!s)
		s = @"?";
	s = [self truncate:s forWidth:horzSpace withAttributes:dict];
	pt.x += horzSpace - [s sizeWithAttributes:dict].width;
	if ([_filter safetySimpleRulesOnly] && ![Rule ruleIsSafe:_ruleE]) {
		dict[NSForegroundColorAttributeName] = [NSColor grayColor];
		dict[NSStrikethroughStyleAttributeName] = @1U;
		[s drawAtPoint:pt withAttributes:dict];
		return;
	}
	if ([hostType isEqual:@"regexp"]) {
		dict[NSForegroundColorAttributeName] = darkOrange;
		[s drawAtPoint:pt withAttributes:dict];
		return;
	}
	NSColor* sel = [NSColor alternateSelectedControlColor];
	sel = [sel colorUsingColorSpace:[NSColorSpace deviceRGBColorSpace]];
	CGFloat r = [sel redComponent];
	CGFloat g = [sel greenComponent];
	CGFloat b = [sel blueComponent];
	NSColor* domainColor;
	if (fabs(r - g) < 0.2 && fabs(r - b) < 0.2) {
		// gray selection color.
		domainColor = [NSColor blueColor];
	} else {
		CGFloat hue, saturation, brightness, alpha;
		[sel getHue:&hue saturation:&saturation brightness:&brightness alpha:&alpha];
		//NSLog(@"hue: %f,  sat: %f,  bright: %f", hue, saturation, brightness);
		hue += 0.5f;
		if (r > 0.9f && g < 0.4f && b < 0.4f) {
			// red hightlight -> cyan text which is too bright
			//brightness *= [self useWhiteInView:controlView] ? 0.8f : 0.5f;
			hue += 0.2f; // make it full blue.
		}
		if (hue >= 1)
			hue -= 1;
		domainColor = [NSColor colorWithCalibratedHue:hue saturation:saturation brightness:brightness alpha:alpha];
	}
	if ([self useWhiteInView:controlView])
		dict[NSForegroundColorAttributeName] = [NSColor whiteColor];
	if ([self isDottedIP:s] || ![self drawColorized:s atPoint:pt withAttributes:dict withDomainColor:domainColor])
		[s drawAtPoint:pt withAttributes:dict];
}

@end
