/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "JavascriptSyntaxHighlighter.h"
#import "GBDebug.h"

#ifdef DEBUG
//#define DEBUG_JS
#endif

static inline NSColor* rgb(int r, int g, int b)
{
	return [NSColor colorWithDeviceRed:(r / 255.0f) green:(g / 255.0f) blue:(b / 255.0f) alpha:1];
}

@interface JsColor : NSObject {
}
@end

@interface JsColor()

@property NSString* name;
@property NSColor* foreground;
@property NSColor* background;

@end

@implementation JsColor

- (instancetype)init:(NSString*)name r:(int)r g:(int)g b:(int)b background:(NSColor*)background
{
	if (!(self = [super init]))
		return nil;
	_name = name;
	_foreground = rgb(r, g, b);
	_background = background;
	return self;
}

- (instancetype)init:(NSString*)name r:(int)r g:(int)g b:(int)b
{
	return [self init:name r:r g:g b:b background:NULL];
}

@end

//===================================================================================

@interface JavascriptSyntaxHighlighter ()

@property NSMutableAttributedString* attributedString;

@property JsColor* uncommittedColor; // retained in globals
@property NSFont* uncommittedFont; // retained in globals
@property NSUInteger uncommittedStart, uncommittedEnd;

@property NSString* txt;
@property NSInteger txtLength;
@property NSInteger indexOfNextChar;
@property BOOL isJavaRegexp;


- (instancetype)initWithAttributedString:(NSMutableAttributedString*)attributedString
				  isJavaRegexp:(BOOL)isJavaRegexp;
+ (void)setupGlobals;

- (void)javaRegexpMain;
- (void)javascriptMain;

@end

@implementation JavascriptSyntaxHighlighter

static JsColor* plainColor;
static JsColor* commentColor;
static JsColor* stringColor;
static JsColor* stringBackslashColor;
static JsColor* regexpStringColor;
static JsColor* regexpGroupParanColor;
static JsColor* regexpGroupOptionsColor;
static JsColor* regexpSpecialsColor;
static JsColor* regexpSpecialsDotColor;
static JsColor* regexpCharClassColor;
static JsColor* regexpBackslashColor;
static JsColor* keywordColor;
static JsColor* errorColor;
static NSMutableDictionary* jsKeywords;
static NSCharacterSet* letterSet;
static NSCharacterSet* letterDigitSet;
static NSFont* plainFont = NULL;
static NSFont* boldFont = NULL;
static NSFont* dotFont = NULL;
static NSFont* bracesFont = NULL;

+ (void)setupGlobals
{
	if (plainColor)
		return;
	NSColor* bg = rgb(255, 240, 240);
	plainColor = [[JsColor alloc] init:@"plain" r:0 g:0 b:0];
	commentColor = [[JsColor alloc] init:@"comment" r:0 g:116 b:0];
	stringColor = [[JsColor alloc] init:@"string" r:196 g:27 b:22];
	stringBackslashColor = [[JsColor alloc] init:@"string-backslash" r:178 g:178 b:178];
	regexpStringColor = [[JsColor alloc] init:@"regexp" r:160 g:25 b:20 background:bg];
	regexpGroupParanColor = [[JsColor alloc] init:@"regexp-group-paran" r:0 g:153 b:0 background:bg];
	regexpGroupOptionsColor = [[JsColor alloc] init:@"regexp-group-opts" r:0 g:153 b:0 background:bg];
	regexpSpecialsColor = [[JsColor alloc] init:@"regexp-specials" r:30 g:60 b:160 background:bg];
	regexpSpecialsDotColor = [[JsColor alloc] init:@"regexp-specials-dot" r:30 g:60 b:255 background:bg];
	regexpCharClassColor = [[JsColor alloc] init:@"regexp-charclass" r:255 g:128 b:0 background:bg];
	regexpBackslashColor = [[JsColor alloc] init:@"string-backslash" r:205 g:185 b:185 background:bg];
	keywordColor = [[JsColor alloc] init:@"keyword" r:158 g:0 b:140];
	errorColor = [[JsColor alloc] init:@"regexp-error" r:255 g:255 b:255 background:rgb(255, 0, 0)];
	//
	plainFont = [NSFont systemFontOfSize:11];
	boldFont = [NSFont boldSystemFontOfSize:12];
	dotFont = [NSFont boldSystemFontOfSize:11]; // much more visible than dot in plain Lucida Grande
	bracesFont = [NSFont boldSystemFontOfSize:11];
	//
	NSString* kw = @"abstract boolean break byte case catch char class const continue debugger "
	/**/ "default delete do double else enum export extends false final finally float "
	/**/ "for function goto if implements import in instanceof int interface let long native "
	/**/ "new null package private protected public return short static super switch "
	/**/ "synchronized this throw throws transient true try typeof var void volatile while with";
	jsKeywords = [NSMutableDictionary dictionaryWithCapacity:60];
	for (NSString* s in [kw componentsSeparatedByCharactersInSet:[NSCharacterSet whitespaceCharacterSet]])
		jsKeywords[s] = @"x";
	//
	letterSet = [NSCharacterSet letterCharacterSet];
	letterDigitSet = [NSCharacterSet alphanumericCharacterSet];
}

+ (void)highlightJavascript:(NSMutableAttributedString*)attributedString
{
	NSAssert(attributedString, @"attributedString");
	JavascriptSyntaxHighlighter* jsh = [[JavascriptSyntaxHighlighter alloc] initWithAttributedString:attributedString isJavaRegexp:NO];
	[jsh javascriptMain];
}

+ (void)highlightJavaRegExp:(NSMutableAttributedString*)attributedString
{
	NSAssert(attributedString, @"attributedString");
	JavascriptSyntaxHighlighter* jsh = [[JavascriptSyntaxHighlighter alloc] initWithAttributedString:attributedString isJavaRegexp:YES];
	[jsh javaRegexpMain];
}


- (instancetype)initWithAttributedString:(NSMutableAttributedString*)attributedString
				  isJavaRegexp:(BOOL)isJavaRegexp;
{
	[JavascriptSyntaxHighlighter setupGlobals];
	if (!(self = [super init]))
		return nil;
	_isJavaRegexp = isJavaRegexp;
	_attributedString = attributedString;
	_uncommittedColor = plainColor;
	_uncommittedStart = 0;
	_txt = [_attributedString string];
	_txtLength = [_txt length];
	_indexOfNextChar = 0;
	return self;
}


- (void)setAttribute:(NSString*)attrName toValue:(id)wantedValue
{
	NSRange range = NSMakeRange(_uncommittedStart, _uncommittedEnd - _uncommittedStart);
	NSRange attrRange = NSMakeRange(0, 0);
	id attr = [_attributedString attribute:attrName
								   atIndex:_uncommittedStart
					 longestEffectiveRange:&attrRange
								   inRange:NSMakeRange(0, _txtLength)];
	NSInteger rangeEnd = range.location + range.length;
	NSInteger attrRangeEnd = attrRange.location + attrRange.length;
	if (attr == wantedValue && attrRange.location <= range.location && attrRangeEnd >= rangeEnd)
		return;
	if (wantedValue) {
		[_attributedString addAttribute:attrName
								  value:wantedValue
								  range:range];
	} else {
		[_attributedString removeAttribute:attrName
									 range:range];
	}
}

- (void)commitTextAttributes
{
#ifdef DEBUG_JS
	DebugNSLog(@"Commits: %d -> %d with %@", _uncommittedStart, _uncommittedEnd, [_uncommittedColor name]);
#endif
	if (_uncommittedEnd <= _uncommittedStart)
		return;
	[self setAttribute:NSFontAttributeName toValue:_uncommittedFont];
	if (_isJavaRegexp && _uncommittedColor == errorColor)
		[self setAttribute:NSForegroundColorAttributeName toValue:[NSColor redColor]]; // white text without colored background is invisible.
	else
		[self setAttribute:NSForegroundColorAttributeName toValue:[_uncommittedColor foreground]];
	if (!_isJavaRegexp) // ugly, don't want background color in host/path/query fields.
		[self setAttribute:NSBackgroundColorAttributeName toValue:[_uncommittedColor background]];
	_uncommittedStart = _uncommittedEnd;
}

- (void)setColor:(JsColor*)color font:(NSFont*)font toIndex:(NSInteger)toIndex
{
	if (color == _uncommittedColor && font == _uncommittedFont) {
		//DebugNSLog(@"Extends uncommit: %d -> %d -> %d of %@", _uncommittedStart, _uncommittedEnd, toIndex, [color name]);
		_uncommittedEnd = toIndex;
	} else {
		[self commitTextAttributes];
		_uncommittedColor = color;
		_uncommittedFont = font;
		_uncommittedEnd = MIN(toIndex, _txtLength);
#ifdef DEBUG_JS
		DebugNSLog(@"Starts uncommitted: %d -> %d using: %@ and %@", _uncommittedStart, _uncommittedEnd, [color name], [font displayName]);
#endif
	}
}

- (void)setColor:(JsColor*)color toIndex:(NSInteger)toIndex
{
	[self setColor:color font:plainFont toIndex:toIndex];
}

- (unichar)peek:(NSInteger)delta
{
	NSInteger idx = _indexOfNextChar + delta;
	if (idx >= 0 && idx < _txtLength)
		return [_txt characterAtIndex:idx];
	else
		return ' ';
}

- (unichar)eat
{
	unichar ch = [self peek:0];
	_indexOfNextChar++;
	return ch;
}

inline static BOOL isDecimalDigit(unichar ch)
{
	return ch >= '0' && ch <= '9';
}

inline static BOOL isHexDigit(unichar ch)
{
	return isDecimalDigit(ch) || (ch >= 'A' && ch <= 'F') || (ch >= 'a' && ch <= 'f');
}

- (void)parseEscapeHexDigits:(int)numDigits
{
	for (int i = 0; i < numDigits; i++) {
		if (!isHexDigit([self eat])) {
			[self setColor:errorColor toIndex:_indexOfNextChar];
			return;
		}
	}
	[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
}

- (void)regexpGroupOptions
{
	[self eat]; // '?'
	unichar ch2 = [self eat];
	switch (ch2) {
		case ':': //  (?:  simple anon group
		case '=': //  (?=  positive look-ahead
		case '!': //  (?!  negative look-ahead
			return;
		case ')': //  (?)  is invalid
			_indexOfNextChar--;
			[self setColor:errorColor toIndex:_indexOfNextChar];
			return;
	}
	if (!_isJavaRegexp) {
		[self setColor:errorColor toIndex:_indexOfNextChar];
		return;
	}
	if (ch2 == '>') //  (?>  anon group without backtracking.
		return;
	if (ch2 == '<') {
		switch ([self peek:0]) {
			case '=': //  (?<=  zero-width positive lookbehind
			case '!': //  (?<!  zero-width negative lookbehind
				[self eat];
				break;
			default:
				[self setColor:errorColor toIndex:_indexOfNextChar];
		}
		return;
	}
	int dash = 0;
	for (int i = 0; _indexOfNextChar < _txtLength; i++) {
		if (i > 0)
			ch2 = [self eat];
		if (ch2 == ':')
			return;
		if (ch2 == ')') {
			_indexOfNextChar--;
			return;
		}
		if (ch2 == '-') {
			if (++dash > 1) {
				[self setColor:errorColor toIndex:_indexOfNextChar];
				return;
			}
			continue;
		}
		if (ch2 == 'i' || ch2 == 'd' || ch2 == 'm' || ch2 == 's' || ch2 == 'u' || ch2 == 'x')
			continue;
		int ch3 = [self peek:0];
		if (!i || ch3 != ')')
			[self setColor:errorColor toIndex:_indexOfNextChar];
		return;
	}
}

- (void)regexpRepeatBraces
{
	int commas = 0;
	NSInteger startIndex = _indexOfNextChar;
	int min = 0;
	int max = 0;
	BOOL hasMin = NO;
	BOOL hasMax = NO;
	for (int i = 0; true; i++) {
		if (i > 200) {
			[self setColor:errorColor font:bracesFont toIndex:startIndex];
			[self setColor:errorColor toIndex:_indexOfNextChar]; // unmatched '{'
			return;
		}
		unichar ch = [self eat];
		if (ch >= '0' && ch <= '9') {
			if (commas > 0) {
				max = max * 10 + ch - '0';
				hasMax = YES;
			} else {
				min = min * 10 + ch - '0';
				hasMin = YES;
			}
			continue;
		}
		if (ch == ',') {
			commas++;
			continue;
		}
		if (ch != '}') {
			[self setColor:errorColor font:bracesFont toIndex:startIndex];
			[self setColor:errorColor toIndex:_indexOfNextChar];
			return;
		}
		// i == 0 means empty {}
		BOOL hasError = !hasMin || (hasMax && max < min) || i == 0 || commas > 1;
		JsColor* color = hasError ? errorColor : regexpSpecialsColor;
		[self setColor:color font:bracesFont toIndex:startIndex];
		[self setColor:color toIndex:_indexOfNextChar - 1];
		[self setColor:color font:bracesFont toIndex:_indexOfNextChar];
		return;
	}
}

- (void)parseRegexp
{
	[self setColor:regexpStringColor toIndex:_indexOfNextChar];
	// http://developer.mozilla.org/en/Core_JavaScript_1.5_Guide/Regular_Expressions
	// http://java.sun.com/j2se/1.5.0/docs/api/java/util/regex/Pattern.html
	int groupNestingLevel = 0;
	int captureGroupCount = 0;
	BOOL hasRepeatableItem = NO;
	int squareCount = 0;
	while (_indexOfNextChar < _txtLength) {
		unichar ch = [self eat];
		if (ch == '\n' || ch == '\r') {
			[self setColor:errorColor toIndex:_indexOfNextChar];
			return;
		}
		if (ch == '\\') {
			hasRepeatableItem = YES;
			ch = [self eat];
			if (_isJavaRegexp) switch (ch) {
				case 'a': // alert/bell (ascii 7)
				case 'e': // escape (ascii 27)
				case 'z': // end of input
				case 'Z': // end of the line
					[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
					continue;
				case 'p': // posix class: \p{Lower}
					if ([self eat] != '{') {
						[self setColor:errorColor toIndex:_indexOfNextChar];
						continue;
					}
					while ([self eat] != '}') {
						if (_indexOfNextChar >= _txtLength) {
							[self setColor:errorColor toIndex:_indexOfNextChar];
							break;
						}
					}
					[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
					continue;
				case 'E': // end of \Q quoting  (unmatched)
					[self setColor:errorColor toIndex:_indexOfNextChar];
					continue;
				case 'Q': // quotes until \E
					[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
					while (_indexOfNextChar < _txtLength) {
						ch = [self eat];
						if (ch == '\\' && [self peek:0] == 'E') {
							[self setColor:regexpStringColor toIndex:_indexOfNextChar - 1];
							[self eat];
							[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
							break;
						}
					}
					[self setColor:regexpStringColor toIndex:_indexOfNextChar]; // missing \E is allowed.
					continue;
			}
			switch (ch) {
				case 'c': // \cM is control-M
					ch = [self eat];
					if ((ch >= 64 && ch <= 64+31) || (ch >= 96 && ch <= 96+31))
						[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
					else
						[self setColor:errorColor toIndex:_indexOfNextChar];
					break;
				case 'x': // \xHH  hexchar
					[self parseEscapeHexDigits:2];
					break;
				case 'u': // \uHHHH  unicode hexchar
					[self parseEscapeHexDigits:4];
					break;
				case 'b': // word boundary
				case 'B': // not \b
					if (squareCount == 0)
						hasRepeatableItem = NO;
					[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
					break;
				case 'd': // digit
				case 'D': // not \d
				case 'f': // form feed
				case 'n': // linefeed
				case 'r': // carriage return
				case 's': // whitespace
				case 'S': // not whitespace
				case 't': // tab
				case 'v': // vertical tab
				case 'w': // alphanum + underscore
				case 'W': // not \w
				case '0': // NUL char.
					[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
					break;
				case '.':
					[self setColor:regexpBackslashColor toIndex:_indexOfNextChar - 1];
					[self setColor:regexpStringColor font:dotFont toIndex:_indexOfNextChar];
					break;
				default:
					if (squareCount == 0 && ch >= '1' && ch <= '9') {
						int num = ch - '0';
						while ([self peek:0] >= '0' && [self peek:0] <= '9')
							num = num * 10 + [self eat] - '0';
						if (num <= captureGroupCount)
							[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
						else
							[self setColor:errorColor toIndex:_indexOfNextChar];
					} else {
						[self setColor:regexpBackslashColor toIndex:_indexOfNextChar - 1];
						[self setColor:regexpStringColor toIndex:_indexOfNextChar];
					}
			}
			continue;
		}
		if (ch == ']') {
			if (squareCount) {
				squareCount--;
				[self setColor:regexpCharClassColor toIndex:_indexOfNextChar];
			} else {
				[self setColor:errorColor toIndex:_indexOfNextChar];
			}
			hasRepeatableItem = YES;
			continue;
		}
		if (ch == '[') {
			if (squareCount > (_isJavaRegexp ? 1 : 0)) {
				[self setColor:errorColor toIndex:_indexOfNextChar];
				continue;
			}
			int idx = 0;
			if ([self peek:0] == '^')
				idx++;
			if ([self peek:idx] == '$')
				idx++;
			if ([self peek:idx] == ']') {
				for (int i = 0; i <= idx; i++) // include ']'
					[self eat];
				[self setColor:errorColor toIndex:_indexOfNextChar]; // empty [] is invalid
				continue;
			}
			squareCount++;
			[self setColor:regexpCharClassColor toIndex:_indexOfNextChar];
			continue;
		}
		if (squareCount) {
			if (ch == '-' && [self peek:0] != ']' && [self peek:-2] != '[' && [self peek:-2] != '^') {
				// interval: [a-z]
				[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
				continue;
			}
			if (ch == '^' && [self peek:-2] == '[') {
				// anchor: [^a]
				[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
				continue;
			}
			if (ch == '$' && [self peek:0] == ']') {
				// anchor: [a$]
				[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
				continue;
			}
			if (_isJavaRegexp && ch == '&' && [self peek:0] == '&' && [self peek:1] == '[') {
				[self eat];
				[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
			} else {
				[self setColor:regexpCharClassColor toIndex:_indexOfNextChar];
			}
			continue;
		}
		if (ch == '(') {
			[self setColor:regexpGroupParanColor font:boldFont toIndex:_indexOfNextChar];
			groupNestingLevel++;
			BOOL anon = ([self peek:0] == '?');
			if (anon)
				[self regexpGroupOptions];
			else
				captureGroupCount++;
			[self setColor:regexpGroupOptionsColor toIndex:_indexOfNextChar];
			hasRepeatableItem = NO;
			continue;
		}
		if (ch == ')') {
			if (groupNestingLevel > 0) {
				groupNestingLevel--;
				[self setColor:regexpGroupParanColor font:boldFont toIndex:_indexOfNextChar];
			} else {
				[self setColor:errorColor toIndex:_indexOfNextChar];
			}
			hasRepeatableItem = YES;
			continue;
		}
		if (ch == '{') {
			if (hasRepeatableItem)
				[self regexpRepeatBraces];
			else
				[self setColor:errorColor font:bracesFont toIndex:_indexOfNextChar];
			continue;
		}
		if (ch == '}') {
			[self setColor:errorColor font:bracesFont toIndex:_indexOfNextChar]; // unmatched '}'
			continue;
		}
		if (ch == '.') {
			hasRepeatableItem = YES;
			[self setColor:regexpSpecialsDotColor font:dotFont toIndex:_indexOfNextChar];
			continue;
		}
		if (ch == '^' || ch == '$' || ch == '|') {
			hasRepeatableItem = NO;
			[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
			continue;
		}
		if (ch == '?' || ch == '*' || ch == '+') {
			if (hasRepeatableItem)
				[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
			else
				[self setColor:errorColor toIndex:_indexOfNextChar];
			continue;
		}
		if (!_isJavaRegexp && ch == '/')
			break;
		[self setColor:regexpStringColor toIndex:_indexOfNextChar];
		hasRepeatableItem = YES;
	}
	if (_isJavaRegexp)
		return;
	[self setColor:regexpSpecialsColor font:boldFont toIndex:_indexOfNextChar];
	while ([self peek:0] == 'g' || [self peek:0] == 'i' || [self peek:0] == 'm')
		[self eat];
	[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
}

- (void)parseString:(unichar)quote
{
	while (true) {
		unichar ch = [self eat];
		if (ch == quote) {
			[self setColor:stringColor toIndex:_indexOfNextChar];
			return;
		}
		if (_indexOfNextChar >= _txtLength || ch == '\r' || ch == '\n') {
			[self setColor:errorColor toIndex:_indexOfNextChar];
			return;
		}
		if (ch == '.') {
			[self setColor:stringColor font:dotFont toIndex:_indexOfNextChar];
			continue;
		}
		if (ch != '\\') {
			[self setColor:stringColor toIndex:_indexOfNextChar];
			continue;
		}
		ch = [self eat];
		if (isDecimalDigit(ch)) {
			// old obsolete octal escape. Mark as error as it's 99% probable a bug in the transform script.
			for (int i = 0; isDecimalDigit(ch) && i < 2; i++)
				ch = [self eat];
			[self setColor:errorColor toIndex:_indexOfNextChar];
			continue;
		}
		switch (ch) {
			case 'x':
				[self parseEscapeHexDigits:2];
				continue;
			case 'u':
				[self parseEscapeHexDigits:4];
				continue;
			case '0': // NUL char.
			case 'b': // backspace, ascii 8
			case 't': // tab
			case 'n': // linefeed
			case 'v': // vertical tab
			case 'f': // form feed
			case 'r': // carriage return
				[self setColor:regexpSpecialsColor toIndex:_indexOfNextChar];
				break;
			default:
				[self setColor:stringBackslashColor toIndex:_indexOfNextChar - 1];
				if (ch == '.')
					[self setColor:stringColor font:dotFont toIndex:_indexOfNextChar];
				else
					[self setColor:stringColor toIndex:_indexOfNextChar];
		}
	}
}

- (void)javaRegexpMain
{
#ifdef DEBUG_JS
	DebugNSLog(@"==============================================================");
	DebugNSLog(@"%@", _txt);
	DebugNSLog(@"==============================================================");
#endif
	[self parseRegexp];
	[self setColor:plainColor toIndex:_txtLength];
	[self commitTextAttributes];
}

- (void)javascriptMain
{
#ifdef DEBUG_JS
	DebugNSLog(@"==============================================================");
	DebugNSLog(@"%@", _txt);
	DebugNSLog(@"==============================================================");
#endif
	BOOL regexpCanStartHere = YES; // whicked hack to avoid creating full tokenizer. Seems to work, though.
	while (_indexOfNextChar < _txtLength) {
		unichar ch = [self eat];
		unichar q;
		NSInteger start;
		switch (ch) {
			case '/':
				if ([self peek:0] == '/') {
					while (ch != 13 && ch != 10 && _indexOfNextChar < _txtLength)
						ch = [self eat];
					[self setColor:commentColor toIndex:_indexOfNextChar];
					continue;
				}
				if ([self peek:0] == '*') {
					[self eat];
					while ([self eat] != '*' || [self peek:0] != '/') {
						if (_indexOfNextChar >= _txtLength)  {
							[self setColor:errorColor toIndex:_indexOfNextChar];
							break;
						}
					}
					[self eat]; // last '/'
					[self setColor:commentColor toIndex:_indexOfNextChar];
					continue;
				}
				if (regexpCanStartHere) {
					[self setColor:regexpSpecialsColor font:boldFont toIndex:_indexOfNextChar];
					[self parseRegexp];
					regexpCanStartHere = NO;
					continue;
				}
				break;
			case '"':
			case '\'':
				[self parseString:ch];
				continue;
			case '(':
			case '=':
			case ',':
			case '?':
			case ':':
			case ';':
			case '\n':
				regexpCanStartHere = YES;
				break;
			case ' ':
			case '\t':
				break;
			case '{':
			case '}':
				[self setColor:plainColor font:bracesFont toIndex:_indexOfNextChar];
				regexpCanStartHere = YES;
				continue;
			default:
				regexpCanStartHere = NO;
				q = [self peek:-2];
				if ([letterSet characterIsMember:ch] && ![letterDigitSet characterIsMember:q]) {
					start = _indexOfNextChar - 1;
					while (_indexOfNextChar < _txtLength) {
						ch = [self eat];
						if (![letterDigitSet characterIsMember:ch]) {
							_indexOfNextChar--;
							break;
						}
					}
					NSInteger len = MIN(_txtLength, _indexOfNextChar - start);
					NSString* s = [_txt substringWithRange:NSMakeRange(start, len)];
					//DebugNSLog(@"token %d -> %d = %d: {{%@}}", start, _indexOfNextChar, len, s);
					if (jsKeywords[s]) {
						[self setColor:keywordColor toIndex:_indexOfNextChar];
						continue;
					}
				}
		}
		[self setColor:plainColor toIndex:_indexOfNextChar];
	}
	[self setColor:plainColor toIndex:_txtLength];
	[self commitTextAttributes];
}

@end
