/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "HistoryController.h"
#import "GBDebug.h"
#import "XmlDoc.h"
#import "Xml.h"
#import "ServerCommunication.h"
#import "AlertHelper.h"
#import "Prefs.h"
#import "Filter.h"
#import "FilterTabController.h"
#import "FilterRuleEdit.h"
#import "AutoZap.h"

@interface HistoryController ()

@property FilterTabController* filterTabController;
@property Prefs* prefs;
@property AlertHelper* alertHelper;
@property ServerCommunication* comm;
@property BOOL visible;
@property BOOL enabled;
@property NSTimer* timer;
@property NSMutableArray* entryList;
@property NSArray *topLevelObjects;


@property (weak) IBOutlet NSView*			subView;
@property (weak) IBOutlet NSSplitView*		splitView;

@property (weak) IBOutlet NSButton*			enableCB;
@property (weak) IBOutlet NSTextField*		emptyHistoryLabel;

@property (weak) IBOutlet HistoryTableView*	historyTableView;

@property (weak) IBOutlet NSTextField*		rulesLabel;
@property (weak) IBOutlet NSTableView*		rulesTableView;
@property (weak) IBOutlet NSButton*			editRuleBtn;



- (void)timerCallback:(NSTimer*)theTimer;

- (NSString*)ruleDescription:(NSInteger)row
		 withFallbackMessage:(BOOL)fallback;

- (void)updateEmptyHistoryMessageDisplay;

- (void)tableViewSelectionDidChange:(NSNotification *)aNotification;

@end

static NSColor* blockedColor;
static NSColor* whitelistedColor;
static NSColor* modifiedColor;
static NSColor* keywordColor;
static NSMutableParagraphStyle* pstyle;

static inline NSColor* rgb(int r, int g, int b)
{
	return [NSColor colorWithDeviceRed:(r / 255.0f) green:(g / 255.0f) blue:(b / 255.0f) alpha:1];
}


@implementation HistoryController

- (instancetype)initWithFilterTabController:(FilterTabController*)filterTabController
				withContainerView:(NSView*)containerView
					   withBundle:(NSBundle*)bundle
{
	if (!(self = [super init]))
		return nil;
	if (!blockedColor) {
		blockedColor = rgb(128, 0, 0);
		whitelistedColor = rgb(0, 128, 0);
		modifiedColor = rgb(0, 0, 192);
		keywordColor = rgb(141, 92, 0);
		pstyle = [[NSParagraphStyle defaultParagraphStyle] mutableCopy];
		[pstyle setLineBreakMode:NSLineBreakByTruncatingTail];
		[pstyle setTabStops:@[]];
		for (int i = 1; i < 10; i++) {
			CGFloat x = i * 25;
			[pstyle addTabStop:[[NSTextTab alloc] initWithType:NSLeftTabStopType location:x]];
		}
	}
	_topLevelObjects = [AutoZap loadNibNamed:@"HistoryList" bundle:[filterTabController mainBundle] owner:self];
	NSAssert(_historyTableView, @"historyTableView is null");
	_filterTabController = filterTabController;
	_comm = [filterTabController comm];
	_prefs = [_comm prefs];
	_alertHelper = [_prefs alertHelper];
	_entryList = [NSMutableArray arrayWithCapacity:200];
	_enabled = [_prefs boolForXPath:@"history/@enabled"];
	[_enableCB setState:_enabled];
	[_editRuleBtn setEnabled:NO]; // doesn't seem to get disable initially by Cocoa in [tableViewSelectionDidChange]
	[containerView addSubview:_subView];
	[_historyTableView setDataSource:self];
	[_historyTableView setDelegate:self];
	[_rulesTableView setDataSource:self];
	[_rulesTableView setDelegate:self];
	[_historyTableView setTarget:self];
	[_rulesTableView setTarget:self];
	[_historyTableView setDoubleAction:@selector(editAction:)];
	[_rulesTableView setDoubleAction:@selector(editAction:)];
	[self notifyUpdatedAuthorizationState];
	[_historyTableView registerForDraggedTypes:@[NSStringPboardType, NSURLPboardType]];
	[_historyTableView setDraggingSourceOperationMask:NSDragOperationCopy forLocal:NO];
	[_historyTableView setDraggingSourceOperationMask:NSDragOperationCopy forLocal:YES];
	[_rulesTableView reloadData];
	[_splitView setDelegate:self];
	[self tableViewSelectionDidChange:NULL];
	[self updateEmptyHistoryMessageDisplay];
	return self;
}

- (void)dealloc
{
	DebugNSLog(@"HistoryList.dealloc");
	[self notifyBecameHidden];
	[_emptyHistoryLabel removeFromSuperview];
	[_subView removeFromSuperview];
}

- (void)startTimerWithImmediateExecute:(BOOL)immediate
{
	if (!_enabled)
		return;
	if (!immediate && [_timer isValid])
		return;
	[self stopTimer];
	if (immediate)
		[self timerCallback:NULL];
	_timer = [NSTimer scheduledTimerWithTimeInterval:2.0
											   target:self
											 selector:@selector(timerCallback:)
											 userInfo:NULL
											  repeats:NO];
}

- (void)stopTimer
{
	if (!_timer)
		return;
	[_timer invalidate];
	_timer = NULL;
}

- (void)notifyBecameVisible
{
	_visible = YES;
	[self startTimerWithImmediateExecute:YES];
}

- (void)notifyBecameHidden
{
	_visible = NO;
	[self stopTimer];
}

- (void)notifyUpdatedAuthorizationState
{
	BOOL hasAuth = [_prefs hasGlimmerAuth];
	[_enableCB setEnabled:hasAuth];
	[self tableViewSelectionDidChange:NULL];
}

- (void)updateEmptyHistoryMessageDisplay
{
	BOOL showEmptyMessage = _enabled && ![_entryList count];
	[_emptyHistoryLabel setHidden:!showEmptyMessage];
}

- (void)timerCallback:(NSTimer*)theTimer
{
	[self stopTimer];
	if (!_enabled) {
		[_entryList removeAllObjects];
		[_historyTableView reloadData];
		[self updateEmptyHistoryMessageDisplay];
		[self tableViewSelectionDidChange:NULL];
		return;
	}
	if (!_visible) {
		DebugNSLog(@"Suspects, Timer callback without being visible");
		return;
	}
	NSString* pathArgs = [NSString stringWithFormat:@"/history/list"];
	NSError* __autoreleasing error = NULL;
	XmlDoc* doc = [_comm curlToXml:pathArgs
			 withActionDescription:@"get history list"
					  withErrorRef:&error];
	[self startTimerWithImmediateExecute:NO];
	NSArray* a = doc ? [[doc rootE] elementsForName:@"entry"] : NULL;
	NSUInteger num = [a count];
	if (!num) {
		[_entryList removeAllObjects];
		[_historyTableView reloadData];
		[self updateEmptyHistoryMessageDisplay];
		[self tableViewSelectionDidChange:NULL];
		return;
	}
	NSString* selectedUrl = NULL;
	NSString* selectedActions = NULL;
	NSInteger row = [_historyTableView selectedRow];
	if (row >= 0) {
		NSXMLElement* e = _entryList[row];
		selectedUrl = [Xml getAttribute:e withName:@"url"];
		selectedActions = [Xml getAttribute:e withName:@"actions"];
	}
	[_entryList setArray:a];
	[_historyTableView reloadData];
	if (selectedUrl) {
		BOOL found = NO;
		for (NSUInteger i = 0; i < num; i++) {
			NSXMLElement* e = a[i];
			if ([selectedUrl isEqual:[Xml getAttribute:e withName:@"url"]]
				&& [selectedActions isEqual:[Xml getAttribute:e withName:@"actions"]]) {
				NSIndexSet* set = [NSIndexSet indexSetWithIndex:i];
				[_historyTableView selectRowIndexes:set byExtendingSelection:NO];
				//DebugNSLog(@"Suspects.timer.selected-row: %d -> %d", row, i);
				found = YES;
				break;
			}
		}
		if (!found) {
			[_historyTableView deselectAll:self];
		}
	}
	[self updateEmptyHistoryMessageDisplay];
}

- (NSArray*)getActionListForSelectedHistoryEntry
{
	NSInteger row = [_historyTableView selectedRow];
	if (row < 0 || row >= (int)[_entryList count])
		return NULL;
	NSXMLElement* e = _entryList[row];
	NSString* encodedActions = [Xml getAttribute:e withName:@"actions"];
	if (![encodedActions length])
		return NULL;
	return [encodedActions componentsSeparatedByString:@"\u2A1F"]; // ⨟
}

- (void)tableViewSelectionDidChange:(NSNotification *)aNotification
{
	NSArray* actionList = [self getActionListForSelectedHistoryEntry];
	NSInteger numRules = [actionList count];
	if (!aNotification || [aNotification object] == _historyTableView) {
		NSString* s;
		NSString* action1 = (numRules == 1) ? actionList[0] : NULL;
		if (numRules == 0)
			s = @"";
		else if ([action1 hasSuffix:@"\u29BFblocked"])
			s = @"1 rule blocked the request";
		else if ([action1 hasSuffix:@"\u29BFwhitelisted"])
			s = @"1 rule whitelisted the request";
		else if (numRules == 1)
			s = @"1 rule modified the request";
		else
			s = [NSString stringWithFormat:@"%ld rules modified the request", numRules];
		[_rulesLabel setStringValue:s];
		[_rulesTableView reloadData];
		if (numRules && [_rulesTableView selectedRow] < 0) {
			[_rulesTableView selectRowIndexes:[NSIndexSet indexSetWithIndex:0]
						byExtendingSelection:NO];
		}
	}
	NSInteger row = numRules ? [_rulesTableView selectedRow] : -1;
	if (row < 0) {
		[_editRuleBtn setEnabled:NO];
		return;
	}
	NSString* s = [self ruleDescription:row withFallbackMessage:NO];
	[_editRuleBtn setEnabled:!!s];
}

- (IBAction)enableAction:(id)sender
{
	_enabled = [_enableCB state];
	DebugNSLog(@"enableAction: %d", _enabled);
	[_prefs setBool:_enabled forXPath:@"history/@enabled"];
	[_prefs markAsDirty];
	[_comm saveDirtyPrefsAndNotify];
	[self updateEmptyHistoryMessageDisplay];
	if (_enabled) {
		[self startTimerWithImmediateExecute:YES];
	} else {
		[self stopTimer];
		[_entryList removeAllObjects];
		[_historyTableView reloadData];
	}
}

- (void)openFilterSheet
{
	NSArray* actions = [self getActionListForSelectedHistoryEntry];
	if (!actions) {
		NSLog(@"HistoryController: openFilterSheet without selected row");
		return;
	}
	NSInteger selectedActionIndex = ([actions count] >= 2) ? [_rulesTableView selectedRow] : 0;
	if (selectedActionIndex < 0)
		return;
	NSString* action = actions[selectedActionIndex];
	NSArray* infos = [action componentsSeparatedByString:@"\u29BF"]; // ⦿
	if ([infos count] < 3) {
		NSLog(@"too few items in action: '%@'", action);
		return;
	}
	int filterId = [infos[1] intValue];
	int ruleId = [infos[2] intValue];
	Filter* filter = [_filterTabController getFilterById:filterId];
	if (!filter) {
		NSLog(@"HistoryController: Filter %d not found", filterId);
		return;
	}
	NSXMLElement* ruleE = [filter ruleWithId:ruleId];
	if (!ruleE) {
		NSLog(@"HistoryController: Didn't find rule %d in filter %d", ruleId, filterId);
		return;
	}
	FilterRuleEdit *editor = [[FilterRuleEdit alloc] initWithFilterTabController:_filterTabController
												 withOwnerFilter:filter
												 withRuleElement:ruleE
											  withSuspectElement:NULL];
	[AutoZap add:editor];
}

- (IBAction)editAction:(id)sender
{
	[self openFilterSheet];
}

- (void)getFormattedActions:(NSXMLElement*)e
					verbose:(BOOL)verbose
					   into:(NSMutableString*)sb
{
	NSString* encodedActions = [Xml getAttribute:e withName:@"actions"];
	NSArray* actionList = [encodedActions componentsSeparatedByString:@"\u2A1F"];
	for (NSUInteger idx = 0; idx < [actionList count]; idx++) {
		NSString* action = actionList[idx];
		if (verbose) {
			if (idx > 0)
				[sb appendString:@"\n"];
			[sb appendFormat:@"Rule #%ld: ", idx + 1];
		}
		NSArray* infos = [action componentsSeparatedByString:@"\u29BF"]; // ⦿
		if ([infos count] < 3) {
			[sb appendString:(verbose ? @"???" : @"?")];
			NSLog(@"too few items in action: '%@'", action);
			continue;
		}
		BOOL first = YES;
		for (NSString* a in [infos[3] componentsSeparatedByString:@"\u2992"]) { // ⦒
			if (first)
				first = NO;
			else if (verbose)
				[sb appendString:@" + "];
			if (verbose)
				[sb appendString:a];
			else if ([a isEqualToString:@"blocked"])
				[sb appendString:@"\u2716"]; // ✖
			else if ([a isEqualToString:@"whitelisted"])
				[sb appendString:@"\u2713"]; // ✓
			else if ([a isEqualToString:@"css-added"])
				[sb appendString:@"c"];
			else if ([a isEqualToString:@"javascript-added"])
				[sb appendString:@"j"];
			else if ([a isEqualToString:@"transformed"])
				[sb appendString:@"t"];
			else if ([a isEqualToString:@"keyword"])
				[sb appendString:@"kw"];
			else if ([a isEqualToString:@"request"])
				[sb appendString:@"r"];
			else
				[sb appendString:@"?"];
		}
	}
}

- (NSString *)tableView:(NSTableView *)aTableView
		 toolTipForCell:(NSCell *)aCell
				   rect:(NSRectPointer)rect
			tableColumn:(NSTableColumn *)aTableColumn
					row:(NSInteger)row
		  mouseLocation:(NSPoint)mouseLocation
{
	if (aTableView != _historyTableView)
		return @"";
	NSXMLElement* e = _entryList[row];
	NSString* referer = [Xml getAttribute:e withName:@"referer"];
	NSMutableString* sb = [[NSMutableString alloc] initWithCapacity:[referer length] + 20];
	[self getFormattedActions:e verbose:YES into:sb];
	if ([referer length]) {
		if ([sb length])
			[sb appendString:@"\n\n"];
		[sb appendString:@"Referer:\n"];
		[sb appendString:referer];
	}
	return sb;
}

- (id)objectValueForHistoryTable:(NSInteger)row
{
	NSXMLElement* e = _entryList[row];
	NSString* actions = [Xml getAttribute:e withName:@"actions"];
	NSColor* color;
	if ([actions hasSuffix:@"\u29BFblocked"])
		color = blockedColor;
	else if ([actions hasSuffix:@"\u29BFwhitelisted"])
		color = whitelistedColor;
	else if ([actions hasSuffix:@"\u29BFkeyword"])
		color = keywordColor;
	else
		color = modifiedColor;
	NSString* url = [Xml getAttribute:e withName:@"url"];
	NSMutableString* sb = [[NSMutableString alloc] initWithCapacity:[url length] + 10];
	[self getFormattedActions:e verbose:NO into:sb];
	NSInteger actionsCharCount = [sb length];
	[sb appendString:@"\t"];
	[sb appendString:url];
	NSMutableDictionary* dict = [NSMutableDictionary dictionaryWithObject:pstyle forKey:NSParagraphStyleAttributeName];
	if (color != blockedColor)
		dict[NSForegroundColorAttributeName] = color;
	NSMutableAttributedString* attrString = [[NSMutableAttributedString alloc]
											  initWithString:sb attributes:dict];
	if (color == blockedColor) {
		[attrString addAttribute:NSForegroundColorAttributeName
						   value:color
						   range:NSMakeRange(0, actionsCharCount)];
	}
	return attrString;
}

- (BOOL)tableView:(NSTableView *)view writeRowsWithIndexes:(NSIndexSet *)rowIndexes toPasteboard:(NSPasteboard *)pboard
{
	DebugNSLog(@"writeRows.%d: %@", (view == _historyTableView), rowIndexes);
	if (view != _historyTableView || [rowIndexes count] != 1)
		return NO;
	NSXMLElement* e = _entryList[[rowIndexes firstIndex]];
	NSString* u = [Xml getAttribute:e withName:@"url"];
	[pboard declareTypes:@[NSStringPboardType, NSURLPboardType] owner:nil];
	[pboard setString:u forType:NSStringPboardType];
	@try {
		[[NSURL URLWithString:u] writeToPasteboard:pboard];
	}
	@catch (NSException* ex) {
		NSLog(@"Couldn't copy URL to pasteboard: %@", u);
	}
	return YES;
}

- (void)copyFromHistoryTable
{
	NSInteger row = [_historyTableView selectedRow];
	if (row < 0 || row >= (int)[_entryList count])
		return;
	NSIndexSet *set = [NSIndexSet indexSetWithIndex:row];
	[self tableView:_historyTableView writeRowsWithIndexes:set toPasteboard:[NSPasteboard generalPasteboard]];
}


// ------------------------------------------------------------------------------- rules

- (NSString*)ruleDescription:(NSInteger)row
		 withFallbackMessage:(BOOL)fallback
{
	NSArray* actionList = [self getActionListForSelectedHistoryEntry];
	if (row < 0 || row >= (int)[actionList count])
		return fallback ? @"???" : NULL;
	NSString* encodedActions = actionList[row];
	NSArray* infos = [encodedActions componentsSeparatedByString:@"\u29BF"]; // ⦿
	if ([infos count] < 4)
		return fallback ? encodedActions : NULL;
	BOOL isPreview = !![infos[0] intValue];
	int filterId = [infos[1] intValue];
	if (filterId <= 0) {
		if (!fallback)
			return NULL;
		if (isPreview)
			return @"(Preview of new rule)";
		else
			return @"(Unidentified filter)";
	}
	Filter* filter = [_filterTabController getFilterById:filterId];
	if (!filter) {
		if (!fallback)
			return NULL;
		if (isPreview)
			return @"(Preview of rule from deleted filter)";
		else
			return @"(Rule from deleted filter)";
	}
	NSString* prefix = [NSString stringWithFormat:@"%@%@ \u27EB ",
						(isPreview ? @"[preview] " : @""), [filter name]];
	int ruleId = [infos[2] intValue];
	if (ruleId <= 0) {
		if (!fallback)
			return NULL;
		return [prefix stringByAppendingString:@"(new unsaved rule)"];
	}
	NSXMLElement* ruleE = [filter ruleWithId:ruleId];
	if (!ruleE) {
		if (!fallback)
			return NULL;
		return [prefix stringByAppendingString:@"(unsaved or deleted rule)"];
	}
	NSString* comments = [Xml stringForXPath:@"comments" withRoot:ruleE];
	if (![comments length])
		return prefix;
	NSRange range = [comments rangeOfCharacterFromSet:[NSCharacterSet newlineCharacterSet]];
	if (range.location != NSNotFound)
		comments = [comments substringToIndex:range.location];
	return [prefix stringByAppendingString:comments];
}

#pragma mark ------------------------ SplitView Delegate

- (BOOL)splitView:(NSSplitView *)sender canCollapseSubview:(NSView *)subview
{
	return NO;
}

- (CGFloat)splitView:(NSSplitView *)sender constrainMaxCoordinate:(CGFloat)proposedMax ofSubviewAt:(NSInteger)offset
{
	//DebugNSLog(@"constrainMaxCoordinate: proposedMax = %f,  ofSubviewAt = %d", proposedMax, offset);
	return [sender frame].size.height - 57;
}

- (CGFloat)splitView:(NSSplitView *)sender constrainMinCoordinate:(CGFloat)proposedMin ofSubviewAt:(NSInteger)offset
{
	//DebugNSLog(@"constrainMinCoordinate: proposedMax = %f,  ofSubviewAt = %d", proposedMin, offset);
	return 100;
}

/*
- (CGFloat)splitView:(NSSplitView *)sender constrainSplitPosition:(CGFloat)proposedPosition ofSubviewAt:(NSInteger)offset
{
	NSLog(@"constrainSplitPosition: proposedPosition = %f,  ofSubviewAt = %d", proposedPosition, offset);
	return proposedPosition;
}
*/

@end

@implementation HistoryController(NSTableDataSource)

- (NSInteger)numberOfRowsInTableView:(NSTableView *)tableView
{
	if (tableView == _historyTableView)
		return [_entryList count];
	if (tableView == _rulesTableView)
		return [[self getActionListForSelectedHistoryEntry] count];
	return 0;
}

- (id)tableView:(NSTableView *)tableView objectValueForTableColumn:(NSTableColumn *)col row:(NSInteger)row
{
	if (tableView == _historyTableView)
		return [self objectValueForHistoryTable:row];
	if (tableView == _rulesTableView) {
		NSString* s = [self ruleDescription:row withFallbackMessage:YES];
		return [NSString stringWithFormat:@"#%ld: %@", row + 1, s];
	}
	return @"???";
}

@end

@implementation HistoryTableView

- (IBAction)copy:(id)sender
{
	id del = [self delegate];
	if ([del isKindOfClass:[HistoryController class]]) {
		HistoryController* hc = (HistoryController*)del;
		[hc copyFromHistoryTable];
	}
}

@end
