/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "SuspectsController.h"
#import "GBDebug.h"
#import "XmlDoc.h"
#import "Xml.h"
#import "ServerCommunication.h"
#import "AlertHelper.h"
#import "Prefs.h"
#import "FilterTabController.h"
#import "FilterRuleEdit.h"
#import "AutoZap.h"

@interface SuspectsController ()

@property FilterTabController* filterTabController;
@property Prefs* prefs;
@property AlertHelper* alertHelper;
@property ServerCommunication* comm;
@property NSBundle* bundle;
@property BOOL visible;
@property BOOL enabled;
@property BOOL showDetails;
@property BOOL showJavascripts;
@property NSTimer* timer;
@property NSMutableArray* suspectList;
@property NSArray *topLevelObjects;

@property (weak) IBOutlet NSView*			subView;

@property (weak) IBOutlet NSButton*			enableCB;
@property (weak) IBOutlet NSTableView*		suspectTableView;
@property (weak) IBOutlet NSButton*			addBtn;
@property (weak) IBOutlet NSButton*			showDetailsCB;
@property (weak) IBOutlet NSButton*			showJavascriptsCB;


- (void)timerCallback:(NSTimer*)theTimer;
- (void)updateTableLayout;

- (void)tableViewSelectionDidChange:(NSNotification *)aNotification;
@end

@implementation SuspectsController

- (instancetype)initWithFilterTabController:(FilterTabController*)filterTabController
				withContainerView:(NSView*)containerView
					   withBundle:(NSBundle*)bundle
{
	if (!(self = [super init]))
		return nil;
	_topLevelObjects = [AutoZap loadNibNamed:@"Suspects" bundle:[filterTabController mainBundle] owner:self];
	NSAssert(_suspectTableView, @"suspectTableView is null");
	_filterTabController = filterTabController;
	_comm = [filterTabController comm];
	_prefs = [_comm prefs];
	_alertHelper = [_prefs alertHelper];
	_bundle = [filterTabController bundle];
	_suspectList = [NSMutableArray arrayWithCapacity:200];
	_showDetails = [_prefs boolForXPath:@"suspects/@show-details"];
	_showJavascripts = [_prefs boolForXPath:@"suspects/@show-javascripts"];
	_enabled = [_prefs boolForXPath:@"suspects/@enabled"];
	[_showDetailsCB setState:_showDetails];
	[_showJavascriptsCB setState:_showJavascripts];
	[_enableCB setState:_enabled];
	[containerView addSubview:_subView];
	[_suspectTableView setDataSource:self];
	[_suspectTableView setDelegate:self];
	[self updateTableLayout];
	[self tableViewSelectionDidChange:NULL];
	[_suspectTableView setTarget:self];
	[_suspectTableView setDoubleAction:@selector(doubleClickInTable:)];
	[self notifyUpdatedAuthorizationState];
	[_suspectTableView registerForDraggedTypes:@[NSStringPboardType, NSURLPboardType]];
	[_suspectTableView setDraggingSourceOperationMask:NSDragOperationCopy forLocal:NO];
	[_suspectTableView setDraggingSourceOperationMask:NSDragOperationCopy forLocal:YES];
	return self;
}

- (void)dealloc
{
	DebugNSLog(@"Suspects.dealloc");
	[self notifyBecameHidden];
	[_subView removeFromSuperview];
}

- (void)startTimerWithImmediateExecute:(BOOL)immediate
{
	if (!_enabled)
		return;
	if (!immediate && [_timer isValid])
		return;
	[self stopTimer];
	if (immediate)
		[self timerCallback:NULL];
	_timer = [NSTimer scheduledTimerWithTimeInterval:2.0
											   target:self
											 selector:@selector(timerCallback:)
											 userInfo:NULL
											  repeats:NO];
}

- (void)stopTimer
{
	if (!_timer)
		return;
	[_timer invalidate];
	_timer = NULL;
}

- (void)notifyBecameVisible
{
	_visible = YES;
	[self startTimerWithImmediateExecute:YES];
}

- (void)notifyBecameHidden
{
	_visible = NO;
	[self stopTimer];
}

- (void)notifyUpdatedAuthorizationState
{
	[_enableCB setEnabled:[_prefs hasGlimmerAuth]];
	[self tableViewSelectionDidChange:NULL];
}

- (void)timerCallback:(NSTimer*)theTimer
{
	[self stopTimer];
	if (!_enabled) {
		[_suspectList removeAllObjects];
		[_suspectTableView reloadData];
		return;
	}
	if (!_visible) {
		DebugNSLog(@"Suspects, Timer callback without being visible");
		return;
	}
	NSString* pathArgs = [NSString stringWithFormat:@"/suspect/list?js=%@", _showJavascripts ? @"true" : @"false"];
	NSError*__autoreleasing error = NULL;
	XmlDoc* doc = [_comm curlToXml:pathArgs
			 withActionDescription:@"get list of suspects"
					  withErrorRef:&error];
	[self startTimerWithImmediateExecute:NO];
	if (!doc)
		return;
	NSArray* a = [[doc rootE] elementsForName:@"suspect"];
	NSUInteger num = [a count];
	if (!num) {
		[_suspectList removeAllObjects];
		[_suspectTableView reloadData];
		[self tableViewSelectionDidChange:NULL];
		return;
	}
	NSString* selectedHost = NULL;
	NSString* selectedPath = NULL;
	NSInteger row = [_suspectTableView selectedRow];
	if (row >= 0) {
		NSXMLElement* e = _suspectList[row];
		selectedHost = [Xml getAttribute:e withName:@"host"];
		selectedPath = [Xml getAttribute:e withName:@"path"];
	}
	[_suspectList setArray:a];
	[_suspectTableView reloadData];
	if (selectedHost) {
		BOOL found = NO;
		for (NSUInteger i = 0; i < num; i++) {
			NSXMLElement* e = a[i];
			if ([selectedHost isEqual:[Xml getAttribute:e withName:@"host"]]
				&& [selectedPath isEqual:[Xml getAttribute:e withName:@"path"]]) {
				NSIndexSet* set = [NSIndexSet indexSetWithIndex:i];
				[_suspectTableView selectRowIndexes:set byExtendingSelection:NO];
				//DebugNSLog(@"Suspects.timer.selected-row: %d -> %d", row, i);
				found = YES;
				break;
			}
		}
		if (!found) {
			[_suspectTableView deselectAll:self];
		}
	}
}

#pragma mark ------------------- IB

- (void)updateTableLayout
{
	[_suspectTableView setRowHeight:_showDetails ? 44 : 17];
}

- (void)tableViewSelectionDidChange:(NSNotification *)aNotification
{
	[_addBtn setEnabled:[_prefs hasGlimmerAuth] && [_suspectTableView selectedRow] >= 0];
}


- (IBAction)showDetailsAction:(id)sender
{
	_showDetails = [_showDetailsCB state];
	[_prefs setBool:_showDetails forXPath:@"suspects/@show-details"];
	[self updateTableLayout];
	[_suspectTableView reloadData];
	if ([_prefs hasGlimmerAuth])
		[_prefs markAsDirty];
}

- (IBAction)showJavascriptsAction:(id)sender;
{
	_showJavascripts = [_showJavascriptsCB state];
	[_prefs setBool:_showJavascripts forXPath:@"suspects/@show-javascripts"];
	if ([_prefs hasGlimmerAuth])
		[_prefs markAsDirty];
	[self startTimerWithImmediateExecute:YES];
}

- (IBAction)enableAction:(id)sender
{
	_enabled = [_enableCB state];
	DebugNSLog(@"enableAction: %d", _enabled);
	[_prefs setBool:_enabled forXPath:@"suspects/@enabled"];
	[_prefs markAsDirty];
	[_comm saveDirtyPrefsAndNotify];
	if (_enabled) {
		[self startTimerWithImmediateExecute:YES];
	} else {
		[self stopTimer];
		[_suspectList removeAllObjects];
		[_suspectTableView reloadData];
	}
}

- (void)openFilterSheet:(NSInteger)row
{
	if (row < 0 || row >= (int)[_suspectList count])
		return;
	NSXMLElement* suspectE = _suspectList[row];
	FilterRuleEdit *editor = [[FilterRuleEdit alloc] initWithFilterTabController:_filterTabController
																 withOwnerFilter:NULL
																 withRuleElement:NULL
															  withSuspectElement:suspectE];
	[AutoZap add:editor];
}

- (IBAction)addAction:(id)sender
{
	[self openFilterSheet:[_suspectTableView selectedRow]];
}

- (IBAction)doubleClickInTable:(id)sender
{
	[self openFilterSheet:[_suspectTableView clickedRow]];
}


- (NSString *)tableView:(NSTableView *)aTableView
		 toolTipForCell:(NSCell *)aCell
				   rect:(NSRectPointer)rect
			tableColumn:(NSTableColumn *)aTableColumn
					row:(NSInteger)row
		  mouseLocation:(NSPoint)mouseLocation
{
	if (aTableView != _suspectTableView || _showDetails)
		return @"";
	NSXMLElement* e = _suspectList[row];
	return [NSString stringWithFormat:@"%@\nFrom: %@",
			[Xml getAttribute:e withName:@"why"],
			[Xml getAttribute:e withName:@"referer"]];
}

- (NSString*)urlForRow:(NSUInteger)row
{
	NSXMLElement* e = _suspectList[row];
	NSString* host = [Xml getAttribute:e withName:@"host"];
	NSString* path = [Xml getAttribute:e withName:@"path"];
	NSString* query = [Xml getAttribute:e withName:@"query"];
	NSString* url = [host stringByAppendingString:path];
	if ([query length])
		url = [[url stringByAppendingString:@"?"] stringByAppendingString:query];
	return url;
}

- (BOOL)tableView:(NSTableView *)view writeRowsWithIndexes:(NSIndexSet *)rowIndexes toPasteboard:(NSPasteboard *)pboard
{
	DebugNSLog(@"writeRows.%d: %@", (view == _suspectTableView), rowIndexes);
	if (view != _suspectTableView || [rowIndexes count] != 1)
		return NO;
	NSString* u = [@"http://" stringByAppendingString:[self urlForRow:[rowIndexes firstIndex]]];
	[pboard declareTypes:@[NSStringPboardType, NSURLPboardType] owner:nil];
	[pboard setString:u forType:NSStringPboardType];
	@try {
		[[NSURL URLWithString:u] writeToPasteboard:pboard];
	}
	@catch (NSException* ex) {
		NSLog(@"Couldn't copy URL to pasteboard: %@", u);
	}
	return YES;
}

- (void)copyFromSuspectsTable
{
	NSInteger row = [_suspectTableView selectedRow];
	if (row < 0 || row >= (NSInteger)[_suspectList count])
		return;
	NSIndexSet *set = [NSIndexSet indexSetWithIndex:row];
	[self tableView:_suspectTableView
		  writeRowsWithIndexes:set
	   toPasteboard:[NSPasteboard generalPasteboard]];
}

@end

@implementation SuspectsController(NSTableDataSource)

- (NSInteger)numberOfRowsInTableView:(NSTableView *)tableView
{
	NSAssert(tableView == _suspectTableView, @"tableView == suspectTableView");
	return [_suspectList count];
}

- (id)tableView:(NSTableView *)tableView objectValueForTableColumn:(NSTableColumn *)col row:(NSInteger)row
{
	NSAssert(tableView == _suspectTableView, @"tableView == suspectTableView");
	if (!_showDetails)
		return [self urlForRow:row];
	NSXMLElement* e = _suspectList[row];
	return [NSString stringWithFormat:@"%@\n       from: %@\n       %@",
			[self urlForRow:row],
			[Xml getAttribute:e withName:@"referer"],
			[Xml getAttribute:e withName:@"why"]];
}

@end

@implementation SuspectsTableView

- (IBAction)copy:(id)sender
{
	id del = [self delegate];
	if ([del isKindOfClass:[SuspectsController class]]) {
		SuspectsController* sc = (SuspectsController*)del;
		[sc copyFromSuspectsTable];
	}
}

@end
