/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "CodeEditHelper.h"
#import "JavascriptSyntaxHighlighter.h"

@interface CodeEditHelper()

@property id<CodeEditDelegate> delegate;
@property BOOL isFormatting;

@end

@implementation CodeEditHelper

static inline NSString* soft(NSString* s)
{
	return s ? s : @"";
}

- (instancetype)initWithDelegate:(id<CodeEditDelegate>)delegate
{
	if (!(self = [super init]))
		return nil;
	_delegate = delegate;
	return self;
}


- (void)syntaxHighlight:(NSMutableAttributedString*)attributedString
		   isJavascript:(BOOL)isJavascript
{
	//DebugNSLog(@"syntaxHighlight%@: %@", (js ? @"Javascript" : @"JavaRegExp"), [textStorage string]);
	//
	// This method is called when changing colors,
	// so we must guard against recursive calls of textStorageDidProcessEditing
	_isFormatting = YES;
	if (isJavascript)
		[JavascriptSyntaxHighlighter highlightJavascript:attributedString];
	else
		[JavascriptSyntaxHighlighter highlightJavaRegExp:attributedString];
	_isFormatting = NO;
}

- (void)syntaxHighlightJavascript:(NSTextStorage*)textStorage
{
	[self syntaxHighlight:textStorage isJavascript:YES];
}

- (void)delayedSyntaxHighlightJavascript:(NSTextStorage*)textStorage
{
	[self performSelectorOnMainThread:@selector(syntaxHighlightJavascript:)
						   withObject:textStorage
						waitUntilDone:NO];
}

- (void)setCodeTabs:(NSTextStorage*)storage
{
	static NSMutableParagraphStyle *style = NULL;
	if (!style) {
		style = [[NSMutableParagraphStyle alloc] init];
		[style setParagraphStyle:[NSParagraphStyle defaultParagraphStyle]];
		[style setTabStops:@[]]; // delete default tabs
		for (int x = 15; x < 500; x += 15) {
			NSTextTab* tab = [[NSTextTab alloc]
							  initWithType:NSLeftTabStopType
							  location:x];
			[style addTabStop:tab];
		}
	}
	[storage addAttribute:NSParagraphStyleAttributeName
					value:style
					range:NSMakeRange(0, [storage length])];
}

- (void)setCodeTextView:(NSTextView*)textView
			   withText:(NSString*)txt
{
	static NSFont* font = NULL; // NSTextView forgets the font when using setString, so force it.
	if (!font) {
		//font = [[NSFont fontWithName:@"Monaco" size:10] retain];
		font = [NSFont systemFontOfSize:11];
	}
	[textView setString:soft(txt)];
	[textView setFont:font];
	[textView setSmartInsertDeleteEnabled:NO];
	[textView setAutomaticQuoteSubstitutionEnabled:NO];
	[textView setAutomaticSpellingCorrectionEnabled:NO];
	[textView setAllowsImageEditing:NO];
	[self setCodeTabs:[textView textStorage]];
}

- (void)textStorageDidProcessEditing:(NSNotification *)noti
{
	if (_isFormatting || ![[noti object] isKindOfClass:[NSTextStorage class]])
		return;
	NSTextStorage* ts = (NSTextStorage*)[noti object];
	if ([_delegate isJavascriptTextStorage:ts])
		[self syntaxHighlight:ts isJavascript:YES];
	[self setCodeTabs:ts];
}

// returns NULL if results in empty string.
- (NSString*)trimmedTextViewString:(NSTextView*)textView
{
	NSString* txt = [CodeEditHelper trimmedString:[textView string]];
	return [txt length] ? txt : NULL;
}

// returns NULL if results in empty string.
+ (NSString*)trimmedString:(NSString*)src
{
	NSCharacterSet* wsp = [NSCharacterSet whitespaceAndNewlineCharacterSet];
	NSInteger len = [src length];
	NSInteger end = len - 1;
	while (end >= 0 && [wsp characterIsMember:[src characterAtIndex:end]])
		end--;
	NSInteger start = 0;
	while (start <= end && [wsp characterIsMember:[src characterAtIndex:start]])
		start++;
	if (start > end)
		return NULL;
	if (start == 0 && end == len - 1)
		return src;
	return [src substringWithRange:NSMakeRange(start, end - start + 1)];
}

@end
