/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals.log;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.util.ResourceUtils;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.misc.ByteBuffer;

public class RollingLogfile extends DatePrefixedLogger
{
	private static final File logDirF = new File("/Library/Logs/GlimmerBlocker");
	private static final File logF = new File(logDirF, "GlimmerBlocker.log");
	private static final File launchdLogF = new File(logDirF, "Launchd.log");

	private static final long mb = 1024 * 1024;

	private static final long archiveTrimSize = 25 * mb;

	private static final long normalModeLogfileRollSize = 2 * mb;
	private static final long normalModeMinLifetime = 5 * 60 * 1000; // don't roll logfiles faster than this.

	// When debug is enabled, don't roll as often because this can easily make the interesting parts
	// be hidden in a rolled file.
	private static final long debugModeLogfileRollSize = 20 * mb;
	private static final long debugModeMinLifetime = 60 * 60 * 1000;

	private static final String dashes =
			"===============================================================================================";
	private static LogStream logStream;

	public RollingLogfile(GlimmerGlobals globals) throws IOException
	{
		if (!logDirF.exists()) {
			System.err.println("GlimmerBlocker is missing log file directory: " + logDirF.getPath());
			System.err.println("PLEASE REINSTALL GLIMMERBLOCKER!!!");
			return;
		}
		if (logStream != null) {
			System.err.println("Has already created RollingLogFile instance.");
			return;
		}
		logStream = new LogStream(this, globals);
		setLogfilePermissions(logF);
		setLogfilePermissions(launchdLogF);
	}

	static class LogStream extends OutputStream implements Logger
	{
		@NotNull private final GlimmerGlobals globals;
		private final Logger logger;
		private final ByteBuffer bb = new ByteBuffer(200);
		private FileOutputStream fos;
		private long currentLogsize;
		private long lastRolltime;
		private boolean isRollingLogfile;

		LogStream(@NotNull Logger parentLogger, @NotNull GlimmerGlobals globals) throws IOException
		{
			this.globals = globals;
			if (launchdLogF.exists() && launchdLogF.length() == 0) {
				System.err.println("Only problems starting the proxy server is written to this file,");
				System.err.println("so it should usually be empty.");
				System.err.println();
				System.err.println("The proxy server logs to " + logF.getName());
				System.err.flush(); // shouldn't be needed, but anyway..
			}
			fos = new FileOutputStream(logF, true);
			currentLogsize = logF.length();
			logger = new PrefixedLogger(parentLogger, "LogFile");
			PrintStream pw = new PrintStream(this, true, "UTF-8");
			System.setErr(pw);
			System.setOut(pw);
			System.err.println(dashes);
		}

		@Override
		public synchronized void write(byte[] bytes, int offset, int len) throws IOException
		{
			bb.append(bytes, offset, len);
			if (len > 0 && bytes[offset + len - 1] == '\n')
				flush();
		}

		public synchronized void write(int ch) throws IOException
		{
			bb.append(ch);
			if (ch == '\n')
				flush();
		}

		@Override
		public synchronized void flush() throws IOException
		{
			if (bb.length() == 0)
				return;
			bb.writeTo(fos);
			currentLogsize += bb.length();
			bb.clear();
			long minLifeTime = globals.debugFlag ? debugModeMinLifetime : normalModeMinLifetime;
			long minRollSize = globals.debugFlag ? debugModeLogfileRollSize : normalModeLogfileRollSize;
			long now = System.currentTimeMillis();
			if (currentLogsize > minRollSize && now > lastRolltime + minLifeTime)
				rollLogfile();
		}

		private synchronized void rollLogfile() throws IOException
		{
			if (isRollingLogfile)
				return;
			try {
				isRollingLogfile = true;
				doRollLogfile();
			}
			finally {
				isRollingLogfile = false;
			}
		}

		private synchronized void doRollLogfile() throws IOException
		{
			currentLogsize = 0;
			lastRolltime = System.currentTimeMillis();
			SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd__HH-mm");
			sdf.setTimeZone(TimeZone.getDefault());
			String date = sdf.format(new Date());
			//
			File tmpF = new File(logDirF, "new-" + date + ".txt");
			File archivedF = new File(logDirF, "old-" + date + ".txt");
			if (archivedF.exists()) {
				ln("Would have rolled logfile, but archive file already exists: " + archivedF.getName());
				return;
			}
			FileOutputStream fos2 = null;
			try {
				fos2 = new FileOutputStream(tmpF);
				setLogfilePermissions(tmpF);
				//
				if (!logF.renameTo(archivedF))
					throw new IOException("Can't rename logfile to " + archivedF);
				//
				if (!tmpF.renameTo(logF))
					throw new IOException("Can't rename tmp logfile to " + logF.getName());
				// It seems like Console.app doesn't detect the renaming, so let it known about 'a deletion'.
				try {
					new FileOutputStream(tmpF).close();
					if (!tmpF.delete())
						ln("Could delete dummy file: " + tmpF.getName());
				}
				catch (Exception ex) {
					ln("Had problem fooling Console.app: " + ExceptionUtils.getExMsg(ex));
				}
				//
				// Might still have buffered stuff in System.err, so use direct write() to get around buffering.
				String s = "\n\n\n" +
						dashes + "\n" +
						DatePrefixedLogger.getDatePrefix() + " [LogFile] " +
						"Rolls logfile. This file will be archived as " + archivedF.getName() + "\n" +
						dashes + "\n";
				fos.write(s.getBytes("UTF-8"));
				//
				fos.close();
				fos = fos2;
				fos2 = null;
				setLogfilePermissions(logF);
				//
				s = dashes + "\n" +
						DatePrefixedLogger.getDatePrefix() + " [LogFile] " +
						"Rolled logfile.  Previous log is archived in " + archivedF.getName() + "\n" +
						dashes + "\n";
				fos.write(s.getBytes("UTF-8"));
			}
			catch (Exception ex) {
				System.err.println("Can't roll logfile: " + ExceptionUtils.getExMsg(ex));
				if (fos2 != null)
					fos2.close();
				if (tmpF != null && tmpF.exists() && !tmpF.delete())
					System.err.println("Can't delete tmp file: " + tmpF);
			}
			zapOldArchivedLogfiles();
		}

		private void zapOldArchivedLogfiles()
		{
			try {
				File[] files = logDirF.listFiles();
				if (files == null) // dir missing!
					return;
				List<File> candidates = new ArrayList<>();
				for (File f : files) {
					String fn = f.getName();
					if (fn.startsWith("old-") && fn.endsWith(".txt"))
						candidates.add(f);
				}
				if (candidates.isEmpty())
					return;
				Collections.sort(candidates);
				Collections.reverse(candidates);
				candidates.remove(0); // don't zap the just-archived file.
				long sumSize = 0;
				boolean didDeleteFile = false;
				for (File f : candidates) {
					if (sumSize > archiveTrimSize) {
						if (!f.delete())
							throw new IOException("Can't delete old logfile: " + f.getName());
						didDeleteFile = true;
						ln("Deleted old logfile: " + f.getName());
					} else {
						// don't let a single huge log file get all the older files to
						// be zapped. So use an upper bound.
						sumSize += Math.min(f.length(), 2 * normalModeLogfileRollSize);
						//ln("Keeps: " + f.getName());
					}
				}
				if (didDeleteFile)
					ln("Total archive size is now: " + (sumSize / mb) + " MB.");
			}
			catch (Exception ex) {
				ln("Got exception zapping old log files: " + ExceptionUtils.getExMsg(ex));
				ln(ExceptionUtils.getStackTraceAsString(ex));
			}
		}

		public void ln(@Nullable String s)
		{
			logger.ln(s);
		}
	}

	private static void setLogfilePermissions(File f)
	{
		ResourceUtils.setPermissions(f, "644");
	}
}
