/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals.network;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.InetAddress;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.globals.network.dns.DnsResolverConfig;
import org.glimmerblocker.proxy.rhino.runtime.ExecutorDelegate;
import org.glimmerblocker.proxy.rhino.runtime.RhinoRuntimeCache;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.RhinoException;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.io.FileContentsUtils;
import dk.vitality.lib.foundation.io.SimpleUnixExec;
import dk.vitality.lib.foundation.misc.StringUtils;
import dk.vitality.lib.foundation.ref.FatRef;
import dk.vitality.lib.foundation.xml.xpath.XmlArgs;

public class NetworkManager implements Logger
{
	@NotNull private NetworkConfig config = new NetworkConfig();

	@NotNull
	public NetworkConfig getConfig()
	{
		return config;
	}

	public synchronized void notifyPrefsChanged() throws Exception
	{
		XmlArgs settingsXA = globals.panelSettingsXA;
		if (settingsXA == null)
			return; // happends at launch.
		String s = settingsXA.getOptNonNullString("network-settings/@site-proxy", "none");
		if (s.equals("static")) {
			setStaticProxy(settingsXA);
		} else if (s.equals("javascript")) {
			runProxyJavascript();
		} else {
			setSimpleProxyConfig(null);
		}
		globals.httpClientCache.purgeAllCurrentConnections();
	}

	public synchronized void validateProxyJavascript(@NotNull String js) throws Exception
	{
		NetworkConfig oldConfig, newConfig;
		synchronized (this) {
			oldConfig = config;
			newConfig = runConfigScript(js);
		}
		if (globals.debugNetworkConfig.get() && !newConfig.equals(oldConfig)) {
			ln("Config script diff:");
			newConfig.logChanges(this, oldConfig);
		}
	}

	//===================================================================

	public NetworkManager(GlimmerGlobals globals)
	{
		this.globals = globals;
		this.logger = new PrefixedLogger(globals, "NetworkConfig");
		Thread t = new Thread(new Listener(), "NetworkConfigurationListener");
		t.start();
	}

	private GlimmerGlobals globals;
	private Logger logger;

	public void ln(@Nullable String s)
	{
		logger.ln(s);
	}

	private void setStaticProxy(@NotNull XmlArgs settingsXA) throws Exception
	{
		String hostname = settingsXA.getReqString("network-settings/@site-proxy-host");
		int port = settingsXA.getReqInt("network-settings/@site-proxy-port");
		if (port <= 0 || port > 65535)
			port = 80;
		try {
			//noinspection ResultOfMethodCallIgnored
			InetAddress.getByName(hostname); // just validate hostname using native dns resolver.
			HostnamePort hp = new HostnamePort(hostname, port);
			setSimpleProxyConfig(hp);
		}
		catch (Exception ex) {
			ln("Can't resolve site-proxy: '" + hostname + "'.  Ignores http-proxy setting.");
			ln("  Error: " + ExceptionUtils.getExMsg(ex));
			setSimpleProxyConfig(null);
		}
	}

	private synchronized void setInterfacesJSON(@NotNull String s) throws Exception
	{
		if (s.equals(config.interfacesJSON)) {
			if (globals.debugNetworkConfig.get() && globals.debugVerbose)
				ln("Got network settings JSON, but is unchanged.");
		} else {
			if (globals.debugNetworkConfig.get()) {
				if (config.hasFullInterfacesJSON())
					ln("Network state changed.");
				else
					ln("Got initial network JSON.");
			}
			if (globals.debugNetworkConfig.get() && globals.debugVerbose)
				ln(s);
			if (globals.debugNetworkConfig.get() && config.hasFullInterfacesJSON())
				dumpDiff(config.interfacesJSON, s);
			setNewConfig(config.withInterfacesJSON(s));
		}
		runProxyJavascript();
	}

	private synchronized void setLocationName(@NotNull String newName) throws Exception
	{
		if (StringUtils.triEquals(config.locationName, newName))
			return;
		setNewConfig(config.withLocationName(newName));
		runProxyJavascript();
	}

	private synchronized void setSimpleProxyConfig(@Nullable HostnamePort hp)
	{
		NetworkConfig cfg = config.withHttpProxy(hp).withSocksProxy(null);
		cfg = cfg.withDnsConfig(DnsResolverConfig.NATIVE_RESOLVER);
		setNewConfig(cfg);
	}

	/**
	 * MUST HAVE synchronized while creating derived config and while calling this method.
	 */
	private void setNewConfig(@NotNull NetworkConfig newConfig)
	{
		if (newConfig == config)
			return;
		newConfig.logChanges(this, config);
		config = newConfig;
		if (globals.httpClientCache != null) // can be null at startup time.
			globals.httpClientCache.purgeAllCurrentConnections();
	}

	class Listener implements Runnable
	{
		@Nullable Process p;

		@SuppressWarnings("InfiniteLoopStatement")
		public void run()
		{
			int num = 0;
			while (true) {
				try {
					if (num++ > 0) {
						Thread.sleep(1000);
						ln("Relaunches NetworkConfigurationListener.");
					}
					check();
				}
				catch (Exception ex) {
					ln("NetworkConfigurationListener got exception: " + ExceptionUtils.getExMsg(ex));
					if (globals.debugNetworkConfig.get() || ex instanceof NullPointerException)
						ln(ExceptionUtils.getStackTraceAsString(ex));
					if (p != null)
						p.destroy();
				}
			}
		}

		void check() throws Exception
		{
			String cmd = GlimmerGlobals.networkConfigListenerBinary.getPath();
			if (globals.debugNetworkConfig.get())
				ln("Launches2 " + cmd);
			p = Runtime.getRuntime().exec(new String[]{cmd, "-continous"});
			if (p == null)
				throw new NullPointerException("Failed launching " + cmd);
			OutputStream os = p.getOutputStream();
			if (os != null)
				os.close();
			BufferedReader r = new BufferedReader(new InputStreamReader(p.getInputStream(), "UTF-8"));
			while (true) {
				String s = r.readLine();
				if (s == null)
					break;
				String value = readJSON(r, s, "location:");
				if (value != null) {
					setLocationName(value);
					continue;
				}
				value = readJSON(r, s, "interfaces:");
				if (value != null) {
					setInterfacesJSON(value);
					continue;
				}
				if (s.startsWith("MSG:")) {
					ln(s.substring(4));
				} else if (s.startsWith("DEBUG:")) {
					if (globals.debugNetworkConfig.get())
						ln(s.substring(6));
				} else if (s.trim().length() > 0) {
					ln("UNHANDLED: " + s);
				}
			}
			r.close();
			p.destroy();
			p = null;
		}

		@Nullable
		String readJSON(@NotNull BufferedReader r,
		                @NotNull String json,
		                @NotNull String expectedPrefix) throws Exception
		{
			if (!json.startsWith(expectedPrefix))
				return null;
			json = json.substring(expectedPrefix.length()).trim();
			if (!json.startsWith("{")) {
				json = StringUtils.zapSuffix(json, ",");
				if (json.startsWith("'"))
					return parseJavascriptStringLiteral(json);
				else
					return json;
			}
			StringBuilder sb = new StringBuilder(100);
			sb.append(json).append('\n');
			while (true) {
				String s = r.readLine();
				if (s == null)
					break;
				sb.append(s).append('\n');
				if (s.startsWith("}"))
					break;
			}
			return sb.toString();
		}
	}

	private String parseJavascriptStringLiteral(String s) throws Exception
	{
		s = s.trim();
		int len = s.length();
		if (s.charAt(0) != '\'')
			throw new Exception("Invalid Javascript string: " + s);
		StringBuilder sb = new StringBuilder(len);
		for (int i = 1; i < len; i++) {
			char ch = s.charAt(i);
			if (ch == '\'')
				return sb.toString();
			if (ch != '\\') {
				sb.append(ch);
				continue;
			}
			ch = s.charAt(++i);
			switch (ch) {
				case 'n':
					sb.append('\n');
					break;
				case 'r':
					sb.append('\r');
					break;
				case 't':
					sb.append('\t');
					break;
				case 'b':
					sb.append('\b');
					break;
				case 'f':
					sb.append('\f');
					break;
				case 'x':
					ch = (char)Integer.parseInt(s.substring(i + 1, i + 3));
					sb.append(ch);
					i += 2;
					break;
				case 'u':
					ch = (char)Integer.parseInt(s.substring(i + 1, i + 5));
					sb.append(ch);
					i += 4;
					break;
				default:
					sb.append(ch);
			}
		}
		throw new Exception("Unterminated javascript string literal: " + s);
	}

	private synchronized void runProxyJavascript()
	{
		try {
			XmlArgs settingsXA = globals.panelSettingsXA;
			String type = settingsXA.getOptNonNullString("network-settings/@site-proxy", "none");
			if (!type.equals("javascript"))
				return;
		}
		catch (Exception ex) {
			ln("Error determining if using javascript proxy: " + ExceptionUtils.getExMsg(ex));
			return;
		}
		try {
			String js = globals.panelSettingsXA.getOptNonNullString("network-settings/proxy-script", "");
			setNewConfig(runConfigScript(js));
		}
		catch (Exception ex) {
			ln("Failed executing site-wide proxy javascript: " + ExceptionUtils.getExMsg(ex));
			if (ex instanceof RhinoException) {
				RhinoException rex = (RhinoException)ex;
				if (rex.lineNumber() > 0)
					ln("  line " + rex.lineNumber() + ", col " + rex.columnNumber());
			}
			if (globals.debugNetworkConfig.get() || ex instanceof NullPointerException)
				ln(ExceptionUtils.getStackTraceAsString(ex));
		}
	}

	private synchronized void dumpDiff(@NotNull String a, @NotNull String b)
	{
		File tmpDirF = new File(System.getProperty("java.io.tmpdir"));
		File aF = new File(tmpDirF, "gb-network-a.js");
		File bF = new File(tmpDirF, "gb-network-b.js");
		try {
			FileContentsUtils.saveFile(aF, a.getBytes("UTF-8"));
			FileContentsUtils.saveFile(bF, b.getBytes("UTF-8"));
			SimpleUnixExec sue = new SimpleUnixExec("diff", "--unified", aF.getPath(), bF.getPath());
			sue.setWantsStdoutAsString(true);
			sue.setWantsStderrAsString(true);
			sue.run();
			// exit code is 0 if no diff.
			if (sue.getResultCode() != 0) {
				String s = sue.getStdoutAsString();
				int idx = s.indexOf("\n@@");
				if (idx > 0)
					s = s.substring(idx + 1); // zap the 2 filename header lines
				ln("state diff:\n" + s);
				ln("-----------------------------------------");
			} else {
				ln("No state diff.");
			}
		}
		catch (Exception ex) {
			ln("Got exception creating network state diff: " + ExceptionUtils.getExMsg(ex));
		}
		if (aF.exists() && !aF.delete())
			ln("Can't delete tmp file: " + aF);
		if (bF.exists() && !bF.delete())
			ln("Can't delete tmp file: " + bF);
	}

	private String previousJS;
	private int jsRuntimeId;

	@NotNull
	private synchronized NetworkConfig runConfigScript(@NotNull String js) throws Exception
	{
		if (globals.debugNetworkConfig.get() && globals.debugVerbose)
			ln("Executes site-wide proxy script.\n");
		if (!js.equals(previousJS)) {
			jsRuntimeId = RhinoRuntimeCache.makeNewRuntimeId();
			previousJS = js;
		}
		long st0 = System.currentTimeMillis();
		NetworkConfig plainConfig = config.withSocksProxy(null).withHttpProxy(null);
		plainConfig = plainConfig.withDnsConfig(DnsResolverConfig.NATIVE_RESOLVER);
		Delegate delegate = new Delegate(plainConfig);
		globals.rhinoRuntimeCache.execute(delegate, jsRuntimeId, js, "site-wide proxy script");
		if (globals.debugNetworkConfig.get() && globals.debugVerbose) {
			long dur = System.currentTimeMillis() - st0;
			ln("Proxy javascript exec time: " + dur + " msecs.");
		}
		return delegate.networkConfigRef.get();
	}

	class Delegate implements ExecutorDelegate<NetworkConfigGlobalObject>
	{
		@NotNull FatRef<NetworkConfig> networkConfigRef;

		Delegate(@NotNull NetworkConfig config)
		{
			this.networkConfigRef = new FatRef<>(config);
		}

		@NotNull
		public NetworkConfigGlobalObject create(@NotNull ScriptableObject scope,
		                                        @NotNull Scriptable defaultPrototype) throws Exception
		{
			return new NetworkConfigGlobalObject(globals, scope, defaultPrototype);
		}

		public void prepare(@NotNull NetworkConfigGlobalObject callback,
		                    @NotNull Context ctx) throws Exception
		{
			callback.prepareNetworkConfigCallback(globals, ctx, networkConfigRef);
		}

		public void executionDone(@NotNull NetworkConfigGlobalObject callback) throws Exception
		{
		}
	}
}
