/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals.subscriptions;

import java.io.ByteArrayInputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.TimeZone;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.http.data.HeadersBuilder;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.util.DateFormat;
import org.glimmerblocker.proxy.util.XmlParserHelper;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.xml.xpath.XmlArgs;

public class FilterNetworkData implements Logger
{
	@NotNull public GlimmerGlobals globals;
	public final int filterId;
	@Nullable public final URL url;
	@Nullable public final String errorMessage;
	@Nullable public final XmlArgs docXA;
	public final boolean hasNewData;
	public final long lastModified;

	public FilterNetworkData(@NotNull GlimmerGlobals globals,
	                         @NotNull Logger parentLogger,
	                         int filterId,
	                         @NotNull String url,
	                         @Nullable XmlArgs currentXA,
	                         boolean allowUpdate) throws Exception
	{
		this.globals = globals;
		this.prefixedLogger = new PrefixedLogger(parentLogger, Integer.toString(++prevId));
		this.filterId = filterId;
		Element filterContainerE = (currentXA != null) ? currentXA.getContextElement() : null;
		setAttr(filterContainerE, "network-unknown-host", null);
		URL u;
		try {
			u = new URL(url);
		}
		catch (Exception ex) {
			errorMessage = "Invalid URL: " + ExceptionUtils.getExMsg(ex);
			setAttr(filterContainerE, "network-error", errorMessage);
			docXA = null;
			hasNewData = false;
			lastModified = 0;
			this.url = null;
			return;
		}
		this.url = u;
		ResultData rd;
		try {
			HeadersBuilder hb = HeadersBuilder.makeEmpty();
			hb.add("user-agent", globals.serverUserAgent);
			if (currentXA != null) {
				String s = currentXA.getOptString("last-modified/@http-raw"); // avoid timezone conversions.
				if (s != null) {
					hb.add("if-modified-since", s);
				} else {
					long md = currentXA.getOptLong("last-modified/@msecs", 0);
					if (md > 0)
						hb.add("if-modified-since", DateFormat.to1123(md));
				}
			}
			rd = globals.curl("GET", u, hb.toHeaders().headerMap, this);
		}
		catch (Exception ex) {
			if (ExceptionUtils.hasCause(ex, UnknownHostException.class)) {
				setAttr(filterContainerE, "network-unknown-host", "1");
				errorMessage = "Unknown host: " + u.getHost();
			} else {
				errorMessage = ExceptionUtils.getExMsg(ex);
			}
			if (globals.debugFlag) {
				ln("Got exception while downloading subscription: " + errorMessage);
				ln(ExceptionUtils.getStackTraceAsString(ex));
			}
			docXA = currentXA;
			hasNewData = false;
			lastModified = 0;
			setAttr(filterContainerE, "url", url);
			setAttr(filterContainerE, "network-error", errorMessage);
			updateLastCheckDate();
			return;
		}
		if (rd.headers.status == HttpURLConnection.HTTP_NOT_MODIFIED) {
			if (currentXA == null) {
				errorMessage = "Invalid 'not modified' response as request was sent without if-modified-since";
				ln(errorMessage);
			} else {
				errorMessage = null;
				if (globals.debugFilters)
					ln("Subscription not-modified");
			}
			docXA = currentXA;
			hasNewData = false;
			lastModified = 0;
			setAttr(filterContainerE, "url", url);
			setAttr(filterContainerE, "network-error", errorMessage);
			setDate("available-update-last-modified", 0);
			updateLastCheckDate();
			return;
		}
		if (rd.headers.status != HttpURLConnection.HTTP_OK) {
			if (rd.headers.status == HttpURLConnection.HTTP_NOT_FOUND)
				errorMessage = "Subscription file does not exist at server";
			else
				errorMessage = "Negative HTTP status returned: " + rd.headers.status + " " + rd.headers.statusMessage;
			ln("Failed updating subscription: " + errorMessage);
			docXA = currentXA;
			hasNewData = false;
			lastModified = 0;
			setAttr(filterContainerE, "url", url);
			setAttr(filterContainerE, "network-error", errorMessage);
			updateLastCheckDate();
			return;
		}
		lastModified = getLastModified(rd);
		if (!allowUpdate) {
			if (globals.debugFilters)
				ln("Has update, but doesn't modify content.");
			errorMessage = null;
			docXA = currentXA;
			hasNewData = true;
			setAttr(filterContainerE, "url", url);
			setAttr(filterContainerE, "network-error", null);
			setDate("available-update-last-modified", lastModified);
			updateLastCheckDate();
			return;
		}
		Document doc;
		try {
			String s = new String(rd.getContent(), "ISO-8859-1").trim(); // never fails
			if (!s.startsWith("<?xml"))
				throw new Exception("The server did not return a subscription file (not xml)");
			if (!s.contains("</filter-data>")) {
				if (s.length() > 250)
					s = s.substring(0, 250);
				System.err.println(s);
				throw new Exception("The server did not return a subscription file (missing </filter-data>)");
			}
			try {
				doc = XmlParserHelper.parse(new ByteArrayInputStream(rd.getContent()));
			}
			catch (Exception ex) {
				throw new Exception("The server returned invalid xml: " + ExceptionUtils.getExMsg(ex), ex);
			}
		}
		catch (Exception ex) {
			errorMessage = ExceptionUtils.getExMsg(ex);
			if (globals.debugFlag) {
				ln("Got exception parsing subscription XML: " + errorMessage);
				ln(ExceptionUtils.getStackTraceAsString(ex));
			}
			hasNewData = false;
			docXA = null;
			setAttr(filterContainerE, "url", url);
			setAttr(filterContainerE, "network-error", errorMessage);
			updateLastCheckDate();
			return;
		}
		try {
			Element filterDataE = doc.getDocumentElement();
			if (filterDataE == null)
				throw new Exception("Missing root node");
			String tag = filterDataE.getTagName();
			if (!tag.equals("filter-data"))
				throw new Exception("Invalid root tag = " + tag);
			int version = getInt(new XmlArgs(doc), "@format-version");
			if (version <= 0 || version > 99999)
				throw new Exception("Invalid format version = " + version);
			//
			//
			if (currentXA != null) {
				for (Element e : currentXA.getElementArray("filter-data"))
					e.getParentNode().removeChild(e);
				filterContainerE.appendChild(filterContainerE.getOwnerDocument().importNode(filterDataE, true));
			} else {
				filterContainerE = doc.createElement("filter-container");
				filterContainerE.setAttribute("filter-id", Integer.toString(filterId));
				filterContainerE.appendChild(filterDataE); // insert as root node
				doc.appendChild(filterContainerE);
				currentXA = XmlArgs.create(doc);
			}
			filterContainerE.setAttribute("url", url);
		}
		catch (Exception ex) {
			errorMessage = "Invalid subscription xml: " + ex.getMessage();
			if (globals.debugFlag) {
				ln("Got exception validating subscription XML: " + errorMessage);
				ln(ExceptionUtils.getStackTraceAsString(ex));
			}
			hasNewData = false;
			docXA = null;
			return;
		}
		docXA = currentXA;
		setDate(docXA, "last-modified", lastModified, rd.headers.getLC("last-modified"));
		setDate("available-update-last-modified", 0);
		updateLastCheckDate();
		errorMessage = null;
		hasNewData = true;
	}

	private void updateLastCheckDate() throws SAXException
	{
		if (docXA == null)
			return;
		long now = System.currentTimeMillis();
		setDate("last-update-check", now);
		if (!docXA.exists("subscribed-date"))
			setDate("subscribed-date", now);
		for (Element e : docXA.getElementArray("filter-data")) // make last element so all dates are at the top.
			docXA.getContextElement().appendChild(e);
	}

	private static void setAttr(Element e, String name, String value)
	{
		if (e == null)
			return;
		if (value != null)
			e.setAttribute(name, value);
		else
			e.removeAttribute(name);
	}

	private int getInt(XmlArgs xa, String xpath) throws Exception
	{
		String name = xpath.startsWith("@") ? xpath.substring(1) : xpath;
		try {
			if (!xa.exists(xpath))
				throw new Exception("Missing '" + name + "' attribute");
			return xa.getReqInt(xpath);
		}
		catch (Exception ex) {
			throw new Exception("Invalid '" + name + "' integer attribute value", ex);
		}
	}

	private void setDate(String name, long value) throws SAXException
	{
		setDate(docXA, name, value);
	}

	public static void setDate(XmlArgs docXA, String name, long value) throws SAXException
	{
		setDate(docXA, name, value, null);
	}

	public static void setDate(XmlArgs docXA, String name, long value, String httpRaw) throws SAXException
	{
		if (docXA == null)
			return; // happends when adding new subscription fails.
		Element setContainerE = docXA.getContextElement();
		setContainerE.removeAttribute("msecs"); // clean-up of obsolete
		setContainerE.removeAttribute("iso"); // clean-up of obsolete
		setContainerE.removeAttribute(name); // clean-up of obsolete
		setContainerE.removeAttribute(name + "-text"); // clean-up of obsolete
		Element e = docXA.getOptElement(name);
		if (value <= 0) {
			if (e != null)
				e.getParentNode().removeChild(e);
			return;
		}
		if (e == null) {
			e = docXA.getDocument().createElement(name);
			setContainerE.insertBefore(e, setContainerE.getFirstChild());
		}
		SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss Z");
		sdf.setTimeZone(TimeZone.getDefault());
		e.setAttribute("iso", sdf.format(new Date(value)));
		e.setAttribute("msecs", Long.toString(value));
		if (httpRaw != null)
			e.setAttribute("http-raw", httpRaw);
		else
			e.removeAttribute("http-raw");
	}

	@SuppressWarnings({"deprecation"})
	private long getLastModified(ResultData rd)
	{
		String s = rd.headers.getLC("last-modified");
		try {
			if (s == null) {
				ln("Missing Last-Modified in response from " + url);
				return 0;
			}
			return Date.parse(s);
		}
		catch (Exception ex) {
			ln("Can't parse last-modified: '" + s + "': " + ExceptionUtils.getExMsg(ex));
			return 0;
		}
	}

	public void ln(@Nullable String s)
	{
		if (!hasPrintedUrlBanner) {
			hasPrintedUrlBanner = true;
			prefixedLogger.ln("Subscription: " + url);
		}
		prefixedLogger.ln("  " + s);
	}

	@NotNull private Logger prefixedLogger;
	private boolean hasPrintedUrlBanner;
	private static int prevId = 0;
}
