/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals.subscriptions;

import java.io.File;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import org.glimmerblocker.proxy.PanelPreferencesUpgrader;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.util.ResourceUtils;
import org.glimmerblocker.proxy.util.XmlParserHelper;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.w3c.dom.Element;
import dk.vitality.lib.foundation.exception.UserErrorException;
import dk.vitality.lib.foundation.io.SimpleUnixExec;
import dk.vitality.lib.foundation.xml.xpath.XmlArgs;

public class FilterSubscriptions implements Logger
{
	public final GlimmerGlobals globals;

	private final Logger logger;
	private final Map<Integer, Object> filterLocks = new HashMap<>();

	private final File dirF = new File("/Library/GlimmerBlocker/Filter subscriptions");

	public FilterSubscriptions(@NotNull GlimmerGlobals globals)
	{
		this.globals = globals;
		this.logger = new PrefixedLogger(globals, "FilterSubs");
	}

	@Nullable
	public XmlArgs getOptFilterContainerXA(int filterId) throws Exception
	{
		File f = getFilterFile(filterId);
		if (!f.exists())
			return null;
		XmlArgs containerXA = new XmlArgs(XmlParserHelper.parse(f));
		if (containerXA.getOptInt("filter-data/@format-version", 0) < 3) {
			Element[] ruleElements = containerXA.getElementArray("filter-data/rule");
			PanelPreferencesUpgrader.upgradeRuleElements(ruleElements);
		}
		return containerXA;
	}

	public synchronized void removeSubscription(int filterId) throws Exception
	{
		File f = getFilterFile(filterId);
		if (f.exists() && !f.delete())
			throw new Exception("Can't delete filter data file: " + f.getPath());
	}

	@Nullable
	public XmlArgs updateSubscription(int filterId,
	                                  boolean forceUpdate,
	                                  long earliestAllowedUpdateTime) throws Exception
	{
		synchronized (getfilterLock(filterId)) {
			if (globals.debugFilters)
				ln("Checks for update to filter " + filterId);
			XmlArgs filterXA = globals.panelSettingsXA.getOptXmlArgs("filters/filter[@id='" + filterId + "']");
			if (filterXA == null)
				throw new Exception("No such subscription: " + filterId);
			String url = filterXA.getReqString("@url");
			String updatePolity = filterXA.getOptString("@update-policy", "manual");
			boolean allowUpdate = forceUpdate || updatePolity.equals("auto");
			XmlArgs containerXA = getOptFilterContainerXA(filterId);
			if (containerXA == null) {
				if (!allowUpdate) {
					if (!url.startsWith("http://glimmerblocker.org/") || !updatePolity.equals("notify"))
						return null;
					// Initial default subscriptions.
					// Allow update to avoid user having "subscription not downloaded" error messages
					// as it's confusing to have default subscriptions which are not downloaded.
					allowUpdate = true;
				}
			} else {
				if (!forceUpdate && !filterXA.getOptBoolean("@enabled")) {
					if (globals.debugFilters)
						ln("Skips updating subscription: not forced and filter is not enabled.");
					return containerXA;
				}
				if (containerXA.getOptBoolean("@network-unknown-host")) {
					long m = System.currentTimeMillis() - 3600 * 1000; // at least once an hour if dns problem.
					earliestAllowedUpdateTime = Math.max(earliestAllowedUpdateTime, m);
				}
				long lastUpdateCheck = containerXA.getOptLong("last-update-check/@msecs", 0);
				if (lastUpdateCheck > earliestAllowedUpdateTime) {
					if (globals.debugFilters) {
						long mins = (lastUpdateCheck - earliestAllowedUpdateTime) / 60 / 1000;
						ln("Skips updating subscription: too soon (" + mins + " minutes)" +
								" since last update check = " + new Date(lastUpdateCheck));
					}
					return containerXA;
				}
				if (globals.debugFilters) {
					ln("Checks subscription for update: last update check = " + new Date(lastUpdateCheck) +
							(earliestAllowedUpdateTime == Long.MAX_VALUE ? ", forced check now." : ""));
				}
			}
			FilterNetworkData data = new FilterNetworkData(globals, this, filterId, url, containerXA, allowUpdate);
			if (allowUpdate) {
				if (data.docXA != null) {
					if (globals.debugFilters)
						ln("Saves updated filter data (allowUpdate), hasNewData = " + data.hasNewData);
					saveFilterData(data);
				}
				return data.docXA;
			}
			if (!data.hasNewData) {
				if (globals.debugFilters)
					ln("Has no update for filter.");
				saveFilterData(data); // save updated check time.
				return data.docXA;
			}
			long ignored = containerXA.getOptLong("ignored-update/@msecs", 0);
			if (data.lastModified != 0 && data.lastModified == ignored) {
				if (globals.debugFlag) {
					ln("Skips updating subscription: has marked update as 'to be ignored'.");
					ln("   available last-modified: " + containerXA.getOptString("ignored-update/@iso"));
				}
				saveFilterData(data); // save updated check time.
				return data.docXA;
			}
			File f = getFilterFile(filterId);
			String[] cmd = {
					"/Library/PreferencePanes/GlimmerBlocker.prefPane/Contents/MacOS/AuthHelper",
					"--ask-filter-update",
					f.getPath()
			};
			if (globals.debugFilters)
				ln("Running " + Arrays.toString(cmd));
			SimpleUnixExec sue = new SimpleUnixExec(cmd);
			sue.run();
			int status = sue.getResultCode();
			if (globals.debugFilters)
				ln("Exit status: " + status);
			if (status != 0)
				return containerXA; // notification not possible.
			long lastModified = data.docXA.getOptLong("last-modified/@msecs", 0);
			if (lastModified <= 0) {
				ln("Missing last-modified header for ignored filter update.");
				return data.docXA;
			}
			if (globals.debugFilters)
				ln("Marks filter with ignored-update=" + lastModified);
			containerXA = getOptFilterContainerXA(filterId); // ensure fresh data if user modified it by other means.
			if (containerXA != null) {
				FilterNetworkData.setDate(containerXA, "ignored-update", lastModified);
				FilterNetworkData.setDate(containerXA, "available-update-last-modified", data.lastModified);
				saveFilterData(containerXA, f);
			}
			return containerXA;
		}
	}

	/**
	 * Caller must have synchronized such that this method cannot execute concurrently for
	 * the same subscription.
	 */
	public void saveFilterData(@NotNull FilterNetworkData data) throws Exception
	{
		if (!dirF.exists()) {
			if (!dirF.mkdir())
				throw new UserErrorException("Can't create folder with saved subscriptions: " + dirF.getPath());
			ResourceUtils.setPermissions(dirF, "ugo+rx");
		}
		File f = getFilterFile(data.filterId);
		if (!f.exists())
			ln("Saves new subscription as: " + f.getPath());
		else if (data.hasNewData)
			ln("Saves updated subscription as: " + f.getPath());
		else if (globals.debugFlag)
			ln("Updates new last-update-time for: " + f.getPath());
		if (data.docXA == null) {
			ln("Can't save subscription: data.docXA is null");
			return;
		}
		saveFilterData(data.docXA, f);
		if (data.hasNewData)
			globals.reloadPanelSettings();
	}

	private void saveFilterData(XmlArgs containerXA, File f) throws Exception
	{
		XmlParserHelper.saveDocument(containerXA.getDocument(), f);
		ResourceUtils.setPermissions(f, "ugo+r");
	}

	@NotNull
	private File getFilterFile(int filterId)
	{
		return new File(dirF, filterId + ".xml");
	}

	@NotNull
	private synchronized Object getfilterLock(int filterId)
	{
		Object lock = filterLocks.get(filterId);
		if (lock == null) {
			lock = new Object();
			filterLocks.put(filterId, lock);
		}
		return lock;
	}

	public void ln(@Nullable String s)
	{
		logger.ln(s);
	}
}
