/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client.cache;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.globals.network.HostnamePort;
import org.glimmerblocker.proxy.globals.network.NetworkConfig;
import org.glimmerblocker.proxy.globals.network.dns.DnsResolver;
import org.glimmerblocker.proxy.http.client.connection.KeepAliveException;
import org.glimmerblocker.proxy.util.HttpException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.ref.FatRef;
import dk.vitality.lib.connectionCache.ConnectionCache;
import dk.vitality.lib.connectionCache.ConnectionCounts;

/**
 * The main cache for performing client http requests.  It caches a {@link HostHttpConnectionCache}
 * for each host:port combination, i.e. maps from a host:port key to a cache of keep-alive connections
 * to that host.
 */
public class HttpClientCache implements Logger
{
	public void execute(@NotNull NetworkConfig networkConfig,
	                    @NotNull String scheme,
	                    @NotNull String host,
	                    int port,
	                    @NotNull HttpRunnable runnable) throws Exception
	{
		doExQuery(networkConfig, scheme, host, port, runnable, null);
	}

	@Nullable
	public <T> T query(@NotNull NetworkConfig networkConfig,
	                   @NotNull String scheme,
	                   @NotNull String host,
	                   int port,
	                   @NotNull HttpQuery<T> query) throws Exception
	{
		return doExQuery(networkConfig, scheme, host, port, null, query);
	}

	@Nullable
	@SuppressWarnings({"CaughtExceptionImmediatelyRethrown"})
	private <T> T doExQuery(@NotNull NetworkConfig networkConfig,
	                        @NotNull String scheme,
	                        @NotNull String host,
	                        int port,
	                        @Nullable HttpRunnable runnable,
	                        @Nullable HttpQuery<T> query) throws Exception
	{
		if (closed)
			throw new Exception("Cache is closed.");
		boolean ssl;
		if (scheme.equals("https")) {
			// proxy connections for https requests uses CONNECT,
			// so they can't be shared with connections used for plain http requests.
			ssl = true;
		} else if (scheme.equals("http")) {
			ssl = false;
		} else {
			throw new IOException("Invalid scheme = '" + scheme + "'");
		}
		/*
			All combinations:
			1) GB -> site.http
			2) GB -> site.ssl

			3) GB -> httpProxy.http -> site.http
			4) GB -> httpProxy.connect -> site.ssl

			5) GB -> socksProxy -> site.http
			6) GB -> socksProxy -> site.ssl

			7) GB -> socksProxy -> httpProxy.http -> site.http
			8) GB -> socksProxy -> httpProxy.connect -> site.ssl

			In case 3 and 7 the site hostnamePort should not be part of the key as the same
			remote proxy connection can be used for multiple sites.
		 */
		HostnamePort httpProxy = networkConfig.httpProxy;
		HostnamePort requestHostPort = (httpProxy == null || ssl) ? new HostnamePort(host, port) : null;
		HttpClientCacheKey key = new HttpClientCacheKey(
				networkConfig.dnsConfig,
				httpProxy, networkConfig.socksProxy,
				requestHostPort,
				ssl);
		int tryCount = 0;
		while (true) {
			Entry e = getEntry(key);
			try {
				if (query != null) {
					return e.connections.query(query);
				} else if (runnable != null) {
					e.connections.execute(runnable);
					return null;
				} else {
					throw new IllegalArgumentException("Both query and runnable are null");
				}
			}
			catch (HttpException ex) {
				if (debugHttpClient.get())
					ln("HttpClientCache forwards HttpException: " + ExceptionUtils.getExMsg(ex));
				throw ex;
			}
			catch (KeepAliveException ex) {
				tryCount++;
				if (tryCount > 20)
					throw (Exception)ex.getCause();
				if (debugHttpClient.get()) {
					ln("Got keep-alive exception, tries again: " + ExceptionUtils.getExMsg(ex));
					if (ex.getCause() != null)
						ln("  Cause: " + ExceptionUtils.getExMsg(ex.getCause()));
					//ln(ExceptionUtils.getStackTraceAsString(ex));
				}
			}
			finally {
				synchronized (e) {
					e.retainCount--;
					e.lastUseTime = System.currentTimeMillis();
				}
			}
		}
	}

	@NotNull
	private synchronized Entry getEntry(HttpClientCacheKey key)
	{
		Entry e = cache.get(key);
		if (e != null) {
			synchronized (e) {
				if (!e.dead) {
					e.retainCount++;
					e.lastUseTime = System.currentTimeMillis();
					return e;
				}
			}
		}
		e = new Entry(dnsResolver, key, connectionLogger, debugHttpClient);
		e.lastUseTime = System.currentTimeMillis();
		e.retainCount = 1;
		cache.put(key, e);
		return e;
	}

	public HttpClientCache(@NotNull DnsResolver dnsResolver,
	                       @NotNull Logger parentLogger,
	                       @NotNull FatRef<Boolean> debugHttpClient)
	{
		this.dnsResolver = dnsResolver;
		this.connectionLogger = new PrefixedLogger(parentLogger, "httpclient");
		this.cacheLogger = new PrefixedLogger(connectionLogger, "cache");
		this.debugHttpClient = debugHttpClient;
		this.flusher = execService.scheduleWithFixedDelay(new Flusher(), 1000, 5000, TimeUnit.MILLISECONDS);
	}

	public void close()
	{
		this.closed = true;
		this.flusher.cancel(false);
		purgeAllCurrentConnections();
	}

	public synchronized void purgeAllCurrentConnections()
	{
		for (Entry e : cache.values()) {
			synchronized (e) {
				e.connections.markAllCurrentAsNonCacheable();
			}
		}
	}

	public void debugDump()
	{
		try {
			for (int i = 0; i < 2; i++) {
				System.gc();
				Thread.sleep(2);
				System.runFinalization();
				Thread.sleep(2);
			}
		}
		catch (Exception ex) {
			cacheLogger.ln("Failed GC: " + ExceptionUtils.getExMsg(ex));
		}
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		pw.println("--------------------------------------------------------------");
		debugDump(pw);
		pw.println("–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––");
		pw.flush();
		ln(sw.toString());
	}

	public void debugDump(@NotNull PrintWriter pw)
	{
		final List<Entry> entries;
		synchronized (this) {
			entries = new ArrayList<>(cache.values());
		}
		if (entries.isEmpty()) {
			pw.println("  cache is empty.");
			return;
		}
		Collections.sort(entries, new Comparator<Entry>()
		{
			public int compare(Entry a, Entry b)
			{
				return a.connections.getCacheName().compareToIgnoreCase(b.connections.getCacheName());
			}
		});
		boolean dumpEmptyEntries = debugHttpClient.get();
		int skipped = 0;
		for (Entry e : entries) {
			synchronized (e) {
				ConnectionCounts counts = e.connections.getConnectionCounts();
				if (!dumpEmptyEntries && e.retainCount == 0
						&& counts.numBusy == 0 && counts.numIdle == 0 && counts.numOpening == 0) {
					skipped++;
					continue;
				}
				pw.print("  ");
				pw.println(e.connections.getCacheName());
				pw.print("         ");
				if (e.dead)
					pw.print("DEAD ");
				if (e.retainCount == 0 && counts.numBusy == 0 && counts.numOpening == 0) {
					pw.print(counts.numIdle + " idle.");
				} else {
					pw.print(e.retainCount == 1 ? "1 transaction, " : e.retainCount + " transactions, ");
					pw.print(counts.numBusy + " busy, ");
					pw.print(counts.numIdle + " idle, ");
					pw.print(counts.numOpening + " opening.");
					if (e.retainCount == 0)
						pw.print(" Time since last use: " + ConnectionCache.age(e.lastUseTime));
				}
				pw.println();
				e.connections.debugDump("         ", pw, false);
			}
		}
		String msg = "Unused host-entries are flushed after " + minUseTimeSecs + " secs.";
		if (skipped > 0)
			msg += " Skipped " + skipped + " " + (skipped == 1 ? "host" : "hosts") + " with no connections.";
		pw.println(msg);
	}

	//=============================================================================================
	private static final int minUseTimeSecs = 15;

	private void flushUnusedCaches()
	{
		long minUseTime = System.currentTimeMillis() - minUseTimeSecs * 1000L;
		final List<Entry> entries;
		synchronized (this) {
			entries = new ArrayList<>(cache.values());
		}
		for (Entry e : entries) {
			synchronized (e) {
				if (!e.shouldDie(minUseTime))
					continue;
			}
			// Must have lock on (this) to change the cache,
			// so the lock on (e) must be release to avoid deadlocks
			// as (this) must be locked before (e).
			synchronized (this) {
				synchronized (e) {
					// must re-test as it could have been gone into use while unlocked.
					if (!e.shouldDie(minUseTime))
						continue;
					e.dead = true;
					if (cache.get(e.key) == e)
						cache.remove(e.key);
				}
			}
			e.connections.markAllCurrentAsNonCacheable();
		}
	}

	private void flushUnusedConnections()
	{
		for (Entry e : cache.values()) {
			synchronized (e) {
				if (e.dead)
					continue;
				e.connections.markAllCurrentAsNonCacheable();
			}
		}
	}


	private class Flusher implements Runnable
	{
		public void run()
		{
			try {
				flushUnusedCaches();
				flushUnusedConnections();
			}
			catch (Throwable t) {
				ln("Got exception in flushing: " + t.getMessage());
				ln(ExceptionUtils.getExMsg(t));
			}
		}
	}

	private static final ScheduledThreadPoolExecutor execService = new ScheduledThreadPoolExecutor(1);

	@NotNull private final DnsResolver dnsResolver;
	@NotNull private final Logger connectionLogger;
	@NotNull private final Logger cacheLogger;
	@NotNull private final FatRef<Boolean> debugHttpClient;
	@NotNull private final ScheduledFuture<?> flusher;
	private boolean closed;

	public void ln(@Nullable String s)
	{
		cacheLogger.ln(s);
	}

	private static class Entry
	{
		@NotNull final HttpClientCacheKey key;
		@NotNull final HostHttpConnectionCache connections;
		long lastUseTime;
		int retainCount;
		boolean dead;

		Entry(@NotNull DnsResolver dnsResolver,
		      @NotNull HttpClientCacheKey key,
		      @NotNull Logger logger,
		      @NotNull FatRef<Boolean> debugHttpClient)
		{
			this.key = key;
			StringBuilder sb = new StringBuilder();
			if (key.socksProxy != null)
				sb.append("socks{").append(key.socksProxy.hostnamePort).append("}.");
			if (key.httpProxy != null)
				sb.append("proxy{").append(key.httpProxy.hostnamePort).append("}.");
			sb.append(key.ssl ? "https" : "http");
			sb.append("://");
			if (key.requestHostPort != null)
				sb.append(key.requestHostPort.hostnamePort);
			else
				sb.append("*");
			String name = sb.toString();
			connections = new HostHttpConnectionCache(name, key, dnsResolver, logger, debugHttpClient);
			connections.shouldLogCreationAndDestroy = false;
		}

		private boolean shouldDie(long minUseTime)
		{
			return retainCount == 0 && (dead || lastUseTime < minUseTime);
		}
	}

	// Must be ConcurrentHashMap and not just HashMap to enable iterating thru the map without
	// synchronizing to 'this'  (only needed for modifying the cache).
	private final Map<HttpClientCacheKey, Entry> cache = new ConcurrentHashMap<>();
}
