/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client.cache;

import java.io.IOException;
import java.net.HttpURLConnection;
import org.glimmerblocker.proxy.globals.network.HostnamePort;
import org.glimmerblocker.proxy.globals.network.dns.DnsResolverConfig;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.util.HttpException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * Immutable.  Fully keyable.
 */
public class HttpClientCacheKey
{
	@NotNull public final DnsResolverConfig dnsConfig;

	/**
	 * The http proxy, if any.
	 */
	@Nullable public final HostnamePort httpProxy;

	/**
	 * The socks proxy, if any.
	 */
	@Nullable public final HostnamePort socksProxy;

	/**
	 * The web-server, e.g. www.apple.com
	 * This will be null if the same connection can be used for requests for multiple sites, e.g.
	 * when using an http proxy for http requests.
	 */
	@Nullable public final HostnamePort requestHostPort;

	/**
	 * The host which GlimmerBlocker makes a tcp connection to when executing the requests.
	 */
	@NotNull public final HostnamePort tcpHost;

	/**
	 * Are the requests using https, i.e. must the socket be an SSL socket instead of plain socket.
	 */
	public final boolean ssl;

	/**
	 * The first http(s)-based host that GB contacts in the
	 * GB → optional-socksProxy → optional-httpProxy → web-server
	 * sequence.
	 */
	@NotNull public final HostnamePort firstHttpHost;

	public HttpClientCacheKey(@NotNull DnsResolverConfig dnsConfig,
	                          @Nullable HostnamePort httpProxy,
	                          @Nullable HostnamePort socksProxy,
	                          @Nullable final HostnamePort requestHostPort,
	                          boolean ssl) throws Exception
	{
		this.dnsConfig = dnsConfig;
		this.httpProxy = httpProxy;
		this.socksProxy = socksProxy;
		this.requestHostPort = requestHostPort;
		this.ssl = ssl;
		//
		if (socksProxy != null)
			tcpHost = socksProxy;
		else if (httpProxy != null)
			tcpHost = httpProxy;
		else if (requestHostPort != null)
			tcpHost = requestHostPort;
		else
			throw new NullPointerException("All of socksProxy, httpProxy and httpServer are null.");
		//
		if (httpProxy != null)
			firstHttpHost = httpProxy;
		else if (requestHostPort != null)
			firstHttpHost = requestHostPort;
		else
			throw new NullPointerException("Both httpProxy and httpServer are null.");
		//
		if (ssl && requestHostPort == null)
			throw new NullPointerException("requestHostPort must not be null for SSL requests.");
		//
		checkPort(socksProxy, 20, "socks-proxy port");
		checkPort(httpProxy, 80, "http-proxy port");
		checkPort(requestHostPort, 80, "request site port");
		//
		if (socksProxy != null && firstHttpHost.hostname.length() > 255)
			throw new IOException("Host names cannot be more than 255 chars when using SOCKS5");
	}

	private static void checkPort(@Nullable HostnamePort hp, int minPort, String desc) throws Exception
	{
		if (hp == null || hp.port >= minPort)
			return;
		Headers h = new Headers(HttpURLConnection.HTTP_BAD_REQUEST, "Invalid " + desc + " = " + hp.hostnamePort);
		throw new HttpException(h);
	}

	public boolean equals(Object o)
	{
		if (this == o)
			return true;
		if (o == null || getClass() != o.getClass())
			return false;
		HttpClientCacheKey key = (HttpClientCacheKey)o;
		return key.ssl == ssl
				&& HostnamePort.equals(key.requestHostPort, requestHostPort)
				&& HostnamePort.equals(key.httpProxy, httpProxy)
				&& HostnamePort.equals(key.socksProxy, socksProxy)
				&& dnsConfig.equals(dnsConfig);
	}

	public int hashCode()
	{
		int hash = dnsConfig.hashCode();
		hash = hash * 31 + (httpProxy != null ? httpProxy.hashCode() : 0);
		hash = hash * 31 + (socksProxy != null ? socksProxy.hashCode() : 0);
		hash = hash * 31 + (requestHostPort != null ? requestHostPort.hashCode() : 0);
		return hash + (ssl ? 3 : 0);
	}
}
