/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client.curl;

import java.net.HttpURLConnection;
import java.net.InetAddress;
import java.net.URL;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.network.NetworkConfig;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.rhino.runtime.ExecutorDelegate;
import org.glimmerblocker.proxy.rule.RequestRule;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.RhinoException;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.ref.FatRef;

class RequestScriptExecutor implements ExecutorDelegate<RequestGlobalObject>, Logger
{
	@NotNull final GlimmerGlobals globals;
	@NotNull final CurlBase curlBase;
	@Nullable InetAddress clientAddr;
	@NotNull final RequestRule rule;
	@NotNull final FatRef<NetworkConfig> networkConfigRef;
	ResultData rd;

	RequestScriptExecutor(@NotNull CurlBase curlBase,
	                      @Nullable InetAddress clientAddr,
	                      @NotNull RequestRule rule)
	{
		this.globals = curlBase.globals;
		this.curlBase = curlBase;
		this.clientAddr = clientAddr;
		this.rule = rule;
		if (rule.js == null)
			throw new NullPointerException();
		this.networkConfigRef = new FatRef<>(curlBase.networkConfig);
		try {
			if (globals.debugFilters && globals.debugVerbose)
				ln("Executes: " + rule.toString() + "\n" + rule.js);
			globals.rhinoRuntimeCache.execute(this, rule.runtimeId, rule.js, rule.toString());
		}
		catch (Throwable ex) {
			ln("Request script failed with exception: " + ExceptionUtils.getExMsg(ex));
			ln("   " + ex.getClass().getName());
			ln("   filter: " + rule.filterName);
			if (ex instanceof RhinoException) {
				RhinoException rex = (RhinoException)ex;
				ln("  line " + rex.lineNumber() + ", col " + rex.columnNumber());
			}
			ln("   " + rule);
			if (!(ex instanceof RhinoException) || (globals.debugFilters && globals.debugVerbose))
				ln(ExceptionUtils.getStackTraceAsString(ex));
			Throwable t = ex;
			while (t != null) {
				if (t instanceof NullPointerException) {
					ln(ExceptionUtils.getStackTraceAsString(t));
					break;
				}
				t = t.getCause();
			}
		}
	}

	@NotNull
	public RequestGlobalObject create(@NotNull ScriptableObject scope,
	                                  @NotNull Scriptable defaultPrototype) throws Exception
	{
		return new RequestGlobalObject(globals, scope, defaultPrototype);
	}

	public void prepare(@NotNull RequestGlobalObject callback, @NotNull Context ctx) throws Exception
	{
		callback.prepareRequestObject(ctx, curlBase, clientAddr,
				curlBase.method, curlBase.url, curlBase.requestHeaders, networkConfigRef);
	}

	public void executionDone(@NotNull RequestGlobalObject gb) throws Exception
	{
		curlBase.networkConfig = networkConfigRef.get();
		if (gb.request != null) {
			String method = gb.request.method;
			if (method != null)
				curlBase.method = method;
			//
			URL url = gb.request.url;
			if (url != null)
				curlBase.url = url;
			//
			HeadersMapObject headers = gb.request.headers;
			if (headers != null)
				curlBase.requestHeaders = headers.extract(HttpURLConnection.HTTP_OK, "OK");
		}
		rd = gb.getOptResultData();
	}

	public void ln(@Nullable String s)
	{
		curlBase.ln(s);
	}
}
