/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.data;

import java.net.HttpURLConnection;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TimeZone;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * HeaderBuilder is to {@link Headers} as StringBuilder is to String.
 */
public class HeadersBuilder
{
	public int status;
	@NotNull public String statusMessage;
	@Nullable public String contentType;
	@Nullable public String contentEncoding;
	public boolean icyProtocol;
	private final Map<String, List<String>> headers = new HashMap<>();

	private static final SimpleDateFormat rfc1123 = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz");

	static {
		rfc1123.setTimeZone(TimeZone.getTimeZone("GMT"));
	}

	/**
	 * 200 OK, without content-type.
	 */
	public static HeadersBuilder makeEmpty()
	{
		return new HeadersBuilder(HttpURLConnection.HTTP_OK, "OK");
	}

	/**
	 * 200 OK, with text/plain content-type.
	 */
	public static HeadersBuilder makeTextPlain()
	{
		final HeadersBuilder hb = makeEmpty();
		hb.contentType = "text/plain; charset=UTF-8";
		return hb;
	}

	/**
	 * Create using the specified HTTP status and text/plain content-type.
	 *
	 * @param status        HTTP status code such as 404.
	 * @param statusMessage HTTP status message such as "OK".
	 */
	public static HeadersBuilder makeErrorTextPlain(int status, @NotNull String statusMessage)
	{
		final HeadersBuilder hb = makeTextPlain();
		hb.status = status;
		hb.statusMessage = statusMessage;
		return hb;
	}

	/**
	 * Make clone.
	 */
	public HeadersBuilder(@NotNull Headers h)
	{
		this.status = h.status;
		this.statusMessage = h.statusMessage;
		this.contentType = h.contentType;
		this.contentEncoding = h.contentEncoding;
		copyAll(h.headerMap);
	}

	public HeadersBuilder(int status, @NotNull String statusMessage)
	{
		this.status = status;
		this.statusMessage = statusMessage;
	}

	/**
	 * Adds an HTTP header. Duplicate headers are added instead of replaced to enable multiple Set-Cookie headers.
	 *
	 * @param nameLC The header name. <b>MUST BE LOWERCASE.</b>
	 * @param value  The header value.
	 */
	public void add(@NotNull String nameLC, @NotNull String value)
	{
		List<String> list = headers.get(nameLC);
		if (list == null) {
			list = new ArrayList<>();
			headers.put(nameLC, list);
		}
		list.add(value);
	}

	/**
	 * Adds an HTTP header formatted as a RFC1123 date.
	 * Duplicate headers are added instead of replaced to enable multiple Set-Cookie headers.
	 *
	 * @param nameLC    The header name. <b>MUST BE LOWERCASE.</b>
	 * @param dateValue The header value (millisecs).
	 */
	@SuppressWarnings("unused")
	public void addDate(@NotNull String nameLC, long dateValue)
	{
		String s;
		synchronized (rfc1123) {
			s = rfc1123.format(new Date(dateValue));
		}
		add(nameLC, s);
	}

	public void addNoCacheHeaders()
	{
		removeAll("pragma");
		removeAll("cache-control");
		add("pragma", "no-cache");
		add("cache-control", "no-cache");
	}

	public void removeAll(@NotNull String nameLC)
	{
		headers.remove(nameLC);
	}

	/**
	 * Removes existing headers whose names is specified in the copied parameter.
	 */
	public void copyAll(@NotNull Map<String, List<String>> headerMap)
	{
		for (String nameLC : headerMap.keySet()) {
			removeAll(nameLC);
			for (String value : headerMap.get(nameLC))
				add(nameLC, value);
		}
	}

	/**
	 * Create a {@link Headers} instance, similar to {@link StringBuilder#toString}.
	 */
	@NotNull
	public Headers toHeaders()
	{
		return new Headers(status, statusMessage, contentType, contentEncoding, headers, icyProtocol);
	}


	/**
	 * HTTP headers which always can be forwarded from the client even when
	 * a completely different URL is requested.
	 */
	public static final Set<String> basicBrowserHeadersLC;

	static {
		Set<String> set = new HashSet<>();
		set.add("user-agent");
		set.add("referer");
		set.add("accept");
		set.add("accept-language");
		set.add("cookie");
		basicBrowserHeadersLC = Collections.unmodifiableSet(set);
	}

	/**
	 * HTTP headers which never must be forwarded, such as connection-handling headers (e.g. Keep-Alive).
	 * All headers whose name start with "Proxy" are ignored by {@link #copyBrowserHeaders} and are not
	 * included in this set.
	 * <p/>
	 * RFC 2616: 13.5.1 End-to-end and Hop-by-hop Headers
	 */
	public static final Set<String> neverCopyHeadersLC;

	static {
		Set<String> set = new HashSet<>();
		set.add("accept-encoding"); // the proxy forces it to gzip
		set.add("keep-alive");
		set.add("connection");
		set.add("host");
		set.add("content-length");
		set.add("te");
		set.add("trailers");
		set.add("transfer-encoding");
		set.add("upgrade");
		neverCopyHeadersLC = Collections.unmodifiableSet(set);
	}

	public enum HeaderSelection
	{
		Minimal, Maximal
	}

	/**
	 * Copy the request headers from the client.
	 */
	public void copyBrowserHeaders(Headers requestHeaders, HeaderSelection selection) throws Exception
	{
		for (String nameLC : requestHeaders.headerMap.keySet()) {
			if (nameLC.startsWith("proxy") || neverCopyHeadersLC.contains(nameLC))
				continue;
			if (selection == HeaderSelection.Minimal && !basicBrowserHeadersLC.contains(nameLC))
				continue;
			for (String value : requestHeaders.headerMap.get(nameLC)) {
				add(nameLC, value);
			}
		}
	}

	/**
	 * Addes or removes "Accept-Encoding: gzip"
	 */
	public void setAllowGzip(boolean allow)
	{
		if (allow)
			add(acceptEncodingHeaderNameLC, "gzip");
		else
			removeAll(acceptEncodingHeaderNameLC);
	}

	public static final String acceptEncodingHeaderNameLC = "accept-encoding"; // I can't always spall.
}
