/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.data;

import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * Immutable structure holding response headers and content.
 */
public class ResultData
{
	/**
	 * The HTTP headers.
	 */
	@NotNull public final Headers headers;

	/**
	 * The content as bytes.
	 */
	@Nullable protected byte[] contentBytes;

	/**
	 * The content as String.
	 */
	@Nullable protected String contentString;

	/**
	 * Create without content.
	 */
	public ResultData(@NotNull Headers headers)
	{
		this.headers = headers;
		this.contentBytes = null;
		this.contentString = null;
	}

	/**
	 * Create with the specified content.
	 *
	 * @param content No copy is made, so the byte array MUST NOT BE MODIFIED by the creator.
	 */
	public ResultData(@NotNull Headers headers, @Nullable byte[] content)
	{
		this.headers = headers;
		this.contentBytes = content;
		this.contentString = null;
	}

	/**
	 * @param headers MUST indicate UTF-8 in content-type
	 */
	public ResultData(@NotNull Headers headers, @NotNull String content)
	{
		this.headers = headers;
		this.contentBytes = null;
		this.contentString = content;
	}

	@Nullable
	public byte[] getContent() throws UnsupportedEncodingException
	{
		if (contentString != null)
			contentBytes = contentString.getBytes("UTF-8");
		return contentBytes;
	}

	/**
	 * Get the content as a string by adhering to the charset specified in the Content-Type header.
	 * <p/>
	 * http://www.whatwg.org/specs/web-apps/current-work/multipage/parsing.html
	 */
	@NotNull
	public String getContentAsText(Logger logger, boolean debugFlag) throws Exception
	{
		if (contentString != null)
			return contentString;
		if (contentBytes == null || contentBytes.length == 0)
			return "";
		CharsetDeterminer cd = new CharsetDeterminer(this, logger, debugFlag);
		return cd.getContentAsText();
	}

	/**
	 * Simple 404 response.
	 */
	public static final ResultData NOT_FOUND;

	static {
		HeadersBuilder hb = HeadersBuilder.makeErrorTextPlain(HttpURLConnection.HTTP_NOT_FOUND, "GlimmerBlocked");
		hb.add("Cache-control", "no-cache, no-store");
		NOT_FOUND = new ResultData(hb.toHeaders());
	}

	/**
	 * Simple 403 response.
	 */
	public static final ResultData FORBIDDEN;

	static {
		HeadersBuilder hb = HeadersBuilder.makeErrorTextPlain(HttpURLConnection.HTTP_FORBIDDEN, "GlimmerBlocked");
		hb.add("Cache-control", "no-cache, no-store");
		FORBIDDEN = new ResultData(hb.toHeaders());
	}

	/**
	 * GIF with light-gray checkerboard pattern
	 */
	public static final ResultData checkerboardPatternGif = makeGif(new int[]{
			0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x04, 0x00, 0x04, 0x00, 0x80, 0x00, 0x00, 0xc8, 0xc8, 0xc8,
			0xff, 0xff, 0xff, 0x21, 0xfe, 0x0e, 0x49, 0x20, 0x77, 0x61, 0x73, 0x20, 0x61, 0x20, 0x62, 0x61,
			0x6e, 0x6e, 0x65, 0x72, 0x00, 0x21, 0xf9, 0x04, 0x01, 0x0a, 0x00, 0x01, 0x00, 0x2c, 0x00, 0x00,
			0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x00, 0x02, 0x05, 0x44, 0x7c, 0x67, 0xb8, 0x05, 0x00, 0x3b
	});

	/**
	 * Transparent 1x1 pixel GIF.
	 */
	public static final ResultData transparentGif = makeGif(new int[]{
			0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x01, 0x00, 0x01, 0x00, 0x80, 0x00, 0x00, 0xff, 0xff, 0xff,
			0x00, 0x00, 0x00, 0x21, 0xf9, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x2c, 0x00, 0x00, 0x00, 0x00,
			0x01, 0x00, 0x01, 0x00, 0x00, 0x02, 0x02, 0x44, 0x01, 0x00, 0x3b
	});

	private static ResultData makeGif(int[] data)
	{
		HeadersBuilder hb = HeadersBuilder.makeEmpty();
		hb.add("Content-type", "image/gif");
		hb.add("Cache-control", "no-cache, no-store");
		byte[] bytes = new byte[data.length];
		for (int i = 0; i < data.length; i++)
			bytes[i] = (byte)data[i];
		return new ResultData(hb.toHeaders(), bytes);
	}
}
