/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.server;

import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;

public class HttpServer
{
	@NotNull public final GlimmerGlobals globals;

	private int listenPort = -1;
	private int serverSocketCount;
	private Dispatcher dispatcherWildcard, dispatcherIP4, dispatcherIP6;

	public HttpServer(@NotNull GlimmerGlobals globals)
	{
		this.globals = globals;
	}

	public synchronized void notifyPrefsChanged() throws Exception
	{
		boolean newWildcard = globals.panelSettingsXA.getOptBoolean("proxy-server/@listen-wildcard");
		int newListenPort = globals.panelSettingsXA.getOptInt("proxy-server/@listen-port", 8228);
		if (newListenPort == listenPort) {
			Dispatcher d = newWildcard ? dispatcherWildcard : dispatcherIP4;
			if (d != null)
				return;
		}
		//
		if (dispatcherWildcard != null)
			dispatcherWildcard.close();
		if (dispatcherIP4 != null)
			dispatcherIP4.close();
		if (dispatcherIP6 != null)
			dispatcherIP6.close();
		//
		// InetAddress.getLocalHost returns external visible IP.
		//
		serverSocketCount++;
		if (newWildcard) {
			dispatcherWildcard = new Dispatcher("HttpServer-*", null, newListenPort);
		} else {
			// InetAddress.getAllByName(null) returns only 127.0.0.1, so we need to explicit listen on IPv6 loopback.
			dispatcherIP4 = new Dispatcher("HttpServer-ip4", "127.0.0.1", newListenPort);
			if (System.getProperty("java.net.preferIPv4Stack") == null)
				dispatcherIP6 = new Dispatcher("HttpServer-ip6", "::1", newListenPort);
			else
				globals.ln("Skips binding to ipv6 localhost due to java.net.preferIPv4Stack");
		}
		listenPort = newListenPort;
	}

	private static String getPrefer(String name)
	{
		String s = System.getProperty(name);
		return (s == null || s.isEmpty()) ? "False" : s;
	}

	public int getListenPort()
	{
		return listenPort;
	}

	private class Dispatcher implements Runnable, Logger
	{
		@NotNull private Logger logger;
		@Nullable private ServerSocket serverSocket;

		Dispatcher(@NotNull String loggerPrefix, @Nullable String bindText, int newListenPort)
		{
			this.logger = new PrefixedLogger(globals, loggerPrefix);
			ServerSocket ss;
			try {
				InetAddress listenAddress = (bindText == null) ? null : InetAddress.getByName(bindText);
				ln("Opens ServerSocket for " + (bindText == null ? "*" : bindText) + ", port " + newListenPort +
						", preferIPv4 = " + getPrefer("java.net.preferIPv4Stack") +
						", preferIPv6 = " + getPrefer("java.net.preferIPv6Addresses"));
				ss = new ServerSocket(newListenPort, 20, listenAddress);
			}
			catch (Throwable ex) {
				ln("Got exception while opening server socket: " + ExceptionUtils.getExMsg(ex));
				if (bindText != null && bindText.equals("::1"))
					ln("Proxy server does not listen on IPv6 localhost.");
				else
					ln("PROXY SERVER IS NOT RUNNING.");
				serverSocket = null;
				return;
			}
			serverSocket = ss;
			new Thread(this, "HttpServer-" + loggerPrefix + "-" + serverSocketCount).start();
		}

		public void run()
		{
			if (serverSocket == null) {
				ln("serverSocket is null");
				return;
			}
			ln("Http server ready.");
			long lastLogMsgTime = 0;
			while (!serverSocket.isClosed()) {
				try {
					Socket skt = serverSocket.accept();
					new HttpServerConnection(globals, skt);
				}
				catch (Throwable ex) {
					long now = System.currentTimeMillis();
					if (now - lastLogMsgTime > 60 * 1000L && !isSocketClosed(ex)) {
						ln("ServerSocket.accept got exception: " + ExceptionUtils.getExMsg(ex));
						ln(ex.getClass().getName());
						lastLogMsgTime = now;
					}
				}
			}
			close();
			if (globals.debugHttpServer)
				ln("Dispatcher done.");
		}

		boolean isSocketClosed(Throwable t)
		{
			return t instanceof SocketException && t.getMessage() != null && t.getMessage().equals("Socket closed");
		}

		void close()
		{
			ServerSocket skt = serverSocket;
			if (skt == null || skt.isClosed())
				return;
			try {
				skt.close();
				ln("Socket closed.");
			}
			catch (Throwable ex) {
				ln("Got exception while closing server socket: " + ExceptionUtils.getExMsg(ex));
			}
		}

		public void ln(@Nullable String s)
		{
			logger.ln(s);
		}
	}
}
