/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.server;

import java.io.IOException;
import java.io.OutputStream;
import org.jetbrains.annotations.NotNull;

class HttpServerContentLengthOutputStream extends OutputStream
{
	@NotNull private final HttpServerRequest req;
	@NotNull private final OutputStream os;
	private long remaining;
	private boolean closed;

	HttpServerContentLengthOutputStream(@NotNull HttpServerRequest req, long remaining) throws IOException
	{
		this.req = req;
		this.os = req.conn.os;
		this.remaining = remaining;
		if (remaining < 0)
			throw new IOException("Invalid response ContentLength = " + remaining);
		checkEOF();
	}

	public void write(int b) throws IOException
	{
		ensureRemaining(1);
		os.write(b);
		remaining--;
		checkEOF();
	}

	@Override
	public void write(@NotNull byte[] data, int offset, int len) throws IOException
	{
		if (len == 0)
			return;
		ensureRemaining(len);
		os.write(data, offset, len);
		remaining -= len;
		checkEOF();
	}

	@Override
	public void close() throws IOException
	{
		if (closed)
			return;
		if (remaining > 0) {
			String msg = "Closing ContentLength-OutputStream when " + remaining + " bytes remain to be sent.";
			if (req.globals.debugHttpServer && !req.keepaliveUsable())
				req.ln(msg);
			req.disableKeepalive(msg);
		}
		try {
			flush();
		}
		finally {
			closed = true; // must be after flush() to avoid making the flush() call a no-op.
			req.notifyHasSentFullResponseBody();
		}
	}

	@Override
	public void flush() throws IOException
	{
		if (closed)
			return;
		os.flush();
	}

	private void ensureRemaining(long numBytes) throws IOException
	{
		if (closed)
			throw new IOException("OutputStream is closed");
		if (remaining >= numBytes)
			return;
		throw new IOException("Trying to send " + numBytes +
				" when only " + remaining + " bytes left of" +
				" promised ContentLength = " + req.responseBodyContentLength);
	}

	private void checkEOF()
	{
		if (remaining == 0)
			req.notifyHasSentFullResponseBody();
	}
}
