/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.rhino.callback;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.http.client.curl.ResponseSubObject;
import org.glimmerblocker.proxy.http.client.curl.StaticRequestSubObject;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.util.HtmlUtils;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.Function;
import org.mozilla.javascript.NativeArray;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import static java.util.regex.Pattern.CASE_INSENSITIVE;
import static java.util.regex.Pattern.DOTALL;
import static java.util.regex.Pattern.compile;

@SuppressWarnings({"NonSerializableFieldInSerializableClass"})
public abstract class ContentGlobalObjectBase extends GlimmerGlobalObjectBase
{
	boolean hasCreatedGlobals;
	private ResponseSubObject response;
	private Headers responseHeaders;

	protected ContentGlobalObjectBase(@NotNull GlimmerGlobals globals,
	                                  @NotNull Scriptable scope,
	                                  @NotNull Scriptable prototype,
	                                  @NotNull CallbackRhinoBinder binder) throws IOException
	{
		super(globals, scope, prototype, binder);
	}

	protected void prepareContentCallback(@NotNull Logger logger,
	                                      @NotNull Context ctx,
	                                      @Nullable String t,
	                                      @NotNull Headers responseHeaders) throws Exception
	{
		super.prepareGlobalBase(logger, ctx);
		this.responseHeaders = responseHeaders;
		if (response != null)
			prepareResponseObject();
		setContent(t);
		if (hasCreatedGlobals)
			return;
		String js = "" +
				"function replace(regexp, replacement)\n" +
				"{\n" +
				"  t = t.replace(regexp, replacement);\n" +
				"}";
		ctx.evaluateString(scope, js, "ContentCallbacksGlobals.js", 1, null);
		hasCreatedGlobals = true;
	}

	@NotNull
	protected abstract StaticRequestSubObject getRequestObject() throws Exception;

	@Override
	public void notifyLeftContext()
	{
		super.notifyLeftContext();
		responseHeaders = null;
	}

	@Nullable
	public ResultData getOptResultData() throws UnsupportedEncodingException
	{
		String t = getContent();
		if (t != null && t.trim().length() > 0) {
			Headers h;
			if (response != null)
				h = response.extract();
			else if (responseHeaders != null)
				h = responseHeaders;
			else
				h = Headers.simpleTextHtml;
			// Don't force content type here, as that would change e.g. javascript transformers
			// into returning a text/html content-type.
			//h = h.copyWithContentType(Headers.simpleTextHtml.contentType);
			return new ResultData(h, t);
		}
		if (response != null && response.dirty()) {
			return new ResultData(response.extract());
		}
		return null;
	}

	@Nullable
	public String getContent()
	{
		Object tt = scope.get("t", scope);
		if (tt == null)
			return null; // Context.toString() makes null into "null".
		return Context.toString(tt);
	}

	public void setContent(@Nullable String t)
	{
		scope.put("t", scope, t);
	}

	//===========================================================================

	@RhinoCallback
	public ResponseSubObject jsGet_response() throws Exception
	{
		if (response == null) {
			response = new ResponseSubObject(this);
			response.bind();
			prepareResponseObject();
		}
		return response;
	}

	private void prepareResponseObject() throws Exception
	{
		response.prepareResponseSubObject(ctx, getRequestObject(), responseHeaders, this);
	}

	//===========================================================================

	@RhinoCallback
	public String jsFunction_transformCharset(String txt, String from, String to) throws Exception
	{
		if (from == null || from.isEmpty())
			throw new Exception("Invalid 'from' charset");
		if (to == null || to.isEmpty())
			throw new Exception("Invalid 'to' charset");
		Charset fromCS, toCS;
		try {
			fromCS = Charset.forName(from);
		}
		catch (Throwable ex) {
			throw new Exception("Invalid 'from' charset: " + ExceptionUtils.getExMsg(ex), ex);
		}
		try {
			toCS = Charset.forName(to);
		}
		catch (Throwable ex) {
			throw new Exception("Invalid 'to' charset: " + ExceptionUtils.getExMsg(ex), ex);
		}
		if (txt == null || txt.isEmpty())
			return txt;
		byte[] b;
		try {
			b = txt.getBytes(fromCS);
		}
		catch (Throwable ex) {
			throw new Exception("Failed converting from '" + from + "' charset: " + ExceptionUtils.getExMsg(ex), ex);
		}
		String s;
		try {
			s = new String(b, toCS);
		}
		catch (Throwable ex) {
			throw new Exception("Failed converting to '" + to + "' charset: " + ExceptionUtils.getExMsg(ex), ex);
		}
		return s;
	}

	//===========================================================================

	protected HtmlHelper getHtmlHelper()
	{
		return new HtmlHelper(getContent());
	}

	public void replaceAll(@NotNull Pattern re, @NotNull String replacement)
	{
		HtmlHelper h = getHtmlHelper();
		if (h.html == null)
			return;
		h.replaceAll(re, replacement);
		setContent(h.html);
	}

	@RhinoCallback
	public void jsFunction_zapScriptElements(boolean unwrapNoscript)
	{
		if (globals.debugFilters && globals.debugVerbose)
			ln("jsFunction_zapScriptElements(" + unwrapNoscript + ")");
		if (!unwrapNoscript) {
			replaceAll(scriptRE, "<script type='text/javascript'>/* GlimmerBlocker zapped script element */</script>");
			return;
		}
		HtmlHelper h = getHtmlHelper();
		if (h.html == null)
			return;
		Matcher m = noscriptRE.matcher(h.html);
		if (!m.find())
			return;
		StringBuilder sb = new StringBuilder(h.html.length());
		int prev = 0;
		while (true) {
			sb.append(h.html.substring(prev, m.start()));
			sb.append("<script type='text/javascript'>/* GlimmerBlocker zapped script element");
			String noscript = m.group(1);
			sb.append(noscript != null ? ", and unwraps noscript content." : ", without noscript content.");
			sb.append(" */</script>");
			if (noscript != null)
				sb.append(noscript).append("<!-- GB: end of noscript -->");
			prev = m.end();
			if (!m.find(m.end()))
				break;
		}
		sb.append(h.html.substring(prev));
		setContent(sb.toString());
	}

	private static final Pattern scriptRE = compile("<script\\b.*?>(.*?)</script\\s*>",
			DOTALL | CASE_INSENSITIVE);
	private static final Pattern noscriptRE = compile("" +
					"<script\\b.*?</script\\s*>" +
					"(?:\\s*<noscript\\b.*?>(.*?)</noscript\\s*>)?",
			DOTALL | CASE_INSENSITIVE);

	@RhinoCallback
	public void jsFunction_zapStyleElements()
	{
		replaceAll(styleRE, "<script type='text/javascript'>/* GlimmerBlocker zapped style element */</script>");
	}

	private static final Pattern styleRE = compile("<style\\b.*?</style\\s*>", DOTALL | CASE_INSENSITIVE);

	@RhinoCallback
	public void jsFunction_zapIframeElements()
	{
		replaceAll(iframeRE, "<script type='text/javascript'>/* GlimmerBlocker zapped iframe element */</script>");
	}

	private static final Pattern iframeRE = compile("<iframe\\b.*?</iframe\\s*>", DOTALL | CASE_INSENSITIVE);

	@RhinoCallback
	public void jsFunction_addScriptElement(String js)
	{
		String html = "<script type='text/javascript'> // GlimmerBlocker script\n" +
				js.trim() + "\n" +
				"</script>\n";
		HtmlHelper h = getHtmlHelper();
		h.insertAtBodyEnd(html, true);
		setContent(h.html);
	}

	@RhinoCallback
	public void jsFunction_addStyleElement(String css)
	{
		String html = "<style type='text/css'> /* GlimmerBlocker js style */\n" +
				css.trim() + "\n" +
				"</style>\n";
		HtmlHelper h = getHtmlHelper();
		h.insertAtHeadEnd(html, true);
		setContent(h.html);
	}

	@RhinoCallback
	public void jsFunction_insertAtHeadStart(String html)
	{
		HtmlHelper h = getHtmlHelper();
		h.insertAtHeadStart(html, true);
		setContent(h.html);
	}

	@RhinoCallback
	public void jsFunction_insertAtHeadEnd(String html)
	{
		HtmlHelper h = getHtmlHelper();
		h.insertAtHeadEnd(html, true);
		setContent(h.html);
	}

	@RhinoCallback
	public void jsFunction_insertAtBodyStart(String html)
	{
		HtmlHelper h = getHtmlHelper();
		h.insertAtBodyStart(html, true);
		setContent(h.html);
	}

	@RhinoCallback
	public void jsFunction_insertAtBodyEnd(String html)
	{
		HtmlHelper h = getHtmlHelper();
		h.insertAtBodyEnd(html, true);
		setContent(h.html);
	}

	//===========================================================================

	@RhinoCallback
	public void jsFunction_setDocumentTitle(String title)
	{
		HtmlHelper h = getHtmlHelper();
		String titleHtml = HtmlUtils.encodeHtml(title);
		if (h.html != null) {
			Matcher m = titleRE.matcher(h.html);
			if (m.find()) {
				setContent(h.html.substring(0, m.end(1)) +
						titleHtml +
						h.html.substring(m.start(2)));
				return;
			}
		}
		h.insertAtHeadEnd("<title>" + titleHtml + "</title>\n", true);
		setContent(h.html);
	}

	private static final Pattern titleRE = Pattern.compile(
			"(<title\\b.*?>).*?(</title\\s*>)", DOTALL | CASE_INSENSITIVE);

	//===========================================================================

	@RhinoCallback
	public String jsGet_attachmentDownloadHostPrefix()
	{
		return globals.javaSettings.getAttachmentDownloadHostPrefix();
	}

	//===========================================================================

	@SuppressWarnings({"UnusedDeclaration"})
	public static Object jsFunction_shp(Context ctx,
	                                    Scriptable thisObj,
	                                    Object[] args,
	                                    Function funObj) throws Exception
	{
		ContentGlobalObjectBase cb = (ContentGlobalObjectBase)thisObj;
		if (args == null || args.length < 1 || args.length > 2)
			throw new Exception("Invalid number of arguments to gb.shp()");
		String html;
		if (args[0] == null) {
			html = "";
		} else if (args[0] instanceof String) {
			html = (String)args[0];
		} else if (args[0] instanceof NativeArray) {
			ScriptableObject a = (ScriptableObject)args[0];
			Number idx = Context.toNumber(a.get("index", a));
			html = Context.toString(a.get("input", a));
			html = html.substring(idx.intValue());
		} else {
			throw new Exception("Invalid type of first argument: " + args[0].getClass().getName());
		}
		return new SimpleHtmlParserCallback(cb, html);
	}

	@SuppressWarnings({"serial"})
	static class SimpleHtmlParserCallback extends ScriptableObject
	{
		final SimpleHtmlParser shp;
		private static final CallbackRhinoBinder binder = new CallbackRhinoBinder(SimpleHtmlParserCallback.class);

		SimpleHtmlParserCallback(@NotNull ContentGlobalObjectBase cb, @NotNull String html)
		{
			shp = new SimpleHtmlParser(cb, html, 0);
			setPrototype(cb.getPrototype());
			binder.bind(this);
		}

		public String getClassName()
		{
			return "SimpleHtmlParser";
		}

		@RhinoCallback
		public String jsFunction_subtree()
		{
			return shp.parseSubTree();
		}

		@RhinoCallback
		public String jsFunction_rest()
		{
			return (shp.idx < shp.html.length()) ? shp.html.substring(shp.idx) : "";
		}
	}
}
