/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.rhino.callback;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.http.data.HeadersBuilder;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.proxyRequest.transform.MimeTypeMatcher;
import org.glimmerblocker.proxy.util.HtmlUtils;
import org.glimmerblocker.proxy.util.JsString;
import org.glimmerblocker.proxy.util.ResourceUtils;
import org.jabhts.runtime.html.HtmlAttr;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.Function;
import org.mozilla.javascript.NativeArray;
import org.mozilla.javascript.Script;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.Undefined;
import dk.vitality.lib.foundation.html.URLUtils;
import dk.vitality.lib.foundation.misc.VersionStringComparator;

public abstract class GlimmerGlobalObjectBase extends CallbackObjectBase
{
	public Script script;
	private Function stringify;
	private static String stringifySource;

	private Logger logger;

	protected GlimmerGlobalObjectBase(@NotNull GlimmerGlobals globals,
	                                  @NotNull Scriptable scope,
	                                  @NotNull Scriptable prototype,
	                                  @NotNull CallbackRhinoBinder binder) throws IOException
	{
		super(scope, prototype, binder, globals);
		if (stringifySource == null) {
			byte[] b = ResourceUtils.getResource("RhinoStringify.js", GlimmerGlobalObjectBase.class);
			stringifySource = new String(b, "UTF-8");
		}
	}

	public void prepareGlobalBase(@NotNull Logger logger,
	                              @NotNull Context ctx) throws Exception
	{
		this.logger = logger;
		super.prepareObjectBase(ctx);
	}

	public void notifyLeftContext()
	{
		super.notifyLeftContext();
		logger = null;
	}

	public void ln(@Nullable String s)
	{
		if (logger != null)
			logger.ln(s);
		else
			System.err.println(s);
	}

	//=====================================================================================

	@RhinoCallback
	public String jsFunction_htmlEncode(String s)
	{
		return HtmlUtils.encodeHtml(s);
	}

	@RhinoCallback
	public String jsFunction_attrEncode(String s)
	{
		return HtmlAttr.escape(s);
	}

	@RhinoCallback
	public String jsFunction_attrQuote(String s)
	{
		return HtmlAttr.quoted(s);
	}

	@RhinoCallback
	public String jsFunction_jsEncode(String s)
	{
		return JsString.encode(s);
	}

	@RhinoCallback
	public String jsFunction_urlEncode(String s)
	{
		return URLUtils.encodeURL(s);
	}

	@RhinoCallback
	public String jsFunction_urlDecode(String s)
	{
		return URLUtils.decodeURL(s, true);
	}

	@RhinoCallback
	public String jsGet_versionString()
	{
		return globals.gbVersionString;
	}

	@RhinoCallback
	public int jsFunction_versionCompare(String s)
	{
		return VersionStringComparator.instance.compare(globals.gbVersionString, s);
	}

	@RhinoCallback
	public void jsFunction_sleep(int msecs) throws InterruptedException
	{
		Thread.sleep(msecs);
	}

	@RhinoCallback
	@SuppressWarnings({"UnusedDeclaration"})
	public static void jsFunction_log(Context cx, Scriptable thisObj, Object[] args, Function funObj)
	{
		GlimmerGlobalObjectBase cb = (GlimmerGlobalObjectBase)thisObj;
		cb.log(args);
	}

	@SuppressWarnings({"UnusedDeclaration"})
	public static void jsFunction_dlog(Context cx, Scriptable thisObj, Object[] args, Function funObj)
	{
		GlimmerGlobalObjectBase cb = (GlimmerGlobalObjectBase)thisObj;
		if (cb.globals.debugFlag)
			cb.log(args);
	}

	private void log(Object[] args)
	{
		if (args == null || args.length == 0) {
			ln("");
			return;
		}
		String fmt = Context.toString(args[0]);
		StringBuilder sb = new StringBuilder(fmt.length() + 10);
		boolean hasWarned = false;
		int argIdx = 1;
		for (int i = 0; i < fmt.length(); i++) {
			char ch = fmt.charAt(i);
			if (ch != '%' || i == fmt.length() - 1) {
				sb.append(ch);
				continue;
			}
			char type = fmt.charAt(++i);
			switch (type) {
				case '%':
					sb.append(ch);
					break;
				case 'd':
				case 'i':
				case 'f':
					if (argIdx >= args.length) {
						sb.append('?');
						hasWarned = warnMissingArg(fmt, i, hasWarned);
						break;
					}
					Number n = Context.toNumber(args[argIdx++]);
					if (type == 'f')
						sb.append(n.doubleValue());
					else
						sb.append(n.longValue());
					break;
				case 's':
					if (argIdx >= args.length) {
						sb.append("%s");
						hasWarned = warnMissingArg(fmt, i, hasWarned);
						break;
					}
					sb.append(Context.toString(args[argIdx++]));
					break;
				case 'o':
					if (argIdx >= args.length) {
						sb.append("%o");
						hasWarned = warnMissingArg(fmt, i, hasWarned);
						break;
					}
					sb.append(jsFunction_stringify(args[argIdx++]));
					break;
				default:
					sb.append('%');
					sb.append(ch);
					makeLogFormatWarning(fmt, i, "Invalid % formatter char:\n");
			}
		}
		if (globals.debugFlag && argIdx < args.length) {
			makeLogFormatWarning(fmt, -1,
					"More arguments (" + args.length + ") than % specifications (" + argIdx + "):\n");
		}
		ln(sb.toString());
	}

	private boolean warnMissingArg(String fmt, int fmtIndex, boolean hasWarned)
	{
		if (!hasWarned)
			makeLogFormatWarning(fmt, fmtIndex, "Too few parameters in gb.log format string:\n");
		return true;
	}

	private void makeLogFormatWarning(String fmt, int fmtIndex, String msg)
	{
		if (!globals.debugFlag)
			return;
		StringBuilder sb = new StringBuilder(fmt.length() * 3);
		sb.append(msg);
		sb.append("  gb.log('");
		int slashes = 0;
		for (int i = 0; i < fmt.length(); i++) {
			char ch = fmt.charAt(i);
			if (ch == '\\' || ch == '\'') {
				sb.append('\\');
				if (i <= fmtIndex)
					slashes++;
			}
			sb.append(ch);
		}
		sb.append("', ...);");
		if (fmtIndex >= 0) {
			sb.append('\n');
			sb.append("  ........");
			for (int i = fmtIndex + slashes; i > 0; i--)
				sb.append('.');
			sb.append("^");
		}
		ln(sb.toString());
	}

	public synchronized String jsFunction_stringify(Object obj)
	{
		if (obj == null)
			return "null";
		if (stringify == null)
			stringify = ctx.compileFunction(scope, stringifySource, "RhinoStringify.js", 1, null);
		obj = stringify.call(ctx, scope, this, new Object[]{obj});
		return Context.toString(obj);
	}

	//=====================================================================================


	@RhinoCallback
	public String jsFunction_curl(String method, String url, Object headers) throws Exception
	{
		if (method == null || method.isEmpty())
			throw Context.reportRuntimeError("1st arg (method) must not be null or empty");
		if (url == null || url.isEmpty())
			throw Context.reportRuntimeError("2nd arg (url) must not be null or empty");
		URL uu;
		try {
			uu = new URL(getCurlBaseUrl(), url);
		}
		catch (Throwable ex) {
			throw Context.reportRuntimeError("2nd arg (url) is not a valid URL.");
		}
		String p = uu.getProtocol();
		if (!p.equals("http") && !p.equals("https"))
			throw Context.reportRuntimeError("2nd arg (url) must be 'http' or 'https' protocol and not '" + p + "'.");
		//-------
		HeadersBuilder hb = HeadersBuilder.makeEmpty();
		hb.setAllowGzip(true);
		boolean seenUA = false;
		if (headers instanceof NativeArray) {
			NativeArray aa = (NativeArray)headers;
			for (Object obj : aa) {
				String h = Context.toString(obj);
				if (h == null || h.isEmpty())
					continue;
				int idx = h.indexOf(':');
				if (idx < 0)
					continue;
				String nameLC = h.substring(0, idx).toLowerCase();
				hb.add(nameLC, h.substring(idx + 1).trim());
				seenUA |= nameLC.equals("user-agent");
			}
		} else if (headers != null && !(headers instanceof Undefined)) {
			throw Context.reportRuntimeError("3nd arg (headers) must be array of strings, null, or omitted.");
		}
		if (!seenUA )
			hb.add("user-agent", globals.serverUserAgent);
		ResultData rd = globals.curl(method, uu, hb.toHeaders(), logger);
		int status = rd.headers.status;
		if (status != HttpURLConnection.HTTP_OK)
			throw Context.reportRuntimeError("Got http status " + status + " from " + url);
		String ct = rd.headers.contentTypeBaseLC;
		if (!MimeTypeMatcher.AnyTextType.matches(rd.headers.status, ct))
			throw Context.reportRuntimeError("Response type is not text but " + ct);
		return rd.getContentAsText(logger, globals.debugVerbose);
	}

	@Nullable
	protected URL getCurlBaseUrl()
	{
		return null;
	}
}
