/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.util;

/**
 * Utilities for encoding strings (including null) as Javascript string literals.<br>
 * If the parameter is null, the methods returns "null". <br>
 * If the parameter is "Hello, world", the methods returns "'Hello, world'".
 * <p/>
 * The returned value includes the required javascript quotes and is 7-bit ascii clean.
 */
public class JsString
{
	/**
	 * Encode the string as a Javascript string literal including quotes (if not null).
	 *
	 * @param s The string to encode. Is allowed to be null, which returns a Javascript <code>null</code>.
	 */
	public static String encode(String s)
	{
		return encode(s, '\'');
	}

	/**
	 * Encode the string as a Javascript string literal including quotes (if not null).
	 *
	 * @param sb The encoded value is appended. The StringBuilder is not reset.
	 * @param s  The string to encode. Is allowed to be null, which returns a Javascript <code>null</code>.
	 */
	public static void encode(StringBuilder sb, String s)
	{
		encode(sb, s, '\'');
	}

	/**
	 * Encode the string as a Javascript string literal including quotes (if not null).
	 *
	 * @param quote Specifies the quoting character used. MUST BE either single quote or double quote character.
	 * @param s     The string to encode. Is allowed to be null, which returns a Javascript <code>null</code>.
	 */
	public static String encode(String s, char quote)
	{
		if (s == null)
			return "null";
		if (s.length() == 0) {
			if (quote == '\'')
				return "''";
			else if (quote == '"')
				return "\"\"";
		}
		StringBuilder sb = new StringBuilder(s.length() + 1);
		encode(sb, s, quote);
		return sb.toString();
	}

	/**
	 * Encode the string as a Javascript string literal including quotes (if not null).
	 *
	 * @param sb    The encoded value is appended. The StringBuilder is not reset.
	 * @param s     The string to encode. Is allowed to be null, which returns a Javascript <code>null</code>.
	 * @param quote Specifies the quoting character used. MUST BE either single quote or double quote character.
	 */
	public static void encode(StringBuilder sb, String s, char quote)
	{
		if (s == null) {
			sb.append("null");
			return;
		}
		int len = s.length();
		sb.ensureCapacity(len + 2);
		sb.append(quote);
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			switch (ch) {
				case '\r':
				case '\n':
					sb.append("\\n");
					break;
				case '\t':
					sb.append("\\t");
					break;
				case '"':
					sb.append("\\x22");
					break;
				case '\'':
					sb.append("\\x27");
					break;
				case '<':
					sb.append("\\x3c");
					break;
				case '>':
					sb.append("\\x3e");
					break;
				case '\\':
					sb.append("\\\\");
					break;
				default:
					if (ch > 255) {
						sb.append("\\u");
						sb.append(String.format("%04x", (int)ch));
					} else if (ch < ' ' || ch >= 127) {
						sb.append("\\x");
						sb.append(String.format("%02x", (int)ch));
					} else {
						sb.append(ch);
					}
			}
		}
		sb.append(quote);
	}
}