/* Copyright (C) 2008-2012 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "AlertHelper.h"
#import "GBDebug.h"
#import "AutoZap.h"

@interface AlertHelper ()

@property NSBundle* mainBundle;
@property NSView* windowView;
@property NSImage* glimmerIcon;
@property NSAlert* alert;
@property BOOL alertIsOpen;
@property BOOL usingSheet;
@property NSObject* callbackDelegate;
@property SEL callbackSelector;
@property (strong, readonly) NSView* accessoryView;
@property NSInteger lastAlertReturnCode;

- (void)releaseDelegate;
- (BOOL)checkHasPreparedAlert:(NSString*)doingWhat;


- (void)simpleAlertDidEnd:(NSAlert *)alert returnCode:(int)returnCode
			  contextInfo:(void *)contextInfo;

@end

//=======================================================================================================

@implementation AlertHelper

- (instancetype)initWindowView:(NSView*)windowView
				withMainBundle:(NSBundle*)mainBundle
{
	if ((self = [super init])) {
		_windowView = windowView;
		_mainBundle = mainBundle;
		NSString* iconPath = [_mainBundle pathForResource:@"tangerine" ofType:@"icns"];
		NSAssert(iconPath, @"Missing tangerine.icns");
		_glimmerIcon = [[NSImage alloc] initWithContentsOfFile:iconPath];
	}
	return self;
}


- (NSWindow*)window
{
	return [_windowView window];
}

- (BOOL)canDisplayAlert
{
	return !_alertIsOpen && [_windowView window];
}

#pragma mark ------------------------ private

- (void)releaseDelegate
{
	_callbackDelegate = NULL;
	_callbackSelector = NULL;
}

- (void)dismissAlert
{
	DebugNSLog(@"dismissAlert, open = %d", _alertIsOpen);
	[self dismissAlertWithInvokeDelegate:NO withReturnCode:0];
}

- (void)dismissAlertWithInvokeDelegate:(BOOL)callDelegate withReturnCode:(NSInteger)returnCode
{
	if (!_alertIsOpen)
		return;
	DebugNSLog(@"dismissAlertWithInvokeDelegate, isOpen=%d, callDelegate=%d, returnCode=%ld", _alertIsOpen, callDelegate, (long)returnCode);
	_alertIsOpen = NO; // otherwise the method is called recursively by simpleAlertDidEnd
	_lastAlertReturnCode = returnCode;
	/*
	 Make the alert sheet go away before the delegate is called, so the delegate is able to show another sheet.
	 Must call [NSApp endSheet] to get the sheet really gone, otherwise we can't show another sheet.
	 [orderOut] performs the animation.
	*/
	if (_usingSheet) {
		[NSApp endSheet:[_alert window]];
		[[_alert window] orderOut:self];
		_usingSheet = NO;
	}
	if (callDelegate && _callbackDelegate) {
		// http://theocacao.com/document.page/264
		NSMethodSignature* sig = [[_callbackDelegate class] instanceMethodSignatureForSelector:_callbackSelector];
		NSInvocation* invo = [NSInvocation invocationWithMethodSignature:sig];
		[invo setTarget:_callbackDelegate];
		[invo setSelector:_callbackSelector];
		[invo setArgument:&returnCode atIndex:2];
		[invo invoke];
		[self releaseDelegate];
	}
	[AutoZap add:_alert];
	_alert = NULL;
	[AutoZap add:_accessoryView];
	_accessoryView = NULL;
}

- (void)simpleAlertDidEnd:(NSAlert *)alert returnCode:(int)returnCode
			  contextInfo:(void *)contextInfo
{
	if (alert == _alert)
		[self dismissAlertWithInvokeDelegate:YES withReturnCode:returnCode];
}

- (BOOL)checkHasPreparedAlert:(NSString*)doingWhat
{
	if (!_alert) {
		NSLog(@"Can't %@ when no alert has been prepared.", doingWhat);
		[GBDebug debugDumpStackframe];
		return NO;
	}
	if (_alertIsOpen) {
		NSLog(@"Can't %@ when alert already is open.", doingWhat);
		[GBDebug debugDumpStackframe];
		return NO;
	}
	return YES;
}


#pragma mark ------------------------ building blocks

- (void)prepareWarningAlertWithTitle:(NSString*)title
			  withInformativeText:(NSString*)infoText
{
	[self prepareAlertWithTitle:title withInformativeText:infoText withStyle:NSWarningAlertStyle];
}

- (void)prepareCriticalAlertWithTitle:(NSString*)title
			   withInformativeText:(NSString*)infoText
{
	[self prepareAlertWithTitle:title withInformativeText:infoText withStyle:NSCriticalAlertStyle];
}

- (void)prepareInformativeAlertWithTitle:(NSString*)title
				  withInformativeText:(NSString*)infoText
{
	[self prepareAlertWithTitle:title withInformativeText:infoText withStyle:NSInformationalAlertStyle];
}

- (void)prepareAlertWithTitle:(NSString*)title
		  withInformativeText:(NSString*)infoText
					withStyle:(NSAlertStyle)alertStyle
{
	if (_alertIsOpen) {
		[GBDebug debugDumpStackframe];
		NSLog(@"Can't prepare another alert while an alert is open.");
		return;
	}
	if (!title) {
		[GBDebug debugDumpStackframe];
		NSLog(@"the title must not be NULL");
		title = @"?"; // better than crashing with NSAssert.
	}
	if (!_alert) {
		_alert = [[NSAlert alloc] init]; // release in delegate
		[_alert setIcon:_glimmerIcon];
	}
	[_alert setMessageText:title];
	if (infoText) // NSAlert can't handle NULL
		[_alert setInformativeText:infoText];
}

- (void)setDelegate:(NSObject*)callbackDelegate
	   withSelector:(SEL)callbackSelector
{
	NSAssert((_callbackDelegate != NULL) == (_callbackSelector != NULL), @"_callbackDelegate / _callbackSelector");
	if (![self checkHasPreparedAlert:@"set delegate"])
		return;
	[self releaseDelegate];
	_callbackDelegate = callbackDelegate;
	_callbackSelector = callbackSelector;
}

- (NSButton*)addButtonWithTitle:(NSString*)title
{
	if (!title) {
		[GBDebug debugDumpStackframe];
		NSLog(@"the button title must not be NULL");
		title = @"?"; // better than crashing with NSAssert.
	}
	if (![self checkHasPreparedAlert:@"add button"])
		return NULL;
	return [_alert addButtonWithTitle:title];
}

- (void)setAccessoryView:(NSView*)view
{
	if (!view) {
		[GBDebug debugDumpStackframe];
		NSLog(@"the view param must not be NULL");
		return;
	}
	if (![self checkHasPreparedAlert:@"set accessory view"])
		return;
	if (view == _accessoryView)
		return;
	_accessoryView = view;
	[_alert setAccessoryView:view];
}

- (void)layoutAlert
{
	if (![self checkHasPreparedAlert:@"layout alert"])
		return;
	[_alert layout];
}

- (void)showPreparedAlert
{
	if (![self checkHasPreparedAlert:@"show alert"])
		return;
	NSWindow* window = [self window];
	_alertIsOpen = YES;
	if (!window) {
		NSInteger returnCode = [_alert runModal];
		[self dismissAlertWithInvokeDelegate:YES withReturnCode:returnCode];
		return;
	}
	_usingSheet = YES;
	[_alert beginSheetModalForWindow:window completionHandler:^(NSModalResponse returnCode){
		[self dismissAlertWithInvokeDelegate:YES withReturnCode:returnCode];
	}];
}

@end
