/* Copyright (C) 2008-2012 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "AutoZap.h"
#include "GBDebug.h"

@interface AutoZap()

@property NSMutableArray *queue;
@property BOOL hasCallback;
@property NSRecursiveLock *lock;

@end

@implementation AutoZap

static AutoZap* zapper = NULL;
static NSRecursiveLock *stickyLock;
static NSMutableArray *stickyArray;

#pragma mark ------------------- sticky

+ (void)initGlobals
{
	if (stickyLock)
		return;
	stickyLock = [[NSRecursiveLock alloc] init];
	stickyArray = [NSMutableArray arrayWithCapacity:10];
}

+ (void)addSticky:(NSObject*)obj
{
	NSAssert(stickyLock, @"You forgot to call AutoZap.initGlobals");
	[stickyLock lock];
	[stickyArray addObject:obj];
	[stickyLock unlock];
	DebugNSLog(@"Sticky ADD of %@: %@", NSStringFromClass(obj.class), obj);
}

+ (void)zapSticky:(NSObject*)obj
{
	[stickyLock lock];
	NSUInteger idx = [stickyArray indexOfObject:obj];
	if (idx == NSNotFound) {
		[stickyLock unlock];
		NSLog(@"Sticky ZAP of UNKNOWN %@: %@", NSStringFromClass(obj.class), obj);
		NSAssert(NO, @"zapStick of unstucked obj");
		return;
	}
	DebugNSLog(@"Sticky ZAP of %@ -> %ld: %@", NSStringFromClass(obj.class), stickyArray.count - 1, obj);
	[AutoZap add:obj]; // zapSticky might be called from within an IBAction handler.
	[stickyArray removeObjectsInRange:NSMakeRange(idx, 1)];
	[stickyLock unlock];
}


#pragma mark ------------------- autorelease

- (instancetype)init
{
	if ((self = [super init])) {
		_queue = [[NSMutableArray alloc] init];
		_hasCallback = NO;
		_lock = [[NSRecursiveLock alloc] init];
	}
	return self;
}

- (void)flushQueue:(NSObject*)dummy
{
	[_lock lock];
	DebugNSLog(@"AutoZap: flushes %lu objects", (unsigned long)[_queue count]);
	_hasCallback = NO;
	if (![_queue count]) {
		[_lock unlock];
		return;
	}
	// AutoZap.add() can be called as result of releasing the queue objects,
	// so use seperate array to avoid having _queue being modified executing its removeAllObjects.
	NSMutableArray *aa = [NSMutableArray arrayWithCapacity:[_queue count]];
	[aa addObjectsFromArray:_queue];
	[_queue removeAllObjects];
	[_lock unlock];
	while ([aa count]) {
#ifdef DEBUG
		NSObject *obj = [aa lastObject];
		NSLog(@"AutoZap: zaps %@: %@", NSStringFromClass(obj.class), obj);
#endif
		[aa removeLastObject];
	}
}


- (void)addToQueue:(NSObject*)obj
{
	[_lock lock];
	DebugNSLog(@"AutoZap: Adds %@: %@", NSStringFromClass(obj.class), obj);
	[_queue addObject:obj];
	if (_hasCallback) {
		[_lock unlock];
		return;
	}
	DebugNSLog(@"AutoZap: Adds callback");
	_hasCallback = YES;
	[_lock unlock];
	[self performSelectorOnMainThread:@selector(flushQueue:)
						   withObject:self
						waitUntilDone:NO];
}

+ (void)add:(NSObject*)obj
{
	if (!obj)
		return;
	if (!zapper) {
		DebugNSLog(@"AutoZap: created singleton");
		zapper = [[AutoZap alloc] init];
	}
	[zapper addToQueue:obj];
	
}

#pragma mark ------------------- NIB

+ (NSArray*)loadNibNamed:(NSString*)name bundle:(NSBundle*)bundle owner:(id)owner
{
	NSAssert(owner, @"owner must not be null");
	NSAssert(bundle, @"bundle arg must not be null");
	NSArray *__autoreleasing tmp = NULL;
	BOOL ok = [bundle loadNibNamed:name owner:owner topLevelObjects:&tmp];
	NSAssert(ok, @"Failed loading nib named '%@' from bundle '%@'", name, bundle);
	return ok ? tmp : NULL;
}


@end
